// -*- Mode: c++ -*-
//Header:
//File: SimpleSexp2.cc
//Author: NODA, Itsuki
//Date: 2001/06/26
//

//ModifyHistory:
// 2001/06/26: Start to create this file
//EndModifyHistory:

/*
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 * Copyright (C) 1999, 2000 Itsuki Noda, Electrotechnical Laboratory, Japan
 */

#include "SimpleSexp2.h"

namespace Itk {

   //======================================================================
   // class static variables
   
   SimpleSexp2::Scanner SimpleSexp2::sharedScanner(cin) ;

   SimpleSexp2::Scanner 
   SimpleSexp2::eternalHeap(static_cast<istream *>(ITK_NULLPTR),ITK_NULLPTR,
						 False) ;

   SimpleSexp2::Heap SimpleSexp2::heap ;

   SimpleSexp2 SimpleSexp2::_Nil(T_Nil,SubString("()")) ;
   SimpleSexp2 * SimpleSexp2::Nil = &SimpleSexp2::_Nil;

   SimpleSexp2 SimpleSexp2::_Eof(T_Atom,SubString("eof")) ;
   SimpleSexp2 * SimpleSexp2::Eof = &_Eof ;

   SimpleSexp2 SimpleSexp2::_Error(T_Error,SubString("error")) ;
   SimpleSexp2 * SimpleSexp2::Error = &_Error ;

   //======================================================================
   // scan As Number

   Int SimpleSexp2::scanAsInt() const {
      Int v ; 
      scanAsIntTo(v) ;
      return v ;
   } ;

   Bool SimpleSexp2::scanAsIntTo(Int & val) const {
      if(isAtom()) {
	 char buf[128] ;
	 str.copyTo(buf) ;
	 char *endptr ;
	 val = strtol(buf,&endptr,0) ;
	 if(isNull(*endptr)) return True ;
      }
      return False ;
   } ;

   Flt SimpleSexp2::scanAsFlt() const {
      Flt v ; 
      scanAsFltTo(v) ;
      return v ;
   } ;

   Bool SimpleSexp2::scanAsFltTo(Flt & val) const {
      if(isAtom()) {
	 char buf[128] ;
	 str.copyTo(buf) ;
	 char *endptr ;
	 val = strtod(buf,&endptr) ;
	 if(isNull(*endptr)) return True ;
      }
      return False ;
   } ;

   SubString SimpleSexp2::scanAsSymbol() const { return str.dup() ; } ;

   SubString SimpleSexp2::scanAsString(Bool naked, Bool copyp) const {
      SubString r ;
      if(copyp) r.copy(str,True) ;
      else      r = str ;
      if(isString()) {
	 r.ignoreChar() ; // chop the first char
	 r.chop() ;       // chop the last char
      }
      return r ;
   } ;

   //--------------------------------------------------
   // copy to buffer

   void SimpleSexp2::copyTo(Buffer & buffer) const {
      buffer.put(str) ;
   }

   //======================================================================
   // list operations

   UInt SimpleSexp2::length() const {
      UInt l = 0 ;
      for(const SimpleSexp2 * p = this ; p->isCons() ; p = p->cdr()) {
	 l++ ;
      }
      return l ;
   } ;

   SimpleSexp2 * SimpleSexp2::nthcdr(UInt n) {
      SimpleSexp2 * r = this ;
      for(UInt i = 0 ; i < n ; i++) { 
	 if(isNullPtr() || !isCons()) {
	    r = ITK_NULLPTR ;
	    break ;
	 }
	 r = r->cdr() ; 
      } 
      return r ;
   } ;

   SimpleSexp2 * SimpleSexp2::nth(UInt n) {
      SimpleSexp2 * r = nthcdr(n) ;
      if(r->isCons()) return r->car() ;
      else return ITK_NULLPTR;
   } ;

   //======================================================================
   // table operations

   //------------------------------------------------------------
   // assoc

   SimpleSexp2 * SimpleSexp2::assoc(const SimpleSexp2 * const key) const {
      SimpleSexp2 * l = const_cast<SimpleSexp2*>(this) ;
      SimpleSexp2 * e = ITK_NULLPTR;
      for( ; l->isCons() ; l = l->rest()) {
	 e = l->first() ;
	 if(e->first()->equal(key)) break ;
      }
      if (l->isCons()) return e ;
      else return Nil ;
   } ;

   //------------------------------------------------------------
   // assoc by string

   SimpleSexp2 * SimpleSexp2::assoc(const SubString & key) const {
      SimpleSexp2 * l = const_cast<SimpleSexp2*>(this) ;
      SimpleSexp2 * e = ITK_NULLPTR;
      for( ; l->isCons() ; l = l->rest()) {
	 e = l->first() ;
	 if(e->first()->equal(key)) break ;
      }
      if (l->isCons()) return e ;
      else return Nil ;
   } ;

   //------------------------------------------------------------
   // assoc Value

   SimpleSexp2 * SimpleSexp2::assocValue(const SimpleSexp2 * const key) const {
      SimpleSexp2 * r = assoc(key) ;
      if(r->isCons()) return r->second() ;
      else return Nil ;
   } ;

   //------------------------------------------------------------
   // assoc Value by String

   SimpleSexp2 * SimpleSexp2::assocValue(const SubString& key) const {
      SimpleSexp2 * r = assoc(key) ;
      if(r->isCons()) return r->second() ;
      else return Nil ;
   } ;

   //------------------------------------------------------------
   // passoc (assoc property list)

   SimpleSexp2 * SimpleSexp2::passoc(const SimpleSexp2 * const key) const {
      SimpleSexp2 * l = const_cast<SimpleSexp2*>(this) ;
      for( ; l->isCons() ; l = l->rest()->rest()) {
	 if(l->first()->equal(key)) break ;
      }
      if (l->isCons()) return l->second() ;
      else return Nil ;
   } ;

   //------------------------------------------------------------
   // passoc (assoc property list) by string

   SimpleSexp2 * SimpleSexp2::passoc(const SubString & key) const {
      SimpleSexp2 * l = const_cast<SimpleSexp2*>(this) ;
      for( ; l->isCons() ; l = l->rest()->rest()) {
	 if(l->first()->equal(key)) break ;
      }
      if (l->isCons()) return l->second() ;
      else return Nil ;
   } ;

   //======================================================================
   // duplicate

   SimpleSexp2 * SimpleSexp2::dup(Scanner & scanner) const {
      if      (isNil()) {
	 return Nil ;
      }
      else if (isAtom()) {
	 return scanner.newAtom(const_cast<const char *>((char*)pname())) ;
      } 
      else if (isCons()) {
	 SimpleSexp2 * carval = car()->dup(scanner);
	 SimpleSexp2 * cdrval = cdr()->dup(scanner);
	 return cons(carval,cdrval,&scanner) ;
      }
      else {
	 ITK_ERR("illegal SimpleSexp2 object to duplicate:" << *this) ;
	 return Nil ;
      }
   } ;

   //======================================================================
   // describe

   void SimpleSexp2::describe(ostream& o, Bool detailp) const {
      if(detailp) {
	 o << "#SimpleSexp2[tag=" ;
	 switch(tag) {
	    case T_Atom : o << "atom" ; break ;
	    case T_Nil  : o << "nil" ; break ;
	    case T_Cons : o << "cons" ; break ;
	    case T_Error : o << "error" ; break ;
	    default      : o << "other" ; break ;
	 }
	 o << ",str=\"" << str << "\"]" << endl ;
	 if(tag == T_Cons) {
	    o << "	car=" ;
	    car()->describe(o,detailp) ;
	    o << "	cdr=" ;
	    cdr()->describe(o,detailp) ;
	 }
      } else {
	 switch(tag) {
	    case T_Atom : {
	       o << str ;
	       break ;
	    } ;
	    case T_Nil :
	    case T_Cons : {
	       o << ITK_SEXP2_CONS_BEGIN_CHAR ;
	       for(const SimpleSexp2 * sx = this ; sx->tag == T_Cons ; 
		   sx = sx->cdr()) {
		  if(sx != this) o << ITK_SEXP2_CONS_SEP_CHAR ;
		  sx->car()->describe(o,detailp) ;
	       }
	       o << ITK_SEXP2_CONS_END_CHAR ;
	       break ;
	    } ;
	    case T_Error : 
	    default : {
	       o << "#SimpleSexp2Error[" << str << "]" ;
	    }
	 }
      }
   } ;
      
} ;
