// -*- Mode: c++ -*-
//Header:
//File: SimpleSexp2.h
//Author: NODA, Itsuki
//Date: 2001/06/26
//

//ModifyHistory:
// 2001/06/26: Start to create this file
//EndModifyHistory:

/*
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 * Copyright (C) 1999, 2000 Itsuki Noda, Electrotechnical Laboratory, Japan
 */

#ifndef _itk_SimpleSexp2_h_
#define _itk_SimpleSexp2_h_
//////////////////////////////////////////////////////////////////////

#include "itk/btype.h"
#include "itk/utility.h"
#include "itk/TimeVal.h"
#include "itk/WithDescriber.h"
#include "itk/SString.h"
#include "itk/Buffer.h"
#include "itk/HeapT.h"
#include "itk/Mutex.h"
#include <fstream.h>
#include <ctype.h>
#include <strstream.h>


//----------------------------------------------------------------------
// char definitions

#define ITK_SEXP2_CONS_BEGIN_CHAR	'('
#define ITK_SEXP2_CONS_END_CHAR		')'

#define ITK_SEXP2_CONS_SEP_CHAR		' '

#define ITK_SEXP2_STRING_BEGIN_CHAR	'"'
#define ITK_SEXP2_STRING_END_CHAR	'"'
#define ITK_SEXP2_STRING_ESCAPE_CHAR	'\\'

#define ITK_SEXP2_COMMENT_BEGIN_CHAR	';'
#define ITK_SEXP2_COMMENT_END_CHAR	'\n'

//----------------------------------------------------------------------
// interface to bison and flex

namespace Itk { class SimpleSexp2 ;} ;

//#define YYSTYPE Itk::Int
#define YYSTYPE Itk::SimpleSexp2 *

extern int SimpleSexp2yyerror(char * s) ;
extern int SimpleSexp2yylex(YYSTYPE *lvalp, void *scanner) ;
extern int SimpleSexp2yyparse(void * scanner) ;

namespace Itk {


   class SimpleSexp2Lexer ;

   extern SimpleSexp2Lexer * newSimpleSexp2Lexer() ;
   extern void assignSimpleSexp2LexerScanner(SimpleSexp2Lexer * lexer,
					     void * scanner) ;
   
   //======================================================================
   // class SimpleSexp2

   class SimpleSexp2 : public WithDescriber {

	 //--------------------------------------------------
	 // local types

	 //----------------------------------------
	 // enum Tag

      public:
	 enum Tag {
	    T_Nil,
	    T_Atom,
	    T_Cons,
	    T_Error
	 } ;

	 //========================================
	 // class Heap & operations

      public:
	 class Heap : public HeapT<SimpleSexp2>, public Mutex { 
	    public:
	       SimpleSexp2 * newCell() { return getPtr() ; } ;
	 } ;
	 
	 static Heap heap ;

	 static SimpleSexp2 * get() {  return heap.getPtr() ; } ; // not used ?

	 static void cleanHeap() { heap.reset() ; } ; // not used ?

	 //========================================
	 // class Scanner

      public:
	 class Scanner : public Mutex {

	       //------------------------------
	       // typedef outside defs
	    public:
	       typedef SimpleSexp2Lexer Lexer ;

	       static Lexer * newLexer() { return newSimpleSexp2Lexer() ;} ;

	       //------------------------------
	       // members
	    public:
	       Lexer * lexer ;
	       Heap * heap ;
	       Bool useBufferP ;  /* use buffer or not for alloc string */
	       Buffer buffer ;
	       istream * istr ;
	       SimpleSexp2 * result ;

	       //------------------------------
	       // constructor
	    public:
	       Scanner(istream * istr = ITK_NULLPTR, Heap * h = ITK_NULLPTR,
		       Bool usebufp = True) {
		  init(istr,h,usebufp) ;
	       } ;
	    public:
	       Scanner(istream & istr, Heap  * h = ITK_NULLPTR,
		       Bool usebufp = True) {
		  init(istr,h,usebufp) ;
	       } ;

	    public:
	       Scanner(Buffer * b, Heap * h = ITK_NULLPTR,
		       Bool usebufp = True) {
		  init(b,h,usebufp) ;
	       } ;

	       Scanner(Buffer & b, Heap * h = ITK_NULLPTR,
		       Bool usebufp = True) {
		  init(b,h,usebufp) ;
	       } ;

	       //------------------------------
	       // init
	    public:
	       void init(istream * is = ITK_NULLPTR, Heap * h = ITK_NULLPTR,
			 Bool usebufp = True) {
		  lexer = newLexer() ;
		  assignSimpleSexp2LexerScanner(lexer,this) ;
		  istr = is ;
		  heap = h ;
		  if(isNull(heap)) heap = new Heap ;
		  useBufferP = usebufp ;
		  if(useBufferP) {
		     initWorkBuffer() ;
		  } else {
		     initWorkBuffer(0) ;
		  }
	       } ;
	       
	    public:
	       void init(istream & istr, Heap  * h = ITK_NULLPTR,
			 Bool usebufp = True) {
		  init(&istr,h,usebufp) ;
	       } ;

	    public:
	       void init(Buffer * b, Heap * h = ITK_NULLPTR,
			 Bool usebufp = True) {
		  istrstream * str = b->newIStrStream() ;
		  init(str,h,usebufp) ;
	       } ;

	       void init(Buffer & b, Heap * h = ITK_NULLPTR,
			 Bool usebufp = True) {
		  init(&b,h,usebufp) ;
	       } ;

	       //------------------------------
	       // reset
	    public:
	       void reset(istream * is) {
		  istr = is ;
	       } ;
		  
	       //------------------------------
	       // call parser

	    public:
	       SimpleSexp2 * scan(Bool resetp = True) {
		  lock() ; {
		     if(resetp) {
			clearWorkBuffer() ;
			heap->reset() ;
		     }
		     if(SimpleSexp2yyparse(this)) {
			if(!result->isEof()) {
			   ITK_WRN(ITK_HERE << 
				   "SimpleSexp2::scan(): yyparse error.") ;
			   result = Error ;
			} else {
			   // in case of EOF, do nothing
			}
		     } else {
			// in the nomal case, do nothing
		     } 
		  } ; unlock() ;
		  return result ;
	       } ;

	       //------------------------------
	       // Work Buffer Operation

	       //--------------------
	       // init
#define ITK_SSEXP2BUFSIZE 8192
	    public:
	       void initWorkBuffer(int size = ITK_SSEXP2BUFSIZE) {
		  resizeWorkBuffer(size) ;
	       } ;

	       //--------------------
	       // resize & reset
	    public:
	       void resizeWorkBuffer(int size) {
		  buffer.alloc(size) ;
		  buffer.clear() ;
	       } ;
	       
	       void clearWorkBuffer() {
		  buffer.clear() ;
	       } ;
	       
	       //--------------------
	       // get new work string
	       //   len = size - 1 
	    public:
	       SubString getWorkSString(Int len) {
		  if(useBufferP) 
		     return buffer.popSubString(len+1) ;
		  else 
		     return SubString(new char[len],len) ;
	       } ;

	    public:
	       char * getWorkCString(int len) { 
		  if(useBufferP) {
		     SubString s = getWorkSString(len) ;
		     s.terminate() ;
		     return (char *)s ;
		  } else {
		     return new char[len] ;
		  }
	       } ;

	       //------------------------------
	       // new Atom 
	    public:
	       SimpleSexp2 * newAtom(const char * text) {
		  SubString str = getWorkSString(strlen(text)) ;
		  str.copy(text) ;
		  return newAtom(str) ;
	       } ;

	    public:
	       SimpleSexp2 * newAtom(SubString str) {
		  return SimpleSexp2::newAtom(str,heap) ;
	       } ;

	       //------------------------------
	       // new Cons
	    public:
	       SimpleSexp2 * cons(SimpleSexp2 * car, SimpleSexp2 * cdr) {
		  return SimpleSexp2::cons(car,cdr,this) ;
	       } ;

	 } ;

	 //--------------------------------------------------
	 // Scanner Operation

      public:
	 static Scanner sharedScanner ;

	 static Scanner eternalHeap ; // used in eternal copy

	 //==================================================
	 //  class SimpleSexp2 body

	 //--------------------------------------------------
	 // static const
      public:
	 static SimpleSexp2  _Nil ;
	 static SimpleSexp2 * Nil ;

	 static SimpleSexp2  _Eof ;
	 static SimpleSexp2 * Eof ;

	 static SimpleSexp2  _Error ;
	 static SimpleSexp2 * Error ;

	 //--------------------------------------------------
	 // member

      public:
	 Tag tag ;
	 SubString str ;

	 SimpleSexp2 * _car ;	// used only if tag == T_Cons 
	 SimpleSexp2 * _cdr ;	// used only if tag == T_Cons 

	 //--------------------------------------------------
	 // constructor
      public:
	 SimpleSexp2() {} ;
	 SimpleSexp2(Tag t, const SubString & s) {
	    tag = t ;
	    str = s ;
	 } ;
	 SimpleSexp2(Tag t, const char * s) {
	    tag = t ;
	    str.copy(s) ;
	 } ;

	 //--------------------------------------------------
	 // scan (static operations) with scanner
      public:
	 static SimpleSexp2 * scan(Scanner * scanner, Bool resetp = True) {
	    return scanner->scan(resetp) ;
	 } ;
	 static SimpleSexp2 * scan(Scanner & scanner, Bool resetp = True) {
	    return scanner.scan(resetp) ;
	 } ;

	 //--------------------------------------------------
	 // scan (static operations) without scanner (using shared scanner)
      public:
	 static SimpleSexp2 * scan(istream * istr, Bool resetp = True) {
	    sharedScanner.reset(istr) ;
	    return sharedScanner.scan(resetp) ;
	 } ;

      public:
	 static SimpleSexp2 * scan(istream & istr, Bool resetp = True) {
	    return scan(&istr,resetp) ;
	 } ;

      public:
	 static SimpleSexp2 * scan(Buffer & buf, Bool resetp = True) {
	    istrstream str((const char *)buf.head,buf.bodyLength()) ;
	    return scan(str,resetp) ;
	 } ;
      public:
	 static SimpleSexp2 * scan(Buffer * buf, Bool resetp = True) {
	    return scan(*buf,resetp) ;
	 } ;

      public:
	 static SimpleSexp2 * scan(const char * str, Bool resetp = True) {
	    SubString s(const_cast<char *>(str)) ;
	    Buffer buf(s) ;
	    return scan(buf,resetp) ;

	    // BUG?
	    // If we use 'scan(Buffer(SubString(s)),resetp)', then
	    // it calls 'scan((char*)Buffer(SubString(s)),resetp)',
	    // and enter infinite loop.
	 } ;
	 
	 //--------------------------------------------------
	 // newatom (static operations)
      public:
	 static SimpleSexp2 * newAtom(SubString s, Heap * h = ITK_NULLPTR) {
	    if(isNull(h)) h = &heap ;
	    
	    SimpleSexp2 * cell = h->newCell() ;
	    cell->tag = T_Atom ;
	    cell->str.set(s) ;
	    return cell ;
	 } ;
      public:
	 static SimpleSexp2 * newAtom(char * s, Heap * h = ITK_NULLPTR) {
	    return newAtom(SubString(s),h) ;
	 } ;
	 
      public:
	 static SimpleSexp2 * newAtom(SubString s, Scanner * scanner) {
	    return newAtom(s,scanner->heap) ;
	 } ;

      public:
	 static SimpleSexp2 * newAtom(char * s, Scanner * scanner) {
	    return newAtom(SubString(s),scanner) ;
	 } ;

	 //--------------------------------------------------
	 // cons (static operations)
      public:
	 static SimpleSexp2 * cons(SimpleSexp2 * carVal,
				   SimpleSexp2 * cdrVal,
				   Heap * h = ITK_NULLPTR) {
	    if(isNull(h)) h = &heap ;

	    SimpleSexp2 * cell = h->newCell() ;
	    cell->tag = T_Cons ;
	    cell->str.set() ;
	    cell->_car = carVal ;
	    cell->_cdr = cdrVal ;
	    return cell ;
	 } ;

      public:
	 static SimpleSexp2 * cons(SimpleSexp2 * carVal,
				   SimpleSexp2 * cdrVal,
				   Scanner * scanner) {
	    return cons(carVal,cdrVal,scanner->heap) ;
	 } ;


	 //--------------------------------------------------
	 // type check

	 Bool isNullPtr() const { return isNull(this) ; } ;

	 Bool isNil() const { return !isNullPtr() && tag == T_Nil ; } ;
	 Bool isAtom() const { return !isNullPtr() && tag == T_Atom ; } ;
	 Bool isCons() const { return !isNullPtr() && tag == T_Cons ; } ;
	 Bool isError() const { return !isNullPtr() && tag == T_Error ; } ;
	 Bool isEof() const { return this == Eof ; } ;
	 
	 Bool isString() const { 
	    return isAtom() 
	       && (str.first() == ITK_SEXP2_STRING_BEGIN_CHAR) ;
	 } ;

	 Bool isInt() const {
	    Int dummy ;
	    return scanAsIntTo(dummy) ;
	 } ;

	 Bool isFlt() const {
	    Flt dummy ;
	    return scanAsFltTo(dummy) ;
	 } ;

	 //--------------------------------------------------
	 // scan As Number

	 Int scanAsInt() const ;

	 Bool scanAsIntTo(Int & val) const ;

	 Flt scanAsFlt() const ;

	 Bool scanAsFltTo(Flt & val) const ;
	 
	 //--------------------------------------------------
	 // scan As Symbol

	 SubString scanAsSymbol() const ;

	 const SubString & scanAsConstSymbol() const {
	    return str ;
	 } ;

	 //--------------------------------------------------
	 // scan As SubString (naked if it start quote mark)
	 
	 SubString scanAsString(Bool naked = True, Bool copyp = False) const ;

	 //--------------------------------------------------
	 // copy to buffer  // ??? work?
	 
	 void copyTo(Buffer & buffer) const ;

	 //------------------------------------------------------------
	 // list operations


	 //--------------------------------------------------
	 // basic
      public:
	 SimpleSexp2 * car() const { return _car ; } ;
	 SimpleSexp2 * cdr() const { return _cdr ; } ;

	 SimpleSexp2 * rplaca(SimpleSexp2 * v) { return _car = v ; } ;
	 SimpleSexp2 * rplacd(SimpleSexp2 * v) { return _cdr = v ; } ;

	 SimpleSexp2 * set(SimpleSexp2 * carval, SimpleSexp2 * cdrval) {
	    _car = carval ; _cdr = cdrval ;
	    return this ; 
	 } ;

	 SubString pname() const { return str ; } ;
	 
	 //--------------------------------------------------
	 // equal

	 Bool equal(const SimpleSexp2 * sexp) const { return equal(*sexp) ; } ;

	 Bool equal(const SimpleSexp2 & sexp) const {
	    if(isNullPtr() || isError()) return False ;
	    switch(tag) {
	       case T_Nil : return sexp.isNil() ; break ;
	       case T_Atom : return sexp.equal(str) ; break ;
	       case T_Cons : {
		  if(sexp.isCons()) {
		     return car()->equal(sexp.car()) 
			&&  cdr()->equal(sexp.cdr()) ;
		  } else {
		     return False ;
		  }
	       }
	       default : return False ;
	    }
	 } ;

	 Bool equal(const char * sym) const {
	    return !isNullPtr() && isAtom() && str == sym ;
	 } ;
	 Bool equal(const SubString & sym) const {
	    return !isNullPtr() && isAtom() && str == sym ;
	 } ;

	 //--------------------------------------------------
	 // length
      public:
	 UInt length() const ;

	 //--------------------------------------------------
	 // list operations
      public:

	 SimpleSexp2 * nthcdr(UInt n) ;

	 SimpleSexp2 * nth(UInt n) ;

	 SimpleSexp2 * first() { return nth(0) ; } ;
	 SimpleSexp2 * second() { return nth(1) ; } ;
	 SimpleSexp2 * rest() { return nthcdr(1) ; } ;
	 SimpleSexp2 * last() { return nthcdr(length()-1) ; } ;
	 SimpleSexp2 * lastone() { return nth(length()-1) ; } ;

	 //--------------------------------------------------
	 // table operation
      public:
	 SimpleSexp2 * assoc(const SimpleSexp2 * const key) const ;
	 SimpleSexp2 * assoc(const SubString & key) const ;
	 SimpleSexp2 * assocValue(const SimpleSexp2 * const key) const ;
	 SimpleSexp2 * assocValue(const SubString & key) const ;

	 SimpleSexp2 * passoc(const SimpleSexp2 * const key) const ;
	 SimpleSexp2 * passoc(const SubString & key) const ;

	 //--------------------------------------------------
	 // duplicate
      public:
	 SimpleSexp2 * dup(Scanner & scanner = sharedScanner) const ;

	 SimpleSexp2 * dup(Scanner * scanner = &sharedScanner) const {
	    return dup(*scanner) ;
	 } ;

	 SimpleSexp2 * dupEternal(Scanner & scanner = eternalHeap) const {
	    return dup(eternalHeap) ;
	 } ;

	 //--------------------------------------------------
	 // describe

      public:
	 virtual void describe(ostream& o, Bool detailp = True) const ;
	 virtual void describePtr(ostream& o, Bool detailp = True) const {
	    describe(o,detailp) ;
	 } ;
   } ;


} ;	 
	 
	    

//////////////////////////////////////////////////////////////////////
#endif














