// -*- Mode: c++ -*-
//Header:
//File: Thread.h
//Author: NODA, Itsuki
//Date: 1999/05/18
//

//ModifyHistory:
// 1999/05/18: Start to create this file
// 1999/11/07: introduce Itk namespace
// 1999/11/07: Divide each class to separate file
//EndModifyHistory:

/*
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 * Copyright (C) 1999, 2000 Itsuki Noda, Electrotechnical Laboratory, Japan
 */

#ifndef _itk_Thread_h_
#define _itk_Thread_h_
//////////////////////////////////////////////////////////////////////

//======================================================================
// includes

extern "C" {
#ifdef THREADED
#include <pthread.h>
#endif
}


#include "itk/btype.h"
#include "itk/WithDescriber.h"
#include "itk/MutexCondVar.h"

namespace Itk {
   
//======================================================================
// Thread

   extern void* defaultStartFunc(void* p) ;

   class Thread {
      public:
#ifdef THREADED
	 pthread_t pid ;
#endif	 
	 MutexCondVar starter ;
	 void* (*func)(void*) ;
	 void* arg ;
	 
      public:
	 Thread() { init() ; } ;
	 Thread(void*(*func)(void*),void* arg) { init(func,arg) ; } ;

	 ~Thread() { destroy() ; } ;

      public:
	 void init() {
	    startByWaitUntilStarter() ;
	 } ;
	 void init(void*(*f)(void*), void* a) {
	    func = f ;
	    arg = a ;
#ifdef THREADED
	    pthread_create(&pid,ITK_NULLPTR,func,arg) ;
#endif
	 } ;

      public:
	 void destroy() {
	 } ;

      public:
	 void start(void*(*f)(void*),void* a) {
	    starter.lock();
	    func = f ;
	    arg = a ;
	    starter.signal() ;
	    starter.unlock() ;
	 } ;
	 
      public:
	 void startByWaitUntilStarter() {
	    starter.lock() ;
	    init(defaultStartFunc,this) ;
	    starter.wait() ;	// wait until the new Thread is ready
	    starter.unlock() ;
	 } ;

      public:
	 void* waitUntilStarter() {
	    starter.lock() ;
	    starter.signal() ;	// inform the Thread is ready
	    starter.wait() ;	// wait starter 
	    starter.unlock() ;

	    return (*func)(arg) ;
	 } ;
   } ;
};

//////////////////////////////////////////////////////////////////////
#endif




