// -*- Mode:C++ -*-
//Header:
//File: BpLLink.cc
//Author: NODA, Itsuki
//Date: 1999/12/31
//

//ModifyHistory:
// 1999/12/31: Start to create this file
//EndModifyHistory:

#include "BpLLink.h"

namespace Nnn {

   //======================================================================
   // Execition

   //----------------------------------------------------------------------
   // forward propagation

   //--------------------------------------------------
   // propagate forward

   void BpLLink::propFore() {
      for(Index i = 0 ; i < n ; i++) {
	 Val & dst = to->unit(i).inp ;
	 for(Index j = 0 ; j < m ; j++) {
	    dst += from->unit(j).out * link(i,j).w ;
	 }
      }
   }

   //--------------------------------------------------
   // propagate backward

   void BpLLink::propBack() {
      for(Index j = 0 ; j < m ; j++) {
	 Val & dst = from->unit(j).err ;
	 for(Index i = 0 ; i < n ; i++) {
	    dst += to->unit(i).dlt * link(i,j).w ;
	 }
      }
   }

   //--------------------------------------------------
   // learn

   Val BpLLink::learn(const Val & eps, const Val & alpha) {
      if(lrate == 0.0) return 0.0 ;	// for speed-up

      Val e = eps * lrate ;

      Val dsum = 0.0 ;
      for(Index i = 0 ; i < n ; i++) {
	 for(Index j = 0 ; j < m ; j++) {
	    Val & dw = link(i,j).dw ;
	    dw *= alpha ;
	    dw += e * to->unit(i).dlt * from->unit(j).out ;
	    dsum += sq(dw) ;
	    link(i,j).w  += dw ;
	 }
      }
      return dsum ;
   } ;

   //======================================================================
   // Management

   //----------------------------------------------------------------------
   // fixFromTo

   Bool BpLLink::fixFromTo(const BpLayerTable & layerTable) {
      from = layerTable.find(fromname) ;
      to = layerTable.find(toname) ;
      if(!isNull(from) && !isNull(to)) {
	 resize(to->n,from->n) ;
	 return True ;
      } else {
	 return False ;
      } 
   } 

   //----------------------------------------------------------------------
   // registerFromTo (register link information into From/To layer)

   void BpLLink::registerFromTo() {
      if(!isNull(from) && !isNull(to)) {
	 from->addToLink(this) ;
	 to->addFromLink(this) ;
      } else {
	 ITK_ERR("can do 'BpLLink::registerFromTo(): from=" << from 
		 << ", to=" << to) ;
	 throw 0 ;
      } 
      return ;
   } 

   //----------------------------------------------------------------------
   // randomize weight
   
   void BpLLink::randomizeWeight(const Val & minval, const Val & maxval) {
      randomizeByTime() ;
      for(Index i = 0 ; i < n ; i++) {
	 for(Index j = 0 ; j < m ; j++) {
	    link(i,j).w = fltRand(minval,maxval) ;
	 }
      }
   } 

   //----------------------------------------------------------------------
   // fixEquiv

   Bool BpLLink::fixEquiv(const BpLLinkTable & linkTable) {
      if(isNull(equivInfo)) return False ;

      SString f, t ;
      if(!equivInfo->assocCvtSString(sformFromLbl,f) ||
	 !equivInfo->assocCvtSString(sformToLbl,t)) {
	 ITK_ERR("Illegal equiv form") ;
	 throw 0 ;
      } ;
      BpLLink * lnk = linkTable.find(f,t) ;
      if(isNull(lnk)) {
	 ITK_ERR("Unknown equiv link") ;
	 throw 0 ;
      } ;
      
      setEquiv(lnk) ;
      
      return True ;
   } 

   //----------------------------------------------------------------------
   // setEquiv

   void BpLLink::setEquiv(BpLLink * lnk) {
      equiv = lnk ;
      _link = lnk->_link ;
   } 

   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void BpLLink::save(ostream & ostr,
		      const Bool & newlinep,
		      const SubString & indent) const {
      SubString indent2 = indent + "  " ;
      //SubString indent3 = indent2 + "  " ;

      StrForm::writeBeginTerm(ostr,sformTermHead) ;

      StrForm::writeAssocRole(ostr,sformFromLbl) ; 
      StrForm::writeString(ostr,fromname) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,sformToLbl) ; 
      StrForm::writeString(ostr,toname) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,sformToSizeLbl) ; 
      ostr << n << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,sformFromSizeLbl) ; 
      ostr << m << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,sformLrnRateLbl) ; 
      ostr << NNN_VAL_FORMAT << lrate << ITK_SST_TERM_SEP ;

      ostr << endl << indent2 ;
      if(isNull(equiv)) {
	 StrForm::writeAssocRole(ostr,sformLinkLbl) ; {
	    SubString indent4 = indent2 + "        " ;
	    StrForm::writeBeginTerm(ostr) ;
	    StrForm::writeBeginList(ostr) ;
	    for(Index i = 0 ; i < n ; i++) {
	       if(i > 0) {
		  StrForm::writeArgSep(ostr) ;
		  ostr << endl << endl << indent4 ; // sep
	       }
	       for(Index j = 0 ; j < m ; j++) {
		  if(j > 0) { StrForm::writeArgSep(ostr) ; } 
		  if(j > 0 && j%4 == 0) 
		     ostr << endl << indent4 ; // indent
		  link(i,j).save(ostr) ;
	       }
	    }
	    StrForm::writeEndTerm(ostr) ;
	 }
      } else {
	 StrForm::writeAssocRole(ostr,sformEquivLbl) ;
	 StrForm::writeBeginTerm(ostr) ;

	 StrForm::writeAssocRole(ostr,sformFromLbl) ;
	 StrForm::writeString(ostr,equiv->fromname) ;
	 StrForm::writeArgSep(ostr) ;

	 StrForm::writeAssocRole(ostr,sformToLbl) ; 
	 StrForm::writeString(ostr,equiv->toname) ;
	 StrForm::writeEndTerm(ostr) ;
      }
	 
      //ostr << endl << indent ;
      StrForm::writeEndTerm(ostr) ;
      if(newlinep) ostr << endl ;
   }
   
   //----------------------------------------------------------------------
   // scan

   void BpLLink::scan(StrForm & sform, 
		      BpLayerTable * layerTable,
		      BpLLinkTable * linkTable) {
      if(!sform.isTerm(sformTermHead)) {
	 ITK_ERR("Illegal BpLLink StrForm:" << sform) ;
	 throw 0 ;
      } 

      sform.assocCvtSString(sformFromLbl,fromname) ;
      sform.assocCvtSString(sformToLbl,toname) ;
      if(!isNull(layerTable)) {
	 fixFromTo(*layerTable) ;
      } ;

      UInt sizeN, sizeM ;
      if(sform.assocCvtUInt(sformToSizeLbl,sizeN) &&
	 sform.assocCvtUInt(sformFromSizeLbl,sizeM)) {
	 resize(sizeN,sizeM) ;
      } else if (!isNull(from) && !isNull(to)) {
	 resize(to->n, from->n) ;
      } 

      sform.assocCvtFlt(sformLrnRateLbl,lrate) ;

      StrForm * linklist = sform.assoc(sformLinkLbl) ;
      if(!isNull(linklist)) {
	 if(!isNull(_link)) {
	    for(Index i = 0 ; i < n ; i++) {
	       for(Index j = 0 ; j < m ; j++) {
		  link(i,j).scan(*(linklist->nthListArg(_idx(i,j)))) ;
	       }
	    }
	 } 
      }

      initInfo = ITK_NULLPTR ;
      sform.assocCvtStrFormPtr(sformInitLbl,initInfo) ;
      if(!isNull(initInfo) && !isNull(_link)) {
	 if(initInfo->isTerm("rand")) {
	    randomizeWeight(initInfo->nthListArg(0)->cvtFlt(),
			    initInfo->nthListArg(1)->cvtFlt()) ;
	 } else {
	    ITK_ERR("illegal init value specifier" << *initInfo) ;
	    throw 0 ;
	 }
	 initInfo = ITK_NULLPTR ;
      }

      equivInfo = ITK_NULLPTR ;
      sform.assocCvtStrFormPtr(sformEquivLbl,equivInfo) ;
      if(!isNull(equivInfo) && !isNull(linkTable)) {
	 fixEquiv(*linkTable) ;
      }
   } ;

} ;








