// -*- Mode:C++ -*-
//Header:
//File: BpNet3SigLyr.h
//Title: BpNet with 3 sigmoid layers
//Author: NODA, Itsuki
//Date: 2000/01/03
//

//ModifyHistory:
// 2000/01/03: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_BpNet3SigLyr_h_
#define _Nnn_BpNet3SigLyr_h_
////////////////////////////////////////////////////////////////////////

#include "BpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // BpNet3SigLyr

   class BpNet3SigLyr : public BpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 static char* const _sformTermHead = "BpNet3SigLyr" ; 
	 virtual char* sformTermHead() const {
	    return _sformTermHead ;
	 } ;

	 static char * const nameInput = "input" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameOutput = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 BpLayer * lyrInput ;
	 BpLayer * lyrHidden ;
	 BpLayer * lyrOutput ;

	 BpLLink * lnkInpHid ;
	 BpLLink * lnkHidOut ;

	 //--------------------------------------------------
	 // constructor
      public:
	 BpNet3SigLyr() { init() ; } ;
	 BpNet3SigLyr(const SString & nm,
		      const UInt & isize,
		      const UInt & hsize,
		      const UInt & osize,
		      const Val & tempval = NNN_DFLT_TEMP,
		      const Val & epsval = NNN_DFLT_EPS,
		      const Val & alphaval = NNN_DFLT_ALPHA) { 
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(isize,hsize,osize) ;
	 } ;

	 //--------------------------------------------------
	 // assignLayer
      public:
	 void assignLayerLLink(const UInt & isize, 
			       const UInt & hsize, 
			       const UInt & osize) {
	    lyrInput = newLayer(nameInput,isize,BpLayer::FT_linear) ;
	    lyrHidden = newLayer(nameHidden,hsize,BpLayer::FT_sigmoid) ;
	    lyrOutput = newLayer(nameOutput,osize,BpLayer::FT_sigmoid) ;
	    
	    lnkInpHid = newLLink(lyrInput,lyrHidden) ;
	    lnkHidOut = newLLink(lyrHidden,lyrOutput) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(const ValVector & inpvec) {
	    lyrInput->propFore(inpvec,temp) ;
	    lyrHidden->propFore(temp) ;
	    lyrOutput->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    lyrOutput->propBack(tchvec,temp) ;
	    lyrHidden->propBack(temp) ;
	    // lyrInput->propBack(inpvec,temp) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Flt learn() {
	    Val r = 0.0 ;
	    r += lyrHidden->learn(eps,alpha) ;
	    r += lyrOutput->learn(eps,alpha) ;
	    r += lnkInpHid->learn(eps,alpha) ;
	    r += lnkHidOut->learn(eps,alpha) ;
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return lyrOutput->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    lyrOutput->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    BpNetBase::scan(sform) ;
	    lyrInput = layer(nameInput) ;
	    lyrHidden = layer(nameHidden) ;
	    lyrOutput = layer(nameOutput) ;
	    lnkInpHid = link(lyrInput,lyrHidden) ;
	    lnkHidOut = link(lyrHidden,lyrOutput) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


