// -*- Mode:C++ -*-
//Header:
//File: Elman3SigLyr.h
//Title: Elman Net with 3 sigmoid layers
//Author: NODA, Itsuki
//Date: 2000/01/03
//

//ModifyHistory:
// 2000/01/03: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_Elman3SigLyr_h_
#define _Nnn_Elman3SigLyr_h_
////////////////////////////////////////////////////////////////////////

#include "BpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // Elman 3 Sigmoid Layer

   class Elman3SigLyr : public BpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 static char* const _sformTermHead = "Elman3SigLyr" ; 
	 virtual char* sformTermHead() const {
	    return _sformTermHead ;
	 } ;

	 static char * const nameContext = "context" ;
	 static char * const nameInput = "input" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameOutput = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 BpLayer * lyrContext ;
	 BpLayer * lyrInput ;
	 BpLayer * lyrHidden ;
	 BpLayer * lyrOutput ;

	 BpLLink * lnkConHid ;
	 BpLLink * lnkInpHid ;
	 BpLLink * lnkHidOut ;

	 ValVector * ctxvec ;

	 //--------------------------------------------------
	 // constructor
      public:
	 Elman3SigLyr() : ctxvec() { init() ; } ;
	 Elman3SigLyr(const SString & nm,
		      const UInt & isize,
		      const UInt & hsize,
		      const UInt & osize,
		      const Val & tempval = NNN_DFLT_TEMP,
		      const Val & epsval = NNN_DFLT_EPS,
		      const Val & alphaval = NNN_DFLT_ALPHA) : ctxvec() { 
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(isize,hsize,osize) ;
	 } ;

	 //--------------------------------------------------
	 // assignLayer
      public:
	 void assignLayerLLink(const UInt & isize, 
			       const UInt & hsize, 
			       const UInt & osize) {
	    lyrContext = newLayer(nameContext,hsize,BpLayer::FT_linear) ;
	    lyrInput = newLayer(nameInput,isize,BpLayer::FT_linear) ;
	    lyrHidden = newLayer(nameHidden,hsize,BpLayer::FT_sigmoid) ;
	    lyrOutput = newLayer(nameOutput,osize,BpLayer::FT_sigmoid) ;
	    
	    lnkConHid = newLLink(lyrContext,lyrHidden) ;
	    lnkInpHid = newLLink(lyrInput,lyrHidden) ;
	    lnkHidOut = newLLink(lyrHidden,lyrOutput) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(const ValVector & inpvec) {
	    if(isNull(ctxvec)) ctxvec = new ValVector(lyrHidden->n) ;

	    lyrHidden->getOut(ctxvec) ;
	    lyrContext->setOut(ctxvec) ;
	    lyrInput->propFore(inpvec,temp) ;
	    lyrHidden->propFore(temp) ;
	    lyrOutput->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    lyrOutput->propBack(tchvec,temp) ;
	    lyrHidden->propBack(temp) ;
	    // lyrInput->propBack(inpvec,temp) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Flt learn() {
	    Val r = 0.0 ;
	    r += lyrHidden->learn(eps,alpha) ;
	    r += lyrOutput->learn(eps,alpha) ;

	    r += lnkConHid->learn(eps,alpha) ;
	    r += lnkInpHid->learn(eps,alpha) ;
	    r += lnkHidOut->learn(eps,alpha) ;
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // run
      public:
	 void run(const ValVector & inpvec) { return propFore(inpvec) ; } ;
	 void runLearn(const ValVector & inpvec,
		       const ValVector & tchvec) {
	    propFore(inpvec) ;
	    propBack(tchvec) ;
	    learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return lyrOutput->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    lyrOutput->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    BpNetBase::scan(sform) ;
	    lyrContext = layer(nameContext) ;
	    lyrInput = layer(nameInput) ;
	    lyrHidden = layer(nameHidden) ;
	    lyrOutput = layer(nameOutput) ;

	    lnkConHid = link(lyrContext,lyrHidden) ;
	    lnkInpHid = link(lyrInput,lyrHidden) ;
	    lnkHidOut = link(lyrHidden,lyrOutput) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


