// -*- Mode:C++ -*-
//Header:
//File: VBpLLink.h
//Author: NODA, Itsuki
//Date: 2000/01/12
//

//ModifyHistory:
// 2000/01/12: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_VBpLLink_h_
#define _Nnn_VBpLLink_h_
////////////////////////////////////////////////////////////////////////

#include "n3btype.h"
#include "n3utility.h"
#include "StrForm.h"
#include "VBpUnitWrap.h"
#include "VBpLinkWrap.h"
#include "VBpLayer.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VBpLLink (Layer-Link)

   class VBpLLinkTable ;

   class VBpLLink : public WithDescriber {

	 //--------------------------------------------------
	 // label string for save/load/scan
      public:
	 static char* const LblTermHead = "VBpLLink" ; 
	 static char* const LblFrom = "from" ; 
	 static char* const LblTo = "to" ; 
	 static char* const LblToSize = "n" ; 
	 static char* const LblFromSize = "m" ; 
	 static char* const LblLrnRate = "lrate" ; 
	 static char* const LblEquiv = "equiv" ;
	 static char* const LblNorm = "norm" ;   // normalize parameter
	 static char* const LblDecay = "decay" ;   // decay parameter
	 static char* const LblW = "w" ;
	 static char* const LblDw = "dw" ;
	 static char* const LblRand = "rand" ; 
	 static char* const LblLayer = "layer" ;
	 static char* const LblLLink = "llink" ;
	 static char* const LblObj = "obj" ;
	 static char* const LblVec = "vec" ;


	 //--------------------------------------------------
	 // member
      public:
	 SubString fromname ;
	 SubString toname ;
	 VBpLayer * from ;
	 VBpLayer * to ;
	 Size n ;
	 Size m ;
	 StrForm * initInfo ;
	 VBpLLink * wEquivLink ;
	 StrForm * wEquivInfo ;
	 Val lrate ; // rate of learning. multiplied with eps
	 Val norm ;  // normalize parameter.  if not positive, no normalize
	 Val decay ; // decay weight value every learnning

	 //VBpLinkWrap *_link ;
	 ValVector *_w ;
	 ValVector *_dw ;

	 //--------------------------------------------------
	 // constructor
      public:
	 VBpLLink() { init() ; } ;
	 VBpLLink(VBpLayer * fromLayer, VBpLayer * toLayer) {
	    init(fromLayer,toLayer) ;
	 } ;

	 //--------------------------------------------------
	 // initialize
      public:
	 void init(VBpLayer * fromLayer = ITK_NULLPTR,
		   VBpLayer * toLayer = ITK_NULLPTR,
		   Bool allocp = True) {
	    n = 0 ; m = 0 ;
	    from = fromLayer ;
	    to = toLayer ;
	    UInt sizeN = 0 ; UInt sizeM = 0 ; 
	    if(!isNull(from)) {
	       fromname = from->name ;
	       sizeM = from->n ;
	    } 
	    if(!isNull(to)) {
	       toname = to->name ;
	       sizeN = to->n ;
	    }
	    if(allocp) {
	       if(sizeN * sizeM > 0) {
		  resize(sizeN,sizeM,False) ;
	       } else {
		  assignVector(ITK_NULLPTR,ITK_NULLPTR) ;
	       } 
	    } else {
	       assignVector(ITK_NULLPTR,ITK_NULLPTR) ;
	    }
	    lrate = 1.0 ;
	    decay = 0.0 ;
	    initInfo = ITK_NULLPTR ;
	    wEquivLink = ITK_NULLPTR ;
	    wEquivInfo = ITK_NULLPTR ;
	 } ;

	 //--------------------------------------------------
	 // assign vectors
      public:
	 void assignVector(ValVector * wv, ValVector * dwv) {
	    _w = wv ; _dw = dwv ;
	 } ;
	 void assignVector(ValVector & wv, ValVector & dwv) {
	    _w = &wv ; _dw = &dwv ;
	 } ;

	 //--------------------------------------------------
	 // resize
      public:
	 Bool resize(Size newN, Size newM, Bool deletep = True) {
	    Size oldN = n ;
	    Size oldM = m ;
	    n = newN ; m = newM ;
	    if(oldN * oldM > newN * newM) return False ;
	    
	    if(deletep) { delete _w ; delete _dw ; }
	    assignVector(new ValVector(n * m), new ValVector(n * m)) ;
	    return True ;
	 } ;

	 //--------------------------------------------------
	 // actual index 
      public:
	 Index _idx(const Index & i) const { return i * m ; } ;
      public:
	 Index _idx(const Index & i, const Index & j) const {
	    return (_idx(i) + j) ;
	 } ;

	 //--------------------------------------------------
	 // access
      public:
	 VBpLinkWrap link(const Index & i, const Index & j) { 
	    return VBpLinkWrap(w(i,j),dw(i,j)) ;
	 } ;

	 Val & w(const Index & i, const Index & j) {
	    return w()[_idx(i,j)] ;
	 } ;
	 const Val & w(const Index & i, const Index & j) const {
	    return w()[_idx(i,j)] ;
	 } ;
	 Val & dw(const Index & i, const Index & j) {
	    return dw()[_idx(i,j)] ;
	 } ;
	 const Val & dw(const Index & i, const Index & j) const {
	    return dw()[_idx(i,j)] ;
	 } ;

	 ValVector & w() { return *_w ; } ;
	 const ValVector & w() const { return *_w ; } ;

	 ValVector & dw() { return *_dw ; } ;
	 const ValVector & dw() const { return *_w ; } ;

	 //--------------------------------------------------
	 // generic access
      public:
	 const ValVector & vec(const SubString & vecname) const {
	    if(vecname == LblW) return w() ;
	    if(vecname == LblDw) return dw() ;
	    
	    ITK_ERR("Illeval vector name:" << vecname) ;
	    throw 0 ;
	 } ;
	 ValVector & vec(const SubString & vecname) {
	    if(vecname == LblW) return w() ;
	    if(vecname == LblDw) return dw() ;
	    
	    ITK_ERR("Illeval vector name:" << vecname) ;
	    throw 0 ;
	 } ;

	 //--------------------------------------------------
	 // processing
      public:
	 void propFore() ;
	 void propBack() ;

	 //--------------------------------------------------
	 // learning
      public:
	 Val learn(const Val & eps, const Val & alpha) ;

	 //--------------------------------------------------
	 // randomize weight 
      public:
	 void randomizeWeight(const Val & minval, const Val & maxval) ;

	 //--------------------------------------------------
	 // normalize weight 
      public:
	 void normalizeWeight(const Val & norm) ;

	 //--------------------------------------------------
	 // save 
      public:
	 virtual void save(ostream & ostr,
			   SaveLevel slevel = SL_NetState,
			   const Bool & newlinep = True,
			   const SubString & indent = SubString("")) const ;

	 //--------------------------------------------------
	 // load
      public:
	 virtual void load(istream & istr, 
			   VBpLayerTable * layerTable = ITK_NULLPTR,
			   VBpLLinkTable * linkTable = ITK_NULLPTR) {
	    StrForm * sform = StrForm::scan(istr) ;
	    scan(*sform, layerTable, linkTable) ;
	    delete sform ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform,
			   VBpLayerTable * layerTable = ITK_NULLPTR,
			   VBpLLinkTable * linkTable = ITK_NULLPTR) ;

	 //--------------------------------------------------
	 // duplicate  & copy
	 //   if sharedp == True, then share weight matrix
      public:
	 VBpLLink & duplicate(const Bool sharedp = False,
			      const VBpLayerTable * lytab = ITK_NULLPTR) const;

      public:
	 VBpLLink & copyTo(VBpLLink & dest, 
			   const Bool sharedp = False,
			   const Bool allparamp = False,
			   const VBpLayerTable * lytab = ITK_NULLPTR ) const {
	    return dest.copyFrom(*this,sharedp,allparamp,lytab) ;
	 } ;
	 VBpLLink & copyFrom(const VBpLLink & src, 
			     const Bool sharedp = False,
			     const Bool allparamp = False,
			     const VBpLayerTable * lytab = ITK_NULLPTR) ;

	 void copyMiscParamFrom(const VBpLLink & src) ;
	 void copyWeightFrom(const VBpLLink & src, 
			     const Bool sharedp = False) ;

	 //--------------------------------------------------
	 // fix from&to layer
      public:
	 Bool fixFromTo(const VBpLayerTable & layerTable) ;

	 //--------------------------------------------------
	 // registerFromTo (register link information into From/To layer)
      public:
	 void registerFromTo() ;

	 //--------------------------------------------------
	 // fix Equiv Link
      public:
	 Bool fixEquiv(const VBpLLinkTable & linkTable) ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#VBpLLink[" 
		 << (isNull(from) ? "(null)" : (const char *)from->name)
		 << "->"
		 << (isNull(to) ? "(null)" : (const char *)to->name)
		 << "]" ;

	    if(detailp) {
	       ostr << endl ;

	       ostr << "  " << "from=" ;
	       if(isNull(from)) ostr << "(null)" ;
	       else             ostr << *from ;
	       ostr << endl ;

	       ostr << "  " << "to=" ;
	       if(isNull(from)) ostr << "(null)" ;
	       else             ostr << *to ;
	       ostr << endl ;

	       if(!isNull(wEquivInfo)) {
		  ostr << "  " << "wEquiv=" << *wEquivInfo << endl ;
	       }

	       ostr << "  " << "size=(" << n << "," << m << ")" << endl ;
	       ostr << "  " << "w=(" ;
	       if(isNull(_w)) ostr << "null" ;
	       else {
		  for(Index i = 0 ; i < n ; i++) {
		     if(i > 0) ostr << endl << "     ";
		     for(Index j = 0 ; j < m ; j++) {
			if(j > 0) ostr << ", " ;
			ostr << NNN_VAL_FORMAT << w(i,j) ;
		     }
		  }
	       }
	       ostr << ")" << endl ;
	    } 
	 } ;
   } ;

   //======================================================================
   // BpLLink Table

   //--------------------------------------------------
   // Index

   class VBpLLinkIndex : public WithDescriber {
	 //--------------------------------------------------
	 // member
      public:
	 SString from ;
	 SString to ;
	 
	 //--------------------------------------------------
	 // constructor
      public:
	 VBpLLinkIndex() : from(), to() {} ;
	 VBpLLinkIndex(const SString & f, const SString & t) {
	    from = f ; to = t ;
	 } ;
	 VBpLLinkIndex(const VBpLayer & f, const VBpLayer & t) {
	    from = f.name ; to = t.name ;
	 } ;
	 VBpLLinkIndex(const VBpLayer * f, const VBpLayer * t) {
	    from = f->name ; to = t->name ;
	 } ;
	 VBpLLinkIndex(const VBpLLink * lnk) {
	    from = lnk->from->name ;
	    to = lnk->to->name ;
	 } ;
	 VBpLLinkIndex(const VBpLLink & lnk) {
	    from = lnk.from->name ;
	    to = lnk.to->name ;
	 } ;
	 
	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr,
			       const Bool detailp = True) const {
	    ostr << "#VBpLLlinkIndex[" << from << "->" << to << "]" ;
	    if(detailp) {
	       ostr << endl ;
	    } 
	 } ;
   } ;
   
   inline Bool MatchIndex(const VBpLLinkIndex & x, const VBpLLinkIndex & y) {
      return x.from == y.from && x.to == y.to ;
   } ;

   //--------------------------------------------------
   // Table

   class VBpLLinkTable : public IndexTableT<VBpLLinkIndex, VBpLLink *> {
	 //--------------------------------------------------
	 // assign
      public:
	 Entry * assign(VBpLLink * lnk) {
	    return IndexTable::assign(VBpLLinkIndex(lnk),lnk) ;
	 } ;
	 Entry * assign(VBpLLink & lnk) { return assign(&lnk) ; } ;

	 //--------------------------------------------------
	 // find
      public:
	 VBpLLink * find(const VBpLayer * from, const VBpLayer * to) const {
	    return IndexTable::find(VBpLLinkIndex(from,to)) ;
	 } ;
	 VBpLLink * find(const VBpLayer & from, const VBpLayer & to) const {
	    return find(&from,&to) ;
	 } ;
	 VBpLLink * find(const SubString & from, const SubString & to) const {
	    return IndexTable::find(VBpLLinkIndex(from,to)) ;
	 } ;
   } ;

} ;

////////////////////////////////////////////////////////////////////////
#endif


