// -*- Mode:C++ -*-
//Header:
//File: VBpNet3L.h
//Title: VBpNet with 3 layers
//Author: NODA, Itsuki
//Date: 2000/01/03
//

//ModifyHistory:
// 2000/01/03: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_VBpNet3L_h_
#define _Nnn_VBpNet3L_h_
////////////////////////////////////////////////////////////////////////

#include "VBpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VBpNet3L

   class VBpNet3L : public VBpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 virtual char* L_TermHead() const { return "VBpNet3L" ; } ;
	 static char * const nameInput = "input" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameOutput = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 VBpLayer * lyrInput ;
	 VBpLayer * lyrHidden ;
	 VBpLayer * lyrOutput ;

	 VBpLLink * lnkInpHid ;
	 VBpLLink * lnkHidOut ;

	 //--------------------------------------------------
	 // constructor
      public:
	 VBpNet3L() { init() ; } ;
	 VBpNet3L(const SString & nm,
		  const UInt & isize,
		  const UInt & hsize,
		  const UInt & osize,
		  const Val & tempval = NNN_DFLT_TEMP,
		  const Val & epsval = NNN_DFLT_EPS,
		  const Val & alphaval = NNN_DFLT_ALPHA) { 
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(isize,hsize,osize) ;
	 } ;

	 //--------------------------------------------------
	 // assignLayer
      public:
	 void assignLayerLLink(const UInt & isize, 
			       const UInt & hsize, 
			       const UInt & osize) {
	    lyrInput = newLayer(nameInput,isize,VBpLayer::FT_linear) ;
	    lyrHidden = newLayer(nameHidden,hsize,VBpLayer::FT_sigmoid) ;
	    lyrOutput = newLayer(nameOutput,osize,VBpLayer::FT_sigmoid) ;
	    
	    lnkInpHid = newLLink(lyrInput,lyrHidden) ;
	    lnkHidOut = newLLink(lyrHidden,lyrOutput) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(const ValVector & inpvec) {
	    lyrInput->propFore(inpvec,temp) ;
	    lyrHidden->propFore(temp) ;
	    lyrOutput->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    lyrOutput->propBack(tchvec,temp) ;
	    lyrHidden->propBack(temp) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 void learn() {
	    lyrHidden->learn(eps,alpha) ;
	    lyrOutput->learn(eps,alpha) ;
	    lnkInpHid->learn(eps,alpha) ;
	    lnkHidOut->learn(eps,alpha) ;
	 } ;

	 //--------------------------------------------------
	 // run

      public:
	 void run(const ValVector & inpvec) { return propFore(inpvec) ; } ;
	 void runLearn(const ValVector & inpvec,
		       const ValVector & tchvec) {
	    propFore(inpvec) ;
	    propBack(tchvec) ;
	    learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return lyrOutput->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    lyrOutput->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    VBpNetBase::scan(sform) ;
	    lyrInput = layer(nameInput) ;
	    lyrHidden = layer(nameHidden) ;
	    lyrOutput = layer(nameOutput) ;
	    lnkInpHid = link(lyrInput,lyrHidden) ;
	    lnkHidOut = link(lyrHidden,lyrOutput) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


