// -*- Mode:C++ -*-
//Header:
//File: VBpNetBase.cc
//Author: NODA, Itsuki
//Date: 1999/12/31
//

//ModifyHistory:
// 1999/12/31: Start to create this file
//EndModifyHistory:

#include "VBpNetBase.h"

namespace Nnn {

   //======================================================================
   // allocation

   //----------------------------------------------------------------------
   // new layer

   //--------------------------------------------------
   // by StrForm

   VBpLayer * VBpNetBase::newLayer(StrForm & sform) {
      VBpLayer * layer = new VBpLayer ;
      layer->scan(sform) ;
      addLayer(layer) ;
      return layer ;
   } ;

   //--------------------------------------------------
   // by Name

   VBpLayer * VBpNetBase::newLayer(const SString & nm, UInt size,
				 VBpLayer::FuncType ftype) {
      VBpLayer * layer = new VBpLayer(nm,size,ftype) ;
      addLayer(layer) ;
      return layer ;
   } ;
      
   //--------------------------------------------------
   // by soruce layer

   VBpLayer * VBpNetBase::newLayer(const VBpLayer * src) {
      VBpLayer * layer = &(src->duplicate()) ;
      addLayer(layer) ;
      return layer ;
   } ;


   //----------------------------------------------------------------------
   // add layer

   void VBpNetBase::addLayer(VBpLayer * layer) {
      layerTable.assign(layer) ;
   } ;

   //----------------------------------------------------------------------
   // get layer

   VBpLayer * VBpNetBase::layer(const SString & nm) {
      return layerTable.find(nm) ;
   } ;

   //----------------------------------------------------------------------
   // new link

   //--------------------------------------------------
   // by StrForm

   VBpLLink * VBpNetBase::newLLink(StrForm & sform) {
      VBpLLink * link = new VBpLLink ;
      link->scan(sform,&layerTable,&linkTable) ;
      addLLink(link) ;
      return link ;
   } ;

   //--------------------------------------------------
   // by from/to layer

   VBpLLink * VBpNetBase::newLLink(VBpLayer * from, VBpLayer * to) {
      VBpLLink * link = new VBpLLink(from,to) ;
      addLLink(link) ;
      return link ;
   } ;

   //--------------------------------------------------
   // by source link

   VBpLLink * VBpNetBase::newLLink(const VBpLLink * src, const Bool addp) {
      VBpLLink * link = &(src->duplicate(&layerTable)) ;
      if(addp) {
	 // *** here, should assign from/to layer 
	 addLLink(link) ;
      }
      return link ;
   } ;

   //----------------------------------------------------------------------
   // add link

   void VBpNetBase::addLLink(VBpLLink * link) {
      linkTable.assign(link) ;
      link->registerFromTo() ;
   } ;
      
   //----------------------------------------------------------------------
   // get link

   VBpLLink * VBpNetBase::link(const SString & from, const SString & to) {
      return linkTable.find(from,to) ;
   } ;

   VBpLLink * VBpNetBase::link(VBpLayer * from, VBpLayer * to) {
      return linkTable.find(from,to) ;
   } ;


   //----------------------------------------------------------------------
   // randomize threshold and weight

   void VBpNetBase::randomize(const Val & minval, const Val & maxval) {
      for(VBpLayerTable::Iterator lyr = layerTable.begin() ;
	  lyr < layerTable.end() ;  lyr++) {
	 lyr.content()->randomizeThreshold(minval,maxval) ;
      }
      for(VBpLLinkTable::Iterator lnk = linkTable.begin() ;
	  lnk < linkTable.end() ; lnk++) {
	 lnk.content()->randomizeWeight(minval,maxval) ;
      } 
   } 

   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void VBpNetBase::save(ostream & ostr, 
			 SaveLevel slevel,
			 const Bool & newlinep,
			 const SubString & indent) {
      SubString indent2 = indent + "  " ;
      SubString indent3 = indent2 + "  " ;
      StrForm::writeBeginTerm(ostr,L_TermHead()) ;

      StrForm::writeAssocRole(ostr,L_Name) ; 
      StrForm::writeString(ostr,name) ;
      StrForm::writeArgSep(ostr) ;

      ostr << endl << indent2 ;

      StrForm::writeAssocRole(ostr,L_Temp) ; 
      ostr << NNN_VAL_FORMAT << temp ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,L_Eps) ; 
      ostr << NNN_VAL_FORM(20,16) << eps ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,L_Alpha) ; 
      ostr << NNN_VAL_FORM(10,6) << alpha ;
      StrForm::writeArgSep(ostr) ;

      ostr << NNN_VAL_FORM(10,6) << endl << indent2 ;
      
      StrForm::writeAssocRole(ostr,L_CntTrial) ; 
      ostr << cntTrial ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,L_CntLearn) ; 
      ostr << cntLearn ;
      StrForm::writeArgSep(ostr) ;

      ostr << endl << indent2 ;
      StrForm::writeAssocRole(ostr,L_Layer) ; {
	 StrForm::writeBeginTerm(ostr) ;
	 StrForm::writeBeginList(ostr) ;
	 for(VBpLayerTable::Iterator i = layerTable.begin() ; 
	     i < layerTable.end() ; i++) {
	    if(!layerTable.isBegin(i)) StrForm::writeArgSep(ostr) ;
	    ostr << endl << indent3 ;
	    i.content()->save(ostr,slevel,False,indent3) ;
	 }
	 ostr << endl << indent2 ;
	 StrForm::writeEndTerm(ostr) ;
      }
      StrForm::writeArgSep(ostr) ;

      ostr << endl << indent2 ;
      StrForm::writeAssocRole(ostr,L_LLink) ; {
	 StrForm::writeBeginTerm(ostr) ;
	 StrForm::writeBeginList(ostr) ;
	 for(VBpLLinkTable::Iterator i = linkTable.begin() ; 
	     i < linkTable.end() ; i++) {
	    if(!linkTable.isBegin(i)) StrForm::writeArgSep(ostr) ;
	    ostr << endl << indent3 ;
	    i.content()->save(ostr,slevel,False,indent3) ;
	 }
	 ostr << endl << indent2 ;
	 StrForm::writeEndTerm(ostr) ;
      }
      StrForm::writeArgSep(ostr) ;

      ostr << endl << indent ;
      StrForm::writeEndTerm(ostr) ;
      if(newlinep) ostr << endl ;
   }
   
   //----------------------------------------------------------------------
   // scan

   void VBpNetBase::scan(StrForm & sform) { 
      if(!sform.isTerm(L_TermHead())) {
	 ITK_ERR("Illegal VBpNetBase StrForm:" << sform) ;
	 throw 0 ;
      } 

      sform.assocCvtSString(L_Name,name) ;

      sform.assocCvtFlt(L_Temp,temp) ;
      sform.assocCvtFlt(L_Eps,eps) ;
      sform.assocCvtFlt(L_Alpha,alpha) ;

      sform.assocCvtUInt(L_CntTrial,cntTrial) ;
      sform.assocCvtUInt(L_CntLearn,cntLearn) ;

      StrForm * layerlist = sform.assoc(L_Layer) ;
      if(!isNull(layerlist)) {
	 for(Index i = 0 ; i < layerlist->lengthListArg() ; i++) {
	    newLayer(*(layerlist->nthListArg(i))) ;
	 } ;
      }

      StrForm * linklist = sform.assoc(L_LLink) ;
      if(!isNull(linklist)) {
	 for(Index i = 0 ; i < linklist->lengthListArg() ; i++) {
	    newLLink(*(linklist->nthListArg(i))) ;
	 } ;
      }
   } ;

   //----------------------------------------------------------------------
   // duplicate

   VBpNetBase & VBpNetBase::duplicate(const Bool sharedp) const {
      VBpNetBase * net = new VBpNetBase() ;
      net->copyFrom(*this,sharedp,True) ;
      return *net ;
   } ;  

   //----------------------------------------------------------------------
   // copyFrom

   VBpNetBase & VBpNetBase::copyFrom(const VBpNetBase & src,
				     const Bool sharedp,
				     const Bool allparamp) {
      VBpNetBase & tsrc = const_cast<VBpNetBase &>(src) ;
      if(allparamp) {
	 init(src.name, src.temp, src.eps, src.alpha) ;
	 cntTrial = src.cntTrial ;
	 cntLearn = src.cntLearn ;
      }

      for(VBpLayerTable::Iterator lsrc = tsrc.layerTable.begin() ;
	  lsrc < tsrc.layerTable.end() ; lsrc++) {
	 VBpLayer * layer = newLayer(lsrc.content()) ;
	 layer->copyFrom(*lsrc.content(),sharedp,allparamp) ;
      }
      
      for(VBpLLinkTable::Iterator lsrc = tsrc.linkTable.begin() ;
	  lsrc < tsrc.linkTable.end(); lsrc++) {
	 VBpLLink * link = newLLink(lsrc.content(),False) ;
	 link->copyFrom(*lsrc.content(),sharedp,allparamp, &layerTable) ;
	 addLLink(link) ;
      }

      return *this ;
   } ;

} ;
