// -*- Mode:C++ -*-
//Header:
//File: VBpUnitWrap.h
//Author: NODA, Itsuki
//Date: 1999/12/16
//

//ModifyHistory:
// 1999/12/16: Start to create this file
// 1999/12/26: reconfigulation
//EndModifyHistory:

#ifndef _Nnn_VBpUnitWrap_h_
#define _Nnn_VBpUnitWrap_h_
////////////////////////////////////////////////////////////////////////

#include "n3btype.h"
#include "n3utility.h"
#include "StrForm.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VBpUnitWrap

   class VBpUnitWrap : public WithDescriber {

	 //--------------------------------------------------
	 // StrForm labels
      public:
	 static char* const LblTermHead = "VBpUW" ; 
	 static char* const LblThr = "h" ;  

	 //--------------------------------------------------
	 // dummy value
      public:
	 static Val _dummy ;

	 //--------------------------------------------------
	 // member
      public:
	 Val & inp ;
	 Val & out ;
	 Val & thr ;
	 Val & err ; // = delta-in
	 Val & dlt ;
	 Val & dh ;

	 //--------------------------------------------------
	 // constructor
      public:
	 VBpUnitWrap(Val & i, Val & o, Val & h, 
		     Val & e, Val & d, Val & c) :
	    inp(i), out(o), thr(h), err(e), dlt(d), dh(c) { } ;

	 VBpUnitWrap(Val & i, Val & o, Val & h, Val & e, Val & d) :
	    inp(i), out(o), thr(h), err(e), dlt(d), dh(_dummy) { } ;

	 VBpUnitWrap(Val & i, Val & o, Val & h, Val & e) :
	    inp(i), out(o), thr(h), err(e), dlt(_dummy), dh(_dummy) { } ;

	 VBpUnitWrap(Val & i, Val & o, Val & h) :
	    inp(i), out(o), thr(h), err(_dummy), dlt(_dummy), dh(_dummy) { } ;

	 VBpUnitWrap(Val & i, Val & o) : inp(i), out(o),
	       thr(_dummy), err(_dummy), dlt(_dummy), dh(_dummy) { } ;

	 VBpUnitWrap() : inp(_dummy), out(_dummy), 
		  thr(_dummy), err(_dummy), dlt(_dummy), dh(_dummy) { } ;
	 
	 VBpUnitWrap(const VBpUnitWrap & s) :
	    inp(s.inp), out(s.out), thr(s.thr), 
	    err(s.err), dlt(s.dlt), dh(s.dh) { } ;

	 //--------------------------------------------------
	 // forward processing
      public:
	 void activateByLinear(const Val & temp) {
	    out = inp / temp ;
	 } ;
	 void activateBySigmoid(const Val & temp) { 
	    out = sigmoid(inp,temp) ; 
	 } ;
	 void activateByBigmoid(const Val & temp) { 
	    out = bigmoid(inp,temp) ; 
	 } ;

	 //--------------------------------------------------
	 // backward processing
      public:
	 void calcError(const Val & tch) {
	    err = tch - out ;
	 } ;
	 Flt sqerr() const { return sq(err) ; } ;

	 void calcDeltaByLinear(const Val & temp) {
	    dlt = err / temp ;
	 } ;
	 void calcDeltaBySigmoid(const Val & temp) {
	    dlt = err * d_sigmoid(inp,out,temp) ;
	 } ;
	 void calcDeltaByBigmoid(const Val & temp) {
	    dlt = err * d_bigmoid(inp,out,temp) ;
	 } ;

	 //--------------------------------------------------
	 // save 
      public:
	 virtual void save(ostream & ostr) const ;

	 //--------------------------------------------------
	 // load
      public:
	 virtual void scan(StrForm & sform) ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#VBpUnitWrap[" 
		 << "i=" << NNN_VAL_FORMAT << inp 
		 << ", h=" << NNN_VAL_FORMAT << thr 
		 << ", o=" << NNN_VAL_FORMAT << out 
		 << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "\t" << "inp(" << &inp << ")=" 
		    << NNN_VAL_FORMAT << inp << endl ;
	       ostr << "\t" << "out(" << &out << ")=" 
		    << NNN_VAL_FORMAT << out << endl ;
	       ostr << "\t" << "thr(" << &thr << ")=" 
		    << NNN_VAL_FORMAT << thr << endl ;
	       ostr << "\t" << "err(" << &err << ")=" 
		    << NNN_VAL_FORMAT << err << endl ;
	       ostr << "\t" << "dlt(" << &dlt << ")=" 
		    << NNN_VAL_FORMAT << dlt << endl ;
	       ostr << "\t" << "dh(" << &dh << ")=" 
		    << NNN_VAL_FORMAT << dh << endl ;
	    } 
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif
