// -*- Mode:C++ -*-
//Header:
//File: XpdVElman5Lb.h
//Title: Expanded VElman Net with 5 sigmoid layers with bypass
//Author: NODA, Itsuki
//Date: 2000/10/23
//

//ModifyHistory:
// 2000/01/05: Start to create this file (originally in VElman5L.h)
// 2000/10/23: Copy from VElman5L.h
//EndModifyHistory:

#ifndef _Nnn_XpdVElman5Lb_h_
#define _Nnn_XpdVElman5Lb_h_
////////////////////////////////////////////////////////////////////////

#include "VElman5Lb.h"
#include "TrainSet4XpdVElman5Lb.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VElman 5 Layer with bypass

   class XpdVElman5Lb : public VBpNetBase {

      public:
	 //--------------------------------------------------
	 // member

	 VElman5Lb * mother ;
	 UInt n ;		/* the number of expansion */

	 VBpLayer ** _context ;
	 VBpLayer ** _input ;
	 VBpLayer ** _inter1 ;
	 VBpLayer ** _inter2 ;
	 VBpLayer ** _output ;

	 VBpLLink ** _con_int1 ;
	 VBpLLink ** _inp_int1 ;
	 VBpLLink ** _con_hid ;
	 VBpLLink ** _inp_hid ;
	 VBpLLink ** _int1_hid ;
	 VBpLLink ** _hid_int2 ;
	 VBpLLink ** _hid_out ;
	 VBpLLink ** _int2_out ;

	 //--------------------------------------------------
	 // constructor
      public:
	 XpdVElman5Lb() { init() ; } ;
	 XpdVElman5Lb(VElman5Lb * elman, UInt _n) {
	    init() ;
	    expand(elman,_n) ;
	 } ;
	 
	 //--------------------------------------------------
	 // expand
      public:
	 void expand(VElman5Lb * elman, UInt _n) {
	    mother = elman ;
	    n = _n ;

	    init(elman->name,elman->temp,elman->eps,elman->alpha) ;
	    
	    _context = allocMultiLayer(n+1, elman->hidden) ;
	    _input =   allocMultiLayer(n,   elman->input);
	    _inter1 =  allocMultiLayer(n,   elman->inter1) ;
	    _inter2 =  allocMultiLayer(n,   elman->inter2) ;
	    _output =  allocMultiLayer(n,   elman->output) ;

	    _con_int1 = allocMultiLLink(n, context(), inter1(),
					elman->con_int1) ;
	    _inp_int1 = allocMultiLLink(n, input(),   inter1(),
					elman->inp_int1) ;
	    _con_hid  = allocMultiLLink(n, context(), hidden(),
					elman->con_hid);
	    _inp_hid  = allocMultiLLink(n, input(),   hidden(),
					elman->inp_hid);
	    _int1_hid = allocMultiLLink(n, inter1(),  hidden(),
					elman->int1_hid);
	    _hid_int2 = allocMultiLLink(n, hidden(),  inter2(),
					elman->hid_int2) ;
	    _hid_out  = allocMultiLLink(n, hidden(),  output(),
					elman->hid_out) ;
	    _int2_out = allocMultiLLink(n, inter2(),  output(),
					elman->int2_out) ;
	 } ;

	 //--------------------------------------------------
	 // alloc multi Layer
      public:
	 VBpLayer ** allocMultiLayer(UInt m, VBpLayer * original) {
	    VBpLayer ** r = new VBpLayer* [m] ;
	    Buffer buf(1024) ;

	    for(UInt i = 0 ; i < m ; i++) {
	       r[i] = &(original->duplicate()) ;

	       buf.clear() ;
	       buf.printf("%s_%ld\000",(char *)(r[i]->name),i) ;
	       buf.terminate() ;
	       r[i]->name = buf.body().dup() ;

	       addLayer(r[i]) ;
	    }
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // alloc multi LLink
      public:
	 VBpLLink ** allocMultiLLink(UInt m, 
				     VBpLayer ** from, 
				     VBpLayer ** to,
				     VBpLLink * original) {
	    VBpLLink ** r = new VBpLLink* [m] ;
	    
	    for(UInt i = 0 ; i < m ; i++) {
	       r[i] = new VBpLLink(from[i],to[i]) ;
	       r[i]->copyMiscParamFrom(*original) ;
	       r[i]->copyWeightFrom(*original) ;
	       addLLink(r[i]) ;
	    }
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // access
      public:

	 ITK_DEF_ACCESS3(VBpLayer **, context(), { return _context;}) ;
	 ITK_DEF_ACCESS3(VBpLayer **, input(), { return _input ; }) ;
	 ITK_DEF_ACCESS3(VBpLayer **, inter1(), { return _inter1 ; }) ;
	 ITK_DEF_ACCESS3(VBpLayer **, hidden(), { return &(_context[1]) ; }) ;
	 ITK_DEF_ACCESS3(VBpLayer **, inter2(), { return _inter2 ; }) ;
	 ITK_DEF_ACCESS3(VBpLayer **, output(), { return _output ; }) ;

	 ITK_DEF_ACCESS(VBpLayer * context(Int i), { return context()[i];}) ;
	 ITK_DEF_ACCESS(VBpLayer * input(Int i), { return input()[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLayer * inter1(Int i), { return inter1()[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLayer * hidden(Int i), { return hidden()[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLayer * inter2(Int i), { return inter2()[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLayer * output(Int i), { return output()[i] ; }) ;

	 ITK_DEF_ACCESS(VBpLLink * con_int1(Int i), { 
	    return _con_int1[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * inp_int1(Int i), { 
	    return _inp_int1[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * con_hid(Int i), { 
	    return _con_hid[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * inp_hid(Int i), { 
	    return _inp_hid[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * int1_hid(Int i), { 
	    return _int1_hid[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * hid_int2(Int i), { 
	    return _hid_int2[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * hid_out(Int i), { 
	    return _hid_out[i] ; }) ;
	 ITK_DEF_ACCESS(VBpLLink * int2_out(Int i), { 
	    return _int2_out[i] ; }) ;
	 
	 //--------------------------------------------------
	 // renewContext
      public:
	 void renewContext(VBpLayer * prehidden) {
	    context(0)->setOut(prehidden->out()) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore1(Int t) {
	    /* input should be set at setInp1 */
	    // input(t)->propFore(*inpvec[t],temp) ; 

	    inter1(t)->propFore(temp) ;
	    hidden(t)->propFore(temp) ;
	    inter2(t)->propFore(temp) ;
	    output(t)->propFore(temp) ;
	 } ;

	 void propFore(TrainSet4XpdVElman5Lb * tset, Bool firstp) {
	    for(Int i = 0 ; i < static_cast<Int>(n) ; i++) {
	       tset->setInp1(this,i,input(i),
			     (i == 0 ? ITK_NULLPTR : output(i-1)),
			     firstp) ;
	       propFore1(i) ;
	    } 
	 } ;

	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack1(Int t, 
			Flt alpha,
			Bool backtoinputp = True) {
	    /* output should be operated in setTch1 */
	    // output(i)->propBack(*tchvec[i],temp) ; 

	    inter2(t)->propBack(temp) ;
	    hidden(t)->propBack(temp) ;
	    inter1(t)->propBack(temp) ;
	    if(backtoinputp) {
	       input(t)->propBack(temp) ;
	       //Flt alpha = 0.1 ;	/* adaptation step */
	       input(t)->adaptRInpByDelta(alpha) ;   
	    }
	 } ;

	 void propBack(TrainSet4XpdVElman5Lb * tset,
		       Bool firstp,
		       Flt alpha,
		       Bool backtoinputp = True){
	    for(Int i = n-1 ; i >= 0 ; i--) {

	       tset->setTch1(this,i,output(i),
			     ((i == static_cast<Int>(n)-1) ? 
			      ITK_NULLPTR : input(i+1)),
			     firstp) ;

	       propBack1(i,alpha,backtoinputp) ;
	    }
	 } ;
	       
	 //--------------------------------------------------
	 // run
      public:
	 void run(TrainSet4XpdVElman5Lb * tset) {
	    propFore(tset,True) ;
	    cntTrial++ ;
	 } ;

	 //--------------------------------------------------
	 // runAdapt
      public:
	 void runAdapt1(TrainSet4XpdVElman5Lb * tset, 
			const Bool firstp,
			const Val & alpha) {
	    propFore(tset,firstp) ;
	    propBack(tset,firstp,alpha) ;
	 } ;

	 void runAdapt(TrainSet4XpdVElman5Lb * tset, 
		       const Int m, 
		       const Val & alpha) {

	    for(Int i = 0 ; i < m ; i++) {
	       runAdapt1(tset,(i==0),alpha) ;
	    }
	 } ;

   } ;
} ;


////////////////////////////////////////////////////////////////////////
#endif
