// -*- Mode:C++ -*-
//Header:
//File: utility.h
//Author: NODA, Itsuki
//Date: 1999/12/16
//

//ModifyHistory:
// 1999/12/16: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_utility_h_
#define _Nnn_utility_h_
////////////////////////////////////////////////////////////////////////

#include "n3btype.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // square

   /* This is defined in itk */
   //inline Val sq(const Val& v) { return v * v ; } ;

   //======================================================================
   // sigmoid functions

   inline Val sigmoid(const Val& x, 
		      const Val& temp = 1.0) {
      return 1.0 / (1.0 + exp(-x/temp)) ;
   } ;
   inline Val d_sigmoid(const Val& x, const Val& y,
		       const Val& temp = 1.0) {
      return y * (1.0 - y) / temp ;
   } ;

   //======================================================================
   // bipolar sigmoid functions

   inline Val bigmoid(const Val& x, 
		      const Val& temp = 1.0) {
      return 2.0 * sigmoid(x,temp) - 1.0 ;
   } ;
   inline Val d_bigmoid(const Val& x, const Val& y,
			const Val& temp = 1.0) {
      return 2.0 * d_sigmoid(x, (y + 1.0)/2.0, temp) ;
   } ;

   //======================================================================
   // liner

   inline Val liner(const Val& x, 
		    const Val& temp = 1.0) {
      return temp * x ;
   } ;
   inline Val d_liner(const Val& x, const Val& y,
		     const Val& temp = 1.0) {
      return temp ;
   } ;

   //======================================================================
   // exceptions

   class Exception {
      public:
	 Int type ;
	 char * name ;
	 void * aux ;
	 char * file ;
	 UInt line ;

      public:
	 Exception(Int t, const char * const nm, void * a, char * f, UInt l) {
	    type = t ;
	    name = strdup(nm) ;
	    aux = a ;
	    file = f ;
	    line = l ;
	 } ;

	 Exception(const Exception & ex) {
	    type = ex.type ;
	    name = ex.name ;
	    aux = ex.aux ;
	    file = ex.file ;
	    line = ex.line ;
	 } ;

      public:
	 friend ostream & operator<< (ostream & ostr, 
				      const Exception & ex) {
	    ostr << "#Exception"
		 << "[" << "type=" << ex.type 
		 << "," << "name=" << ex.name 
		 << "," << "aux=" << ex.aux 
		 << "," << "at=" << ex.file << ":" << ex.line
		 << "]" ;
	    return ostr ;
	 } ;
   } ;

#define NNN_XPT(type,name,aux) {\
      Nnn::Exception * ex = new Nnn::Exception(type,name,aux,__FILE__,__LINE__) ; \
      throw(*ex) ; \
   }

   const Flt TooBigChange = 1.0e5 ;

} ;
////////////////////////////////////////////////////////////////////////
#endif
