## -*- Mode: Ruby -*-

require "itkUtility" ;
require "StrForm.rb" ;
require 'thread' ;

class MutexCV < ConditionVariable
  
  def initialize()
    super ;
    @mutex = Mutex.new() ;
  end

  def lock()
    @mutex.lock() ;
  end

  def unlock()
    @mutex.unlock() ;
  end

  def wait()
    super(@mutex) ;
  end
end

module Nnn

  ##======================================================================
  ## class LogScanner

  class LogScanner

    ##------------------------------------------------------------
    ## local class TrialEntry

    class TrialEntry 
      attr :trialCount	,true ;
      attr :learnCount	,true ;
      attr :err		,true ;
      attr :vv		,true ;

      def initialize(tc = 0, lc = 0, e = 0, vf = 0)
	@trialCount = tc ;
	@learnCount = lc ;
	@err = e ;
	#@vform = Array.new() ;
	@vv = Array.new() ;
	vf.each_index do |i|
	  @vv[i] = Array.new() ;
#	  @vform[i] = vf[i] ;
	  for j in 0 .. (vf[i].hash['_'].to_i - 1) do
	    @vv[i][j] = vf[i].array[j].to_f ;
	  end
	end
      end
    end
      
    ##------------------------------------------------------------
    ## initialize

    attr :file		,true ;
    attr :stream	,true ;
    attr :scanner	,true ;

    attr :trialN	,true ;
    attr :trialFormat	,true ;
    attr :trialCount	,true ;
    attr :learnCount	,true ;
    attr :err		,true ;
    attr :vform		,true ;
#    attr :trial		,true ;
    attr :trialPos	,true ;
    attr :trialSeekN	,true ;
    attr :mutexcv	,true ;
    attr :readThread	,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize(file)
      init(file) ;
    end

    def init(file)
      @file = file ;
      @stream = open(@file) ;
      @scanner = StrForm::StreamScanner.new(@stream) ;
      @vform = Array.new() ;
      @trialCount = 0 ;
      @learnCount = 0 ;
      @err = 0.0 ;
#      @trial = [] ;
      @trialPos = [] ;
      @trialSeekN = -1 ;
      @mutexcv = MutexCV.new() ;
      @readThread = nil ;
    end

    ##------------------------------------------------------------
    ## seek to next Trial entry

    def seekToNextTrial()
      while(TRUE) do
	form = StrForm::scan(@scanner) ;
	break if(form.type == 'term' && form.head == 'Trial') ;

	if(form.type == 'term' && form.head == 'SqEr')
	  @learnCount = form.hash['count'].to_i ;
	  @err = form.hash['err'].to_f ; 
	end
      end
      @trialN = form.hash['n'].to_i ;
      @trialFormat = form.hash['format'].to_s ;
    end

    ##------------------------------------------------------------
    ## read Trial entry

    def readTrialLine()
      markTrialLinePos() ;
      _readTrialLine() ;
#      @trial[@trialCount] = 
#	TrialEntry.new(@trialCount,@learnCount,@err,@vform) ;
      @trialCount += 1 ;
    end

    def _readTrialLine()
      StrForm::scan(@scanner) ; ## skip '~'
      @vform[0] = StrForm::scan(@scanner) ; 
      StrForm::scan(@scanner) ; ## skip '/'
      @vform[1] = StrForm::scan(@scanner) ; 
      StrForm::scan(@scanner) ; ## skip '/'
      @vform[2] = StrForm::scan(@scanner) ; 
    end

    ##------------------------------------------------------------
    ## mark Trial Entry

    def markTrialLinePos()
      @trialPos[@trialCount] = @scanner.pos() ;
    end

    ##------------------------------------------------------------
    ## mark Trial Entry

    def rereadTrialLine(tcount)
      cpos = @scanner.pos() ;
      @scanner.seek(@trialPos[tcount]) ;
      _readTrialLine() ;
      @scanner.seek(cpos) ;
    end

    ##------------------------------------------------------------
    ## seek to next Trial entry

    def getTrial0(n)
      until(@trialCount > n)
	seekToNextTrial() ;
	for i in 1..@trialN do
	  readTrialLine() ;
	end
      end
      return @trial[n] ;
    end

    def getTrial1(n)
      spawnReadThread() ;
      @mutexcv.lock() ;
      if(@trialCount <= n)
	@trialSeekN = n ;
	@mutexcv.broadcast() ;
	@mutexcv.wait() ;
      end
      trial = @trial[n] ;
      @mutexcv.unlock() ;
      return trial ;
    end

    def getTrial(n)
      spawnReadThread() ;
      @mutexcv.lock() ;
      while(@trialCount <= n)
	@trialSeekN = n ;
	@mutexcv.broadcast() ;
	@mutexcv.wait() ;
      end
      rereadTrialLine(n) ;
      trial = TrialEntry.new(@trialCount,@learnCount,@err,@vform) ;
      @mutexcv.unlock() ;
      return trial ;
    end

    ##------------------------------------------------------------
    ## spawn readTrial

    def spawnReadThread()
      if(@readThread == nil) 
	@mutexcv.lock() ;
	@readThread = Thread.new() {
	  loopReadTrial() ;
	}
	@mutexcv.wait() ;
	@mutexcv.unlock() ;
      else
      end
    end

    ##------------------------------------------------------------
    ## background reader

    def loopReadTrial()
      @mutexcv.lock() ;
      @mutexcv.broadcast() ;
      while (TRUE)
	while(@trialCount > @trialSeekN)
	  @mutexcv.wait() ;
	end
	
	@mutexcv.unlock() ;
	until(@trialCount > @trialSeekN)
	  seekToNextTrial() ;
	  for i in 1..@trialN do
	    @mutexcv.lock() ;
	    readTrialLine() ;
	    @mutexcv.broadcast() ;
	    @mutexcv.unlock() ;
	  end
	end
	@mutexcv.lock() ;
      end
      @mutexcv.unlock() ;
    end

  end

end

