// -*- Mode: c++ -*-
//Header:
//File: selfLearn.cc
//Author: NODA, Itsuki
//Date: 2001/04/01
//EndHeader:

//======================================================================
// Headers

#include "VElman5Lb.h"
#include "TrainSetBase.h"
#include "itk/Geo2D.h"

using namespace Itk ;
using namespace Nnn ;

typedef Geo2D::Vec XYVec ;

//======================================================================
// options

#define NetFileBase "selfLearn"

char * InNetFile = NetFileBase ".net" ;
char * OutNetFile = NetFileBase ".net" ;
char * TmpDumpFile = NetFileBase ".dump.net" ;

char * LogFile = NetFileBase ".log" ;
Bool LogP = True ;
UInt LogN = 10000 ;
Bool NewLogP = False ;

UInt CheckN = 0 ;

UInt LearnN = 100000 ;
UInt TestN = 50 ;

UInt LoopN = 1 ;

char * DataFile = "a.sdt" ;

void getopt(int & argc, char ** & argv) {
   OptArg opt[] = {
      {"infile",    OptArg::cstringVal,	&InNetFile, 0,	"netfile_for_input"},
      {"outfile",   OptArg::cstringVal,	&OutNetFile,0, 	"netfile_for_output"},
      {"data",	    OptArg::cstringVal,	&DataFile,0, 	"datafile"},
      {"learnN",    OptArg::IntVal,	&LearnN,     0,	"learn_N"},
      {"testN",     OptArg::IntVal,	&TestN,      0,	"test_N"},
      {"loopN",	    OptArg::IntVal,	&LoopN,      0,	"loop_N"},
      {"logfile",   OptArg::cstringVal,	&InNetFile, 0,	"logfile"},
      {"log",       OptArg::OnOff,	&LogP, 0,	"[on/off]"},
      {"logN",      OptArg::IntVal,	&LogN, 0,	"log_N"},
      {"newlog",    OptArg::OnOff,	&NewLogP, 0,	"[on/off]"},
      {"checkN",    OptArg::IntVal,	&CheckN, 0,	"check_N ; check internal layer's output"},
      {ITK_NULLPTR}} ;

   myGetOpt(opt,argc,argv) ;
}
      
//======================================================================
// training set

//--------------------------------------------------
class TsTurnDashKick : public TrainSetBase {
      
   public:

      //--------------------------------------------------
      // Index Util

      enum StateIndex {
	 ix_PPosX, ix_PPosY,
	 ix_BDirX, ix_BDirY,
	 ix_HDirX, ix_HDirY,
	 ix_PVelX, ix_PVelY,
	 ix_BPosX, ix_BPosY,
	 ix_BVelX, ix_BVelY,
	 ix_SI_Max
      } ;

      enum CommandIndex {
	 ix_TurnX, ix_TurnY,
	 ix_DashX, ix_DashY,
	 ix_KickX, ix_KickY,
	 ix_CI_Max
      } ;

      static int ixCoach(StateIndex ix) { return ix ; } ;
      static int ixPlayer(StateIndex ix) { return ix + ix_SI_Max; } ;
      static int ixCommand(CommandIndex ix) { return ix + 2 * ix_SI_Max ; } ;
      static int ixMax() { return 2 * ix_SI_Max + ix_CI_Max ; } ;

      //--------------------------------------------------
      // World State

      static const Val DirRadius = 10.0 ;

      class WorldState {
	    //----------------------------------------
	    // FieldState
	 public:
	    class Field {
		  //------------------------------
		  // member
	       public:
		  Int mask ;
		  XYVec ppos ;
		  XYVec pvel ;
		  Val bdir ;
		  Val hdir ;
		  XYVec bpos ;
		  XYVec bvel ;

		  //------------------------------
		  // access
	       public:
		  ITK_DEF_ACCESS(Val & pposX(), { return ppos.x ; }) ;
		  ITK_DEF_ACCESS(Val & pposY(), { return ppos.y ; }) ;
		  ITK_DEF_ACCESS(Val & pvelX(), { return pvel.x ; }) ;
		  ITK_DEF_ACCESS(Val & pvelY(), { return pvel.y ; }) ;

		  Val bdirX() const { return DirRadius * Cos(bdir) ; } ;
		  Val bdirY() const { return DirRadius * Sin(bdir) ; } ;

		  Val hdirX() const { return DirRadius * Cos(hdir) ; } ;
		  Val hdirY() const { return DirRadius * Sin(hdir) ; } ;
	       
		  ITK_DEF_ACCESS(Val & bposX(), { return bpos.x ; }) ;
		  ITK_DEF_ACCESS(Val & bposY(), { return bpos.y ; }) ;
		  ITK_DEF_ACCESS(Val & bvelX(), { return bvel.x ; }) ;
		  ITK_DEF_ACCESS(Val & bvelY(), { return bvel.y ; }) ;

		  //------------------------------
		  // setToVector
	       public:
		  void setToVector(ValVector & v, const Int & offset = 0) {
		     v[offset + ix_PPosX] = pposX() ;
		     v[offset + ix_PPosY] = pposY() ;
		     v[offset + ix_BDirX] = bdirX() ;
		     v[offset + ix_BDirY] = bdirY() ;
		     v[offset + ix_HDirX] = hdirX() ;
		     v[offset + ix_HDirY] = hdirY() ;
		     v[offset + ix_PVelX] = pvelX() ;
		     v[offset + ix_PVelY] = pvelY() ;
		     v[offset + ix_BPosX] = bposX() ;
		     v[offset + ix_BPosY] = bposY() ;
		     v[offset + ix_BVelX] = bvelX() ;
		     v[offset + ix_BVelY] = bvelY() ;

		     return ;
		  } ;

		  //------------------------------
		  // scan
	       public:
		  Bool scan(SimpleSexp * s) {
		     SimpleSexp * d = s->rest() ; // skip label

		     mask = 
			d->assocValue("m")->scanAsInt() ;

		     ppos.x = 
			d->assocValue("pp")->nth(0)->scanAsFlt() ;
		     ppos.y = 
			d->assocValue("pp")->nth(1)->scanAsFlt() ;

		     bdir = d->assocValue("bd")->scanAsFlt() ;
		     hdir = d->assocValue("hd")->scanAsFlt() ;

		     pvel.x = 
			d->assocValue("pv")->nth(0)->scanAsFlt() ;
		     pvel.y = 
			d->assocValue("pv")->nth(1)->scanAsFlt() ;

		     bpos.x = 
			d->assocValue("bp")->nth(0)->scanAsFlt() ;
		     bpos.y = 
			d->assocValue("bp")->nth(1)->scanAsFlt() ;

		     bvel.x = 
			d->assocValue("bv")->nth(0)->scanAsFlt() ;
		     bvel.y = 
			d->assocValue("bv")->nth(1)->scanAsFlt() ;

		     return True ;
		  } ;
		     
	    } ;

	    //----------------------------------------
	    // Command
	 public:
	    class Command {
		  //------------------------------
		  // member
	       public:
		  Int mask ;
		  Val tPow ;	/* turn power */
		  Val dPow ;  /* dash power */
		  Val kPow ;  /* kick power */
		  Val kDir ;  /* kick dir */
		  Val hPow ;  /* head power */

		  //------------------------------
		  // access
	       public:
		  Val turnX(const Field & fs) const { 
		     return -tPow * Sin(fs.bdir) ;  } ;
		  Val turnY(const Field & fs) const { 
		     return  tPow * Cos(fs.bdir) ;  } ;

		  Val dashX(const Field & fs) const { 
		     return dPow * Cos(fs.bdir) ;  } ;
		  Val dashY(const Field & fs) const { 
		     return dPow * Sin(fs.bdir) ;  } ;
		  
		  Val kickX(const Field & fs) const { 
		     return kPow * Cos(fs.bdir + kDir) ;  } ;
		  Val kickY(const Field & fs) const { 
		     return kPow * Sin(fs.bdir + kDir) ;  } ;

		  //------------------------------
		  // setToVector
	       public:
		  void setToVector(ValVector & v, 
				   const Field & fs,
				   const Int & offset = 0) {
		     v[offset + ix_TurnX] = turnX(fs) ;
		     v[offset + ix_TurnY] = turnY(fs) ;
		     v[offset + ix_DashX] = dashX(fs) ;
		     v[offset + ix_DashY] = dashY(fs) ;
		     v[offset + ix_KickX] = kickX(fs) ;
		     v[offset + ix_KickY] = kickY(fs) ;
		     
		     return ;
		  } ;

		  //------------------------------
		  // scan
	       public:
		  Bool scan(SimpleSexp * s) {
		     SimpleSexp * d = s->rest() ; // skip label
		     
		     mask = d->assocValue("m")->scanAsInt();
		     tPow = d->assocValue("tp")->scanAsFlt() ;
		     dPow = d->assocValue("dp")->scanAsFlt() ;
		     kPow = d->assocValue("kp")->scanAsFlt() ;
		     kDir = d->assocValue("kd")->scanAsFlt() ;
		     hPow = d->assocValue("hp")->scanAsFlt() ;

		     return True ;
		  } ;
	    } ;

	    //========================================
	    // World State: body
	    
	    //----------------------------------------
	    // member
	 public:
	    Int time ;
	    Field coachView ;
	    Field playerView ;
	    Command command ;

	    //----------------------------------------
	    // set to Vector
	 public:
	    void setToVector(ValVector & v, const Int & offset = 0) {
	       coachView.setToVector(v,offset + 0) ;
	       playerView.setToVector(v,offset + ix_SI_Max) ;
	       command.setToVector(v,coachView,offset + (2 * ix_SI_Max)) ;
	    } ;

	    //----------------------------------------
	    // scan
	 public:
	    Bool scan(SimpleSexp * s) {
	       SimpleSexp * d = s->rest() ; // skip label
	       
	       time = d->assocValue("t")->scanAsInt() ;
	       
	       SimpleSexp * ch = d->assocValue("ch") ;
	       coachView.scan(ch) ;

	       SimpleSexp * pl = d->assocValue("pl") ;
	       playerView.scan(pl) ;

	       SimpleSexp * cm = d->assocValue("cm") ;
	       command.scan(cm) ;

	       return True ;
	    } ;
      } ;

      //--------------------------------------------------
      // Streams

      SubString datfile ;
      istream * datstrm ;
      Buffer buffer ;

      //--------------------------------------------------
      // for parser

      SimpleSexp *data ;
      SimpleSexp::Heap sxHeap ;

      //--------------------------------------------------
      // World State

      WorldState world0 ;
      WorldState world1 ;

      //--------------------------------------------------
      // constructor ;

      TsTurnDashKick() {} ;
      TsTurnDashKick(const SubString & dfile) {
	 open(dfile) ;
      } ;

      //--------------------------------------------------
      // set size

      virtual UInt sizeInp() const { return ixMax() ; } ;
      virtual UInt sizeTch() const { return ixMax() * 2 ; } ;

      //--------------------------------------------------
      // open

      Bool open(const SubString & dfile) {
	 datfile = dfile ;
	 return open0(True) ;
      } ;

      Bool open0(Bool allocp = True) {
	 datstrm = new ifstream(datfile) ;
	 Bool r = !isNull(datstrm) ;
	 if(r) {
	    if(allocp) buffer.alloc(4096) ;
	 }
	 return r ; 
      } ;

      Bool reopen() {
	 delete datstrm ;
	 return open0(False) ;
      } ;

      //--------------------------------------------------
      // readScan

      Bool read1(Bool retryp = False) {
	 buffer.clear() ;
	 data = SimpleSexp::scan(datstrm,&buffer,&sxHeap) ;
	 if(data->isNullPtr()) {
	    if(retryp) {
	       ITK_WRN("No Data!!") ;
	       return False ;
	    } else {
	       ITK_WRN("reopen data file") ;
	       reopen() ;
	       return read1(True) ;
	    }
	 } else {
	    return True ;
	 }
      } ;

      Bool scan1() {
	 world1.scan(data) ;

	 return True ;
      } ;

      //--------------------------------------------------
      // shift

      void shift() {
	 world0 = world1 ;
      } ;

      //--------------------------------------------------
      // cycle

      void cycle() {
	 shift() ;
	 read1() ;
	 scan1() ;
      } ;

      virtual void get(ValVector & inp, ValVector & tch) {
	 cycle() ;

	 world0.setToVector(inp,0) ;
	 
	 world0.setToVector(tch,0) ;
	 world1.setToVector(tch,ixMax()) ;
      } ;
} ;

//======================================================================
// logging util

#define Logging(logp,logstr,coutp,form) {\
   if(logp) {\
      logstr << form << endl ;\
   }\
   if(coutp) {\
      cout << form << endl ;\
   }\
}


//======================================================================
// adaptation
#if 0
void adapt(VElman5Lb & net, const TsTurnDash & origtset) {
   VBpNetBase & net2 = net.duplicate() ;
   net.copyTo(net2) ;

   TsTurnDash tset = origtset ;
   
   ValVector inp0(tset.sizeInp()) ;
   ValVector inp(tset.sizeInp()) ;
   ValVector tch(tset.sizeTch()) ;

   tset.get(inp,tch) ;

   inp0.setAll(0.0) ;

   inp0 = inp ;
   inp0[4] = inp0[5] = inp0[6] = inp0[7] = 0.0 ;

   for(UInt i = 0 ; i < 10 ; i++) {
      net.propFore(inp0,(i == 0)) ;
      net.propBack(tch,True) ;
   
      cout << "~" << i << "\t"
	   << inp0 << "/" << net.input->dlt() << "/" << inp << endl ;

      inp0.inc(net.input->dlt()) ;
   }

   net.copyFrom(net2) ;
   delete & net2 ;
}

void adapt2(VElman5Lb & net, const TsTurnDash & origtset) {
   VBpNetBase & net2 = net.duplicate() ;
   net.copyTo(net2) ;

   TsTurnDash tset = origtset ;
   
   ValVector inp0(tset.sizeInp()) ;
   ValVector inp(tset.sizeInp()) ;
   ValVector tch(tset.sizeTch()) ;
   ValVector inpmask(tset.sizeInp()) ;
   ValVector tchmask(tset.sizeTch()) ;

   tset.get(inp,tch) ;

   inpmask.set(1) ;
   tchmask.set(1) ;

   inp0 = inp ;
   inp0[4] = inp0[5] = inp0[6] = inp0[7] = 0.0 ;
   inpmask[4] = inpmask[5] = inpmask[6] = inpmask[7] = 0.0 ;

   net.runAdapt(inp0,inpmask,tch,tchmask,0.3,10) ;

   cout << "~" 
	<< inp0 << "/" << net.input->rinp() << "/" << inp << endl ;

   net.copyFrom(net2) ;
   delete & net2 ;
}
#endif // #if 0

//======================================================================
// run

TsTurnDashKick TSet ;

void run(VElman5Lb & net, Bool logp, ostream & logstr) {

   ValVector inp(TSet.sizeInp()) ;
   ValVector tch(TSet.sizeTch()) ;

   Val err = 0.0 ;

   for(UInt i = 1 ; i < 60000 ; i++) {
      TSet.get(inp,tch) ;
      ITK_DBG(tch) ;
   } 
   exit(0) ;

   for(UInt i = 1 ; i <= LearnN ; i++) {
      TSet.get(inp,tch) ;
      Val dsum = net.runLearn(inp,tch) ;
      if(dsum > 1.0e10) { NNN_XPT(0,"too big change",&net) ; }
      const Val r = 0.9 ;
      err *= r ;
      err += net.sumError() ;
      if(net.cntLearn % LogN == 0) {
	 Logging(logp,logstr,True,
		 "#SqEr["
		 << "count=" << setfill(' ') << setw(7) << net.cntLearn << ", "
		 << "err=" << setprecision(10) << sqrt(err) << "]") ;
	 net.save(OutNetFile) ;
      }
   } 
   
   ValVector out(TSet.sizeTch()) ;

   if(TestN > 0) {
      Logging(logp,logstr,True,
	      "#Trial["
	      << "n=" << TestN << ", "
	      << "format=" << "\"inp/out/tch\"" << "]") ;
   }

   for(UInt i = 0 ; i < TestN ; i++) {
      TSet.get(inp,tch) ;
      net.run(inp) ;
      net.getNetOutput(out) ;
      if(i % 1 == 0) {
	 Logging(logp,logstr,True,
		 "~" << inp << "/" << out << "/" << tch) ;
      }
   }

   if(CheckN > 0) {
      Logging(logp,logstr,True,
	      "#Trial["
	      << "n=" << TestN << ", "
	      << "format=" << "\"inter1/inter2\"" << "]") ;
   }

   VBpLayer * inter1 = net.findLayer("inter1") ;
   VBpLayer * inter2 = net.findLayer("inter2") ;

   for(UInt i = 0 ; i < CheckN ; i++) {
      TSet.get(inp,tch) ;
      net.run(inp) ;
      net.getNetOutput(out) ;
      if(i % 1 == 0) {
	 Logging(logp,logstr,True,
		 "~" << inter1->out() << "/" << inter2->out()) ;
      }
   }
   
   // adapt2(net,TSet) ;

}

//--------------------------------------------------

void load_run_save() { 

   TSet.open(DataFile) ;

   ofstream logstr ;
   if(LogP) {
      if(NewLogP) 
	 logstr.open(LogFile) ;
      else
	 logstr.open(LogFile,ios::app) ;
   }

   VElman5Lb net ;

   net.load(InNetFile) ;
   Logging(LogP,logstr,True,
	   "#Load["
	   << "file=\"" << InNetFile << "\", "
	   << "time=\"" << TimeVal::now().stamp() << "\"]") ;

   for(UInt i = 0 ; i < LoopN ; i++) {
      try {
	 run(net,LogP,logstr) ;
      } catch (Nnn::Exception & ex) {
	 net.save(TmpDumpFile,SL_AllState) ;
	 Logging(LogP,logstr,True,
		 "#Xpt"
		 << "[" << "time=" << TimeVal::now().stamp() 
		 << "," << "xpt=" << ex 
		 << "]") ;
	 exit(1) ;
      }
   }

   //net.save(OutNetFile,SL_AllState) ;
   net.save(OutNetFile) ;
   Logging(LogP,logstr,True,
	   "#Save["
	   << "file=\"" << OutNetFile << "\", "
	   << "time=\"" << TimeVal::now().stamp() << "\"]") ;

#if 0
   //test net.duplicate and copy operation
   {
      nop() ;
      net.VBpNetBase::save("net1.all.0",SL_AllState) ;
    
      VBpNetBase & net2 = net.duplicate() ;
      net2.VBpNetBase::save("net2.all.0",SL_AllState) ;

      net.copyTo(net2) ;
      net.VBpNetBase::save("net1.all.1",SL_AllState) ;
      net2.VBpNetBase::save("net2.all.1",SL_AllState) ;
   }
#endif


   if(LogP)
      logstr.close() ;
}


//======================================================================
// main routine

int main(int argc, char* argv[]) {

   getopt(argc,argv) ;

   load_run_save() ;


   return 0;
}





