## -*- Mode: Ruby -*-

require 'gtk' ;
require "itkUtility" ;
require "GtkBufferedDrawingArea.rb" ;

module NnnViewer

  ##======================================================================
  ## class UnitWidget

  class XYPlane2

    DFLT_sizeX = 500 ;
    DFLT_sizeY = 500 ;

    attr :parent	,true ;
    attr :frame		,true ;
    attr :vbox		,true ;
    attr :notebox	,true ;
    attr :image		,true ;

    attr :notes		,true ;
    attr :label		,true ;
    attr :sizeX		,true ;
    attr :sizeY		,true ;
    attr :offsetX	,true ;
    attr :offsetY	,true ;
    attr :scaleX	,true ;
    attr :scaleY	,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize(parent, sizeX=DFLT_sizeX, sizeY=DFLT_sizeY, name=NIL)
      @parent = parent ;
      @sizeX = sizeX ;
      @sizeY = sizeY ;
      @offsetX = @sizeX / 2 ;
      @offsetY = @sizeY / 2 ;
      @scaleX =  4 ;
      @scaleY = -4 ;  ## for natural XY cordination
      
      @frame = Gtk::Frame::new(name) ;
      @parent.add(@frame) ;
      
      @vbox = Gtk::VBox::new() ;
      @frame.add(@vbox) ;
      
      @notebox = Gtk::HBox::new() ;
      @vbox.add(@notebox) ;

      @image = Gtk::BufferedDrawingAreaWithBack::new(sizeX,sizeY) ;
      @vbox.add(@image) ;

      @label = Hash.new() ;

      @colors = [
	"blue",
	"red",
	"DarkBlue",
	"DarkRed",
	[0x0000, 0x0000, 0xFFFF],
	[0xFFFF, 0x0000, 0x0000],
	[0x0000, 0x7777, 0x0000],
	[0x0000, 0xFFFF, 0xFFFF],
	[0xFFFF, 0x0000, 0xFFFF],
	[0xFFFF, 0xFFFF, 0x0000]] ;

      @isSetuped = FALSE ;

    end

    ##------------------------------------------------------------
    ## set name

    def setName(name)
      @frame.set_label(name) ;
    end

    ##------------------------------------------------------------
    ## access utility

    def buffer
      @image.buffer ;
    end

    def back
      @image.back ;
    end

    def saveBuffer()
      @image.saveToBack() ;
    end

    def restoreBuffer()
      @image.restoreFromBack() ;
    end

    ##------------------------------------------------------------
    ## assign new label

    class Label
      attr :id		,true ;
      attr :pname	,true ;
      attr :shape	,true ;		## should be rectangle or circle
      attr :gc		,true ;
      attr :x		,true ;
      attr :y		,true ;
      attr :note	,true ;

      def initialize(id,x,y,gc,pname,parent,shape = 'circle_fill')
	@id = id ;
	@x = x ;
	@y = y ;
	@gc = gc ;
	@pname = pname ;
	@shape = shape ;

	alignment = Gtk::Alignment::new(0,0,0,0) ;
	parent.add(alignment) ;

	hbox = Gtk::HBox::new() ;
	alignment.add(hbox) ;
	
	color = Gtk::BufferedDrawingArea::new(10,10) ;
	hbox.add(color) ;

	@note = Gtk::Label::new(pname) ;
	hbox.add(note) ;

	alignment.show_all() ;
	color.clear(gc) ;

      end
    end

    def newLabel(id,x,y,shape='circle_fill',pname=NIL)
      pname = id if (pname == NIL) ;
      w = label[id] = Label.new(id,x,y,
				##@image.newColoredGCbyRGB(*(@colors.shift())),
				@image.newColoredGC(@colors.shift()),
				pname,
				@notebox,
				shape) ;
      _moveLabel(w,x,y,TRUE) ;
    end

    ##------------------------------------------------------------
    ## setup

    def setup()
      @image.setup() ;

      drawBackground() ;

      saveBuffer() ;
      @isSetuped = TRUE ;
    end

    ##------------------------------------------------------------
    ## draw Background

    def drawBackground()
      @image.clear(@image.whiteGC) ;
      @image.draw_line(@image.fgc,ax(-50),ay(0),  ax(50),ay(0)) ;
      @image.draw_line(@image.fgc,ax(0),  ay(-50),ax(0), ay(50)) ;
    end

    ##------------------------------------------------------------
    ## set label name

    def setLabelName(id,name)
      label[id].pname = name ;
    end

    ##------------------------------------------------------------
    ## move label

    def moveLabel(id,x,y,firstp=FALSE)
      w = label[id] ;
      if(w == NIL) 
	return newLabel(id,x,y) ;
      else
	return _moveLabel(w,x,y,firstp) ;
      end
    end
    
    ##------------------------------------------------------------
    ## raw move label

    def _moveLabel(w,x,y,firstp=FALSE)
      w.x = x ;
      w.y = y ;
    end

    ##------------------------------------------------------------
    ## flush changes to window

    def flush()
      setup() if(!@isSetuped) ;

      restoreBuffer() ;
      
      label.each do |k,v|
	case v.shape
	when "circle_fill" 
	  @image.draw_circle_centered(v.gc,TRUE,ax(v.x),ay(v.y),10) ;
	when "circle" 
	  @image.draw_circle_centered(v.gc,nil,ax(v.x),ay(v.y),10) ;
	when "rectangle_fill"
	  @image.draw_rectangle_centered(v.gc,TRUE,ax(v.x),ay(v.y),10,10) ;
	when "rectangle"
	  @image.draw_rectangle_centered(v.gc,nil,ax(v.x),ay(v.y),10,10) ;
	else
	  @image.draw_point(v.gc,ax(v.x),ay(v.y)) ;
	end
      end
      
      @image.exposeAll() ;
    end

    ##------------------------------------------------------------
    ## calc actual xy

    def ax(x)
      (x * @scaleX) + @offsetX ;
    end

    def ay(y)
      (y * @scaleY) + @offsetY ;
    end
  end

end

		
