## -*- Mode: Ruby -*-

##Usage:
#	$scanner = StrForm::StreamScanner.new($stdin) ;
#	x = StrForm::scan($scanner) ;
#	p x ;
#	$stdout << x << "\n" ;
#

##======================================================================
## Module StrForm

module StrForm

  ##============================================================
  ## Token

  class Token
    attr :type	,true ;
    attr :str	,true ;

    def initialize(t,s)
      @type = t ;
      @str = s;
    end

    def to_s
      "#Token[%s:%s]" % [@type,@str]
    end

  end

  ##============================================================
  ## ScannerBase

  class ScannerBase
    attr :stack		,true ;
    attr :lineN		,true ;
    attr :charN		,true ;
    attr :charInLine	,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize()
      @stack = [] ;
      @lineN = 0 ;
      @charN = 0 ;
      @charInLine = 0 ;
    end

    ##------------------------------------------------------------
    ## basic operation

    ##----------------------------------------
    ## eof

    def eof()
      raise "undefined eof()" ;
    end

    ##----------------------------------------
    ## _getc

    def _getc()
      raise "undefined _getc()" ;
    end

    ##----------------------------------------
    ## _ungetc

    def _ungetc(c)
      raise "undefined _ungetc()" ;
    end

    ##----------------------------------------
    ## _pos

    def _pos() 
      raise "undefined _pos()" ;
    end

    ##----------------------------------------
    ## _seek

    def _seek(n,mode=File::SEEK_SET) 
      raise "undefined _seek()" ;
    end

    ##----------------------------------------
    ## getc

    def getc()
      if(eof()) 
	return nil ; 
      else
	c = _getc() ;
	@charN += 1 ;
	@lineN += 1 if (c == ?\n) ;
	return c ;
      end
    end

    ##----------------------------------------
    ## ungetc

    def ungetc(c)
      if(c != nil) then 
	_ungetc(c) ;
	@charN -= 1 ;
	@lineN -= 1 if (c == ?\n) ;
      end
    end

    ##----------------------------------------
    ## peek

    def peek()
      c = getc() ;
      ungetc(c) ;
      c ;
    end

    ##----------------------------------------
    ## _pos

    def pos() 
      return _pos() ;
    end

    ##----------------------------------------
    ## _seek

    def seek(n,mode=File::SEEK_SET) 
      return _seek(n,mode) ;
    end

    ##------------------------------------------------------------
    ## ungetToken

    def ungetToken(token)
      @stack.push(token) ;
    end

    ##------------------------------------------------------------
    ## getToken

    def getToken()

      token = @stack.pop() ;
      if(token != nil) 
	return token ;	
      end

      skipSpacesComments() ;
      return Token.new('EOF',nil) if(eof()) ;

      c = peek() ;
      if (c == ?\#)   # )
	return Token.new('TermPrefix',		"%c" % getc()) ;
      elsif (c == ?\[)
	return Token.new('TermArgBegin',	"%c" % getc()) ;
      elsif (c == ?\])
	return Token.new('TermArgEnd',		"%c" % getc()) ;
      elsif (c == ?\|)
	return Token.new('TermArgGuard',	"%c" % getc()) ;
      elsif (c == ?\,)
	return Token.new('TermArgSep',		"%c" % getc()) ;
      elsif (c == ?\=)
	return Token.new('TermArgEqual',	"%c" % getc()) ;
      elsif (c == ?\')  # '
	return Token.new('String',		getTokenString()) ;
      elsif (c == ?\")	# "
	return Token.new('String',		getTokenString()) ;
      elsif (c == ?\%)
	return Token.new('CommentBegin',	"%c" % getc()) ;
      else
	return Token.new('Atom',		getTokenAtom()) ;
      end
    end

    ##------------------------------------------------------------
    ## getTokenString

    def getTokenString()
      b = getc() ;
      str = "%c" % b ;
      endp = FALSE ;
      until(endp)
	return str if eof() ;
	c = getc() ;
	str = "%s%c" % [str,c] ;
	if(c == b) then endp = TRUE end
      end
      return str ;
    end

    ##------------------------------------------------------------
    ## getTokenAtom

    def getTokenAtom()
      str = "" ;
      while(TRUE)
	return str if (eof()) ;
	c = getc() ;
	cstr = "%c" % c ;
	if(/^[\s]$/ =~ cstr || ("\#\[\|\]\,\=\'\"\s\n\t\f".index(cstr) != nil))
	  ungetc(c) ;
	  return str ;
	end
	str = "%s%c" % [str,c] ;
      end
    end

    ##------------------------------------------------------------
    ## skips

    ##----------------------------------------
    ## skipSpaces

    def skipSpaces()
      c = 0 ;
      while(TRUE)
	return c if (eof()) ;
	cstr = "%c" % peek() ;
	if(!("\s\n\t\f".index(cstr))) then
	  return c;
	end
	getc() ;
	c += 1 ;
      end 
    end 

    ##----------------------------------------
    ## skipComments

    def skipComments()
      c = 0 ;
      if(peek() == ?\%)
	 until(peek() == ?\n || eof())
	   getc() ;
	   c += 1 ;
	 end
       end
       return c ;
    end

    ##----------------------------------------
    ## skipSpacesComments

    def skipSpacesComments()
      skipSpaces() ;
      if(skipComments > 0)
	skipSpacesComments()
      end
    end
    
  end

  ##============================================================
  ## StreamScanner

  class StreamScanner < ScannerBase

    attr :stream	,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize(str)
      super() ;
      @stream = str ;
    end

    ##------------------------------------------------------------
    ## basic operation

    ##----------------------------------------
    ## eof

    def eof()
      return @stream.eof? ;
    end

    ##----------------------------------------
    ## _getc

    def _getc()
      return @stream.getc() ;
    end

    ##----------------------------------------
    ## _ungetc

    def _ungetc(c)
      @stream.ungetc(c) ;
    end

    ##----------------------------------------
    ## _pos

    def _pos()
      return @stream.pos() ;
    end

    ##----------------------------------------
    ## _seek

    def _seek(n,mode=File::SEEK_SET)
      return @stream.seek(n,mode) ;
    end

  end

  ##============================================================
  ## Form

  class Form

    attr :head 		,true ;   # token when string or atom
    attr :type		,true ;   # atom, string, or term
    attr :hash	        ,true ;
    attr :array		,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize(type,head = NIL)
      @type = type ;
      @head = head ;
      if(type == 'term')
	@hash = Hash.new() ;
	@array = Array.new() ;
      end
    end

    ##------------------------------------------------------------
    ## to_s

    def to_s
      case @type
      when "atom", "string"
	return @head ;
      when "term"
	return term_to_s ;
      else
	return "#Form[????]" ;
      end
    end

    ##----------------------------------------
    ## term_to_s

    def term_to_s
      str = ("\#%s\[" % @head) ;

      firstp = TRUE ;
      @hash.each do |k,v|
	if(firstp)
	  firstp = FALSE ;
	else
	  str += "," ;
	end
	str += ("%s=%s" % [k,v.to_s]) ;
      end

      firstp = TRUE ;
      @array.each do |v|
	if(firstp)
	  firstp = FALSE ;
	  str += "|" ;
	else
	  str += "," ;
	end
	str += v.to_s ;
      end

      str += "\]" ;

      return str ;
    end

    ##------------------------------------------------------------
    ## to_i

    def to_i
      if(@type == 'atom')
	return @head.to_i ;
      else
	return 0 ;
      end
    end

    ##------------------------------------------------------------
    ## to_f

    def to_f
      if(@type == 'atom')
	return @head.to_f ;
      else
	return 0 ;
      end
    end


    ##----------------------------------------
    ## eof

    def eof()
      return @type == 'EOF' ;
    end

  end

  ##============================================================
  ## Scan

  ##------------------------------------------------------------
  ## scanStream

  def scanStream(stream)
    scanner = StreamScanner.new(stream) ;
    return scan(scanner) ;
  end 
  module_function :scanStream ;

  ##------------------------------------------------------------
  ## scan Toplevel
  
  def scan(scanner)
    begin
      token = scanner.getToken() ;
      case token.type
      when 'TermPrefix'
	return scanTerm(scanner) ;
      when 'String'
	return Form.new('string',token.str) ;
      when 'Atom'
	return Form.new('atom',token.str) ;
      when 'EOF'
	return Form.new('eof',"") ;
      else
	raise "illegal token type:" + token.to_s ;
      end

    rescue
      $stderr << "Error in scan:" ;
      $@.each do |pos|
	$stderr << "\n\t" << pos ;
      end
      $stderr << ":" << $! << "\n" ;
      $stderr << "\tline=%d, char=%d" % [scanner.lineN, scanner.charN] << "\n";
      exit ;
    end
  end
  module_function :scan ;

  ##------------------------------------------------------------
  ## scan Term
  
  def scanTerm(scanner) ;

    ##----------------------------------------
    ## scan head

    token = scanner.getToken() ;
    if(token.type == 'TermArgBegin') 
      form = Form.new('term',"") ;
    else
      form = Form.new('term',token.str) ;
      token = scanner.getToken() ;
    end
    if(token.type != 'TermArgBegin') 
      raise "illegal term begin format" + token.to_s ;
    end

    ##----------------------------------------
    ## scan role/value part

    while(TRUE)  

      ##------------------------------
      ## check end and skip separator

      token=scanner.getToken() ;
      if(token.type == 'TermArgEnd')
	return form ;
      elsif(token.type == 'TermArgGuard')
	break ;
      elsif(token.type == 'TermArgSep')
	token=scanner.getToken() ;
	if(token.type == 'TermArgEnd')
	  return form ;
	end
      end

      ##------------------------------
      ## check role

      if(token.type != 'Atom' && token.type != 'String')
	raise "illegal role name" + token.to_s ;
      end

      role = token.str ;

      ##------------------------------
      ## get equal symbol

      token = scanner.getToken() ;

      if(token.type != 'TermArgEqual')
	if(token.type == 'TermArgGuard')	## special operation
	  scanner.ungetToken(token) ;
	  form.hash['_'] = role ;
	  next ;
	end

	raise "illegal role-value format" + token.to_s ;
      end

      ##------------------------------
      ## scan value

      value = scan(scanner) ;
      form.hash[role] = value ;
      
    end

    ##----------------------------------------
    ## scan enumerable arg part

    while(TRUE) 
      token = scanner.getToken() ;
      if(token.type == 'TermArgEnd')
	return form ;
      elsif(token.type == 'TermArgSep')
	token = scanner.getToken() ;
      end
      
      scanner.ungetToken(token) ;
      value = scan(scanner) ;
      form.array.push(value) ;
    end

  end
  module_function :scanTerm ;

end

