## -*- Mode: Ruby -*-

require 'gtk' ;
include Math ;

def abs(v)
  return (v > 0) ? v : -v ;
end

##======================================================================
## class UnitWidget

class UnitWidget < Gtk::DrawingArea

  attr :sizeX		,true ;
  attr :sizeY		,true ;

  attr :valmax		,true ;
  attr :val		,true ;

  ##------------------------------------------------------------
  ## initialize

  def initialize(sizeX = 10, sizeY = 10)
    super() ;

    @xsize = sizeX ;
    @ysize = sizeY ;

    @valmax = 1.0 ;
    @val = 0.0 ;

    @mutex = Mutex.new() ;

    signal_connect("expose_event") { |w,e| exposeEvent(w,e) }
    set_usize(@xsize,@ysize) ;

  end

  ##------------------------------------------------------------
  ## newColoredGCbyRGB

  def newColoredGCbyRGB(r,g,b)
    color = Gdk::Color.new(r,g,b) ;
    @colormap.alloc_color(color,TRUE,TRUE) ;
    gc = Gdk::GC::new(window) ;
    gc.copy(@bgc) ;
    gc.set_foreground(color) ;
#    gc.set_background(color) ;
    gc ;
  end

  ##------------------------------------------------------------
  ## setup

  def setup()
    @buffer = Gdk::Pixmap::new(window, @xsize, @ysize, -1) ;
    @colormap = Gdk::Colormap.get_system() ;
    @bgc = self.style.bg_gc(self.state) ;
    @fgc = self.style.fg_gc(self.state) ;
    @gcBlack = newColoredGCbyRGB(0x0000, 0x0000, 0x0000) ;
    @gcWhite = newColoredGCbyRGB(0xFFFF, 0xFFFF, 0xFFFF) ;
    @gcRed   = newColoredGCbyRGB(0xFFFF, 0x0000, 0x0000) ;
    @gcGreen = newColoredGCbyRGB(0x0000, 0xFFFF, 0x0000) ;
    @gcBlue  = newColoredGCbyRGB(0x0000, 0x0000, 0xFFFF) ;
    @gcGrey  = newColoredGCbyRGB(0x8000, 0x8000, 0x8000) ;
  end

  ##------------------------------------------------------------
  ## expose Event

  def exposeEvent(w,e) 
    setup() if (@buffer.nil?) ;

    #redraw() ;
    w.window.draw_pixmap(@bgc,
			 @buffer,e.area.x,e.area.y,
			 e.area.x,e.area.y,e.area.width,e.area.height) ;
  end

  ##------------------------------------------------------------
  ## clear

  def clear(gc = @gcGrey)
    b = @buffer ;
    b.draw_rectangle(gc,TRUE,0,0,@xsize,@ysize) ;
    b.draw_rectangle(@gcBlack,FALSE,0,0,@xsize-1,@ysize-1) ;
  end

  ##------------------------------------------------------------
  ## redraw

  def redraw(exposep = FALSE)

    setup() if (@buffer.nil?) ;

    if(@val > 0) 
      r = @val/@valmax ;
      bggc = @gcWhite ;
      fggc = @gcRed ;
    else
      r = (1 + @val/@valmax) ;
      bggc = @gcBlue ;
      fggc = @gcWhite ;
    end

    x0 = (1 - r) * (@xsize / 2) + 1 ;
    dx = r * @xsize - 1;
    y0 = (1 - r) * (@ysize / 2) + 1 ;
    dy = r * @ysize - 1;
    
    clear(bggc) ;
    @buffer.draw_rectangle(fggc,TRUE,x0,y0,dx,dy) ;

    if(exposep) 
      draw(Gdk::Rectangle.new(0,0,@xsize,@ysize)) ;
    end
  end

  ##------------------------------------------------------------
  ## changeVal

  def changeVal(newval) 
    @val = newval ;
    redraw(TRUE) ;
  end

end





    
