// -*- Mode: c++ -*-
//Header:
//File: splayer
//Author: NODA, Itsuki
//Date: 2001/04/03
//EndHeader:

//======================================================================
// Headers

#include "MentalWorld.h"
#include "TrainSetBase.h"
#include "itk/Geo2D.h"

using namespace Itk ;
using namespace Nnn ;

typedef Geo2D::Vec Vec ;

//======================================================================
// options

#define NetFileBase "splayer"

char * InNetFile = NetFileBase ".net" ;
char * OutNetFile = NetFileBase ".net" ;
char * TplNetFile = NetFileBase ".tpl" ;
char * TmpDumpFile = NetFileBase ".dump.net" ;
char * XpdNetFile = NetFileBase ".xpd.net" ;

char * LogFile = NetFileBase ".log" ;

char * SdtFile = "slog.sdt" ;

UInt LoopN = 10 ;
UInt LearnN = 100000 ;
UInt TestN = 50 ;
UInt AdaptN = 20 ;
UInt ExpandN = 10 ;
Flt Beta = 0.1 ;

Bool NewLogP = False ;
Bool LogP = True ;
UInt LogN = 10000 ;

Flt Noise = 0.1 ;

Bool RenewP = False ;

UInt PlanN = 0 ;

Bool TestP = False ;

//Flt RealOffRate = 0.33 ;
Flt RealOffRate = 0.0 ;
Flt NoisedOffRate = 0.2 ;

//UInt FrameN = 2 ;
UInt FrameN = 5 ;


//======================================================================
// getopt

void getopt(int & argc, char ** & argv) {
   OptArg opt[] = {
      {"infile",    OptArg::cstringVal,	&InNetFile, 0,	"netfile_for_input"},
      {"outfile",   OptArg::cstringVal,	&OutNetFile,0, 	"netfile_for_output"},
      {"tplfile",   OptArg::cstringVal,	&TplNetFile,0, 	"netfile_for_template"},
      {"logfile",   OptArg::cstringVal,	&LogFile, 0,	"logfile"},

      {"sdtfile",   OptArg::cstringVal,	&SdtFile, 0,	"sdtfile"},


      {"loopN",	    OptArg::IntVal,	&LoopN,      0,	"loop_N"},
      {"learnN",    OptArg::IntVal,	&LearnN,     0,	"learn_N"},
      {"testN",     OptArg::IntVal,	&TestN,      0,	"test_N"},
      {"adaptN",    OptArg::IntVal,	&AdaptN,     0, "adapt_N"},
      {"frameN",    OptArg::IntVal,	&FrameN,     0, "frame_N"},
      {"expandN",   OptArg::IntVal,	&ExpandN,    0, "expand_N"},
      {"planN",	    OptArg::IntVal,	&PlanN,      0, "plan_N"},

      {"beta", 	    OptArg::FltVal,	&Beta,       0, "beta ; beta value for adaptation."},

      {"log",       OptArg::OnOff,	&LogP, 0,	"[on/off]"},
      {"logN",      OptArg::IntVal,	&LogN, 0,	"log_N"},
      {"newlog",    OptArg::OnOff,	&NewLogP, 0,	"[on/off]"},

      {"noise",	    OptArg::FltVal,	&Noise, 0,   "noise ; noise for pot input."},
      {"realoff",   OptArg::FltVal,	&RealOffRate,0, "rate ; off rate of real sensor value."},
      {"noisedoff", OptArg::FltVal,	&NoisedOffRate,0, "rate ; off rate of noised sensor value."},

      

      {"renew",	    OptArg::OnOff,	&RenewP,   0,   "renew from .tpl file"},

      {"test",	    OptArg::OnOff,	&TestP,   0,   "for test"},

      {ITK_NULLPTR}} ;

   myGetOpt(opt,argc,argv) ;

   if(RenewP) {
      NewLogP = True ;
      InNetFile = TplNetFile ;
   }
   
}

//======================================================================
// TrainSet

/*
 *(DT (t 0)
    (ch (FS (m 0)                       ;;;; coach's view
          (pp (0.000000 0.000000))      ; 2 player pos
	  (bd 0.000000)                 ; 2 player bdir
	  (hd 0.000000)                 ; 2 player hdir
	  (pv (0.000000 0.000000))      ; 2 player vel
	  (bp (0.000000 0.000000))      ; 2 ball pos
	  (bv (0.000000 0.000000))))    ; 2 ball vel
    (pl (FS (m 0)                       ;;;; player's view
          (pp (0.000000 0.000000))      ; 2 player pos
	  (bd 0.000000)                 ; 2 player bdir
	  (hd 0.000000)                 ; 2 player hdir
	  (pv (0.000000 0.000000))      ; 2 player vel
	  (bp (0.000000 0.000000))      ; 2 ball pos
	  (bv (0.000000 0.000000))))    ; 2 ball vel
    (cm (CM (m 1)(c t)                  ;;;; action
          (tp 180.000000)               ; 2 turn val
	  (dp 0.000000)                 ; 2 dash val
	  (kp 0.000000)                 ; 2 kick val
	  (kd 0.000000)                 ; 2 kick dir
	  (hp 0.000000))))              ; 2 head pow
 */

class SoccerTSet : public MwTrainSetBase {

      //------------------------------------------------------------
      // SoccerTSet:: for MwTrainSetBase
   public:
      virtual const char * name() const { return "Soccer" ; } ;
      virtual Int sensorN() const { return 6 ; } ;
      virtual Int actionN() const { return 2 ; } ;

      //============================================================
      // SoccerTSet:: class FieldState
   public:
      class FieldState : public WithDescriber {
	 public:
	    Bool validp ; /* data is valid or not */
	    Flt bdir ;  /* body dir */
	    Flt hdir ;  /* head dir */
	    Vec ppos ;	/* player pos */
	    Vec pvel ;	/* player vel */
	    Vec bpos ;  /* ball pos */
	    Vec bvel ;  /* ball vel */

	 public:
	    void reset(Bool vp = False) {
	       validp = vp ;
	       bdir = hdir = 0.0 ;
	       ppos.set(0,0) ; pvel.set(0,0) ; bpos.set(0,0) ; bvel.set(0,0) ;
	    } ;
	       
	 public:
	    void scan(SimpleSexp * sexp) {
	       if(sexp->isNil()) { /* if there are no data, it is not valid */
		  validp = False ;
		  return ;
	       }
	       
	       SimpleSexp * alist = sexp->rest() ; /* skip head */

	       SimpleSexp * fs = alist->assoc("FS")->rest() ; 
	       {
		  validp = (fs->assocValue("m")->scanAsInt() > 0) ;
		     
		  bdir = fs->assocValue("bd")->scanAsFlt() ;
		  hdir = fs->assocValue("hd")->scanAsFlt() ;

		  SimpleSexp * pp = fs->assocValue("pp") ;
		  ppos.x = pp->first()->scanAsFlt() ;
		  ppos.y = pp->second()->scanAsFlt() ;

		  SimpleSexp * pv = fs->assocValue("pv") ;
		  pvel.x = pv->first()->scanAsFlt() ;
		  pvel.y = pv->second()->scanAsFlt() ;
		  
		  SimpleSexp * bp = fs->assocValue("bp") ;
		  bpos.x = bp->first()->scanAsFlt() ;
		  bpos.y = bp->second()->scanAsFlt() ;

		  SimpleSexp * bv = fs->assocValue("bv") ;
		  bvel.x = bv->first()->scanAsFlt() ;
		  bvel.y = bv->second()->scanAsFlt() ;
	       }
	    } ;

	 public:
	    virtual void describe(ostream& o, Bool detailp = True) const {
	       describe_(o,detailp) ;
	    } ;

	    void describe_(ostream& o, Bool detailp = True, 
			   const char * indent = "\t") const {
	       o << "#FieldState[" << validp << "]" ;

	       if(detailp) {
		  o << endl ;
		  o << indent << "ppos=" << ppos 
		    << ", "   << "pvel=" << pvel << endl ;
		  o << indent << "dir: body=" << bdir 
		    << ", head=" << hdir << endl ;
		  o << indent << "bpos=" << bpos 
		    << ", "   << "bvel=" << bvel << endl ;
	       }
	    };
      } ;

      //============================================================
      // SoccerTSet:: class CommandState
   public:
      class CommandState : public WithDescriber {
	 public:
	    Bool validp ; /* data is valid or not */
	    enum CommandType { NoCom, Turn, Dash, Kick, Head } type ;
	    Flt tpow ; /* turn power */
	    Flt dpow ; /* dash power */
	    Flt kpow ; /* kick power */
	    Flt kdir ; /* kick dir */
	    Flt hpow ; /* head power */

	 public:
	    void reset(Bool vp = False) {
	       validp = vp ;
	       type = NoCom ;
	       tpow = dpow = kpow = kdir = hpow = 0.0 ;
	    } ;
	       
	 public:
	    void scan(SimpleSexp * sexp) {

	       if(sexp->isNil()) {  /* if there are no data, it is not valid */
		  validp = False ; 
		  return ; 
	       }

	       SimpleSexp * alist = sexp->rest() ; /* skip head */

	       SimpleSexp * cm = alist->assoc("CM")->rest() ; 
	       {
		  SimpleSexp * m = cm->assocValue("m") ;
		  validp = (m->scanAsInt() > 0) ;

		  SimpleSexp * c = cm->assocValue("c") ;
		  type = 
		     c->equal("t") ? Turn :
		     c->equal("d") ? Dash :
		     c->equal("k") ? Kick :
		     c->equal("h") ? Head : NoCom ;
		  
		  tpow = cm->assocValue("tp")->scanAsFlt() ;
		  dpow = cm->assocValue("dp")->scanAsFlt() ;
		  kpow = cm->assocValue("kp")->scanAsFlt() ;
		  kdir = cm->assocValue("kd")->scanAsFlt() ;
		  hpow = cm->assocValue("hp")->scanAsFlt() ;
	       }
	    } ;

	 public:
	    virtual void describe(ostream& o, Bool detailp = True) const {
	       describe_(o,detailp) ;
	    } ;

	    void describe_(ostream& o, Bool detailp = True, 
			   const char * indent = "\t") const {
	       o << "#CommandState[" << validp 
		 << ", command=" << ( type==Turn ? "turn" :
				      type==Dash ? "dash" :
				      type==Kick ? "kick" :
				      type==Head ? "head" : "nop" )
		 << "]" ;

	       if(detailp) {
		  o << endl ;
		  o << indent << "turn: " << tpow << endl ; 
		  o << indent << "dash: " << dpow << endl ;
		  o << indent << "kick: " << kpow << "," << kdir << endl ;
		  o << indent << "head: " << hpow << endl ;
	       }
	    };
      } ;

      //============================================================
      // SoccerTSet:: class LogEntry
   public:
      class LogEntry : public WithDescriber {
	 public:
	    Int time ;
	    
	    FieldState coach ;
	    FieldState player ;
	    CommandState command ;

	 public:
	    void init() {
	       reset(-1,False) ;
	    } ;

	    void reset(Int t = 0, Bool validp = False) {
	       time = t ;
	       coach.reset(validp) ;
	       player.reset(validp) ;
	       command.reset(validp) ;
	    } ;

	 public:
	    void scan(SimpleSexp * sexp) {
	       SimpleSexp * alist = sexp->rest() ; /* skip head "DT" */
	       
	       SimpleSexp * tval = alist->assocValue("t") ;
	       time = tval->scanAsInt() ;
	       
	       coach.scan(alist->assoc("ch")) ;
	       player.scan(alist->assoc("pl")) ;
	       command.scan(alist->assoc("cm")) ;

	       //ITK_DBG(*sexp) ;
	       //describe(cerr) ;
	    } ;

	 public:
	    virtual void describe(ostream& o, Bool detailp = True) const {
	       describe_(o,detailp) ;
	    } ;

	    void describe_(ostream& o, Bool detailp = True, 
			   const char * indent = "\t") const {
	       o << "#LogEntry[" << time << "]" ;

	       if(detailp) {
		  o << endl << "  Coach:" ;
		  coach.describe_(o,True,indent) ;

		  o << endl << "  Player:" ;
		  player.describe_(o,True,indent) ;

		  o << endl << "  Command:" ;
		  command.describe_(o,True,indent) ;
	       }
	    };
      } ;

      //------------------------------------------------------------
      // SoccerTSet:: static constants
   public:
      static const UInt DfltEntryN = 64 ;
      
   public:
      static const Flt pscale = 0.02 ; /* scale for network pos input */
      static const Flt vscale = 2.00 ; /* scale for network vel input */
      static const Flt ascale = 0.01 ; /* scale for network acc input */
      static const Flt dscale = 1.00 ; /* scale for network dir input */

      //------------------------------------------------------------
      // SoccerTSet:: members
   public:
      LogEntry * entry_ ;
      UInt entryN ;
      UInt entryI ;
      
      ITK_DEF_ACCESS(LogEntry & entry(UInt i), { 
	 return entry_[i % entryN] ; 
      });
      ITK_DEF_ACCESS(LogEntry & entry(), { return entry(entryI-1) ; }) ;

      istream * lstr ;
      Buffer buffer ;
      SimpleSexp::StreamInput * ginp ;

      static const Int runupLength = 10 ;
      Int runup ;

      //------------------------------------------------------------
      // SoccerTSet:: constructor and init
   public:
      SoccerTSet() { myinit() ; } ;

      void myinit() {
	 init() ;

	 entryN = DfltEntryN ;
	 entry_ = new LogEntry[entryN] ;
	 entryI = 0 ;

	 buffer.alloc(4096) ;
	 lstr = ITK_NULLPTR ;
	 ginp = ITK_NULLPTR ;

	 rewind() ;
      } ;

   public:
      void rewind() {
	 if(!isNull(ginp)) delete ginp ;
	 if(!isNull(lstr)) delete lstr ;
	 
	 lstr = new ifstream(SdtFile) ;
	 ginp = new SimpleSexp::StreamInput(lstr,&buffer) ;

	 runup = runupLength ;
      } ; 

      //------------------------------------------------------------
      // SoccerTSet:: class XptEndOfSdt

      class XptEndOfSdt : public Itk::Exception {
	 public:
	    virtual const char * name() const {
	       return "EndOfSdt" ; } ;

	    static Itk::Exception* root(char* const f, 
				   UInt const l, char* const d) {
	       static XptEndOfSdt _root ;
	       _root.set(f,l,d) ;
	       return &_root ;
	    } ;
	    
      } ;

      //------------------------------------------------------------
      // SoccerTSet:: step
   public:
      void step() {
	 
	 entryI++ ;

	_STEP_:

	 if(runup > 0) {
	    runup-- ;
	    entry().reset(0,True) ;
	 } 
	 else {
	    buffer.clear() ;
	    SimpleSexp * sexp = SimpleSexp::scan(*ginp) ;
	 
	    if(isNull(sexp)) {
	       if(True) {
		  rewind() ;
		  goto _STEP_ ;
	       } else {
		  throw ITK_XPT(XptEndOfSdt,"end of file of .sdt.") ;
	       }
	    }

	    entry().scan(sexp) ;
	 }

      } ;

      //------------------------------------------------------------
      // SoccerTSet:: cycle
   public:
      virtual void cycle(ValVector & sensor, ValVector & senMask,
			 ValVector & action, ValVector & actMask) {
	 step() ;
	 LogEntry & e = entry() ;

	 inpMask().set(1.0) ;
	 /* real value is not used for input value for learning */
	 if(fltRand(1.0) < RealOffRate) {
	    inpMask().set(0.0, 0, 4) ;
	 } else {
	    inpMask().set(1.0, 0, 4) ;
	 }

	 /* coach's view */

	 sensor[0] = pscale * e.coach.bpos.x ;
	 sensor[1] = pscale * e.coach.bpos.y ;
	 sensor[2] = vscale * e.coach.bvel.x ;
	 sensor[3] = vscale * e.coach.bvel.y ;
	 
	 if(e.coach.validp) senMask.set(1.0, 0,4) ;
	 else               senMask.set(0.0, 0,4) ;

	 /* player's view */

	 sensor[4] = pscale * e.player.bpos.x ;
	 sensor[5] = pscale * e.player.bpos.y ;

	 if(e.player.validp) senMask.set(0.3, 4,6) ; /* noised */
	 //if(e.player.validp) senMask.set(0.7, 4,6) ; /* noised */
	 else                senMask.set(0.0, 4,6) ;

	 /* action */

	 Flt bdir = e.coach.bdir ;

	 /* kick */
	 action[0] = ascale * e.command.kpow * Cos(bdir + e.command.kdir) ;
	 action[1] = ascale * e.command.kpow * Sin(bdir + e.command.kdir) ;

	 if(e.command.validp && e.coach.validp) actMask.set(1.0) ;
	 else                                   actMask.set(0.0) ;

      }  ;
} ;

//======================================================================
// logging util

#define Logging(logp,logstr,coutp,form) {\
   if(logp) {\
      logstr << form << endl ;\
   }\
   if(coutp) {\
      cout << form << endl ;\
   }\
}


//======================================================================
// run_learn_test

void run_learn_test(MentalWorld & mw, SoccerTSet & tset, ostream & logstr) {
   Val err = 0.0;
   for(UInt i = 0 ; i < LearnN ; i++) {
      //Val dsum = mw.runLearn2frame(tset,AdaptN,Beta) ;
      Val dsum = mw.runLearnNframe(tset,FrameN,AdaptN,Beta) ;

      dsum += 1; // dummy
      err *= 0.9 ;
      err += (mw.frame[FrameN-1].expect->sumError() 
	      + mw.frame[FrameN].expect->sumError()) ;
      if(mw.cntLearn() % LogN == 0) {
	 Logging(LogP,logstr,True,
		 "#SqEr["
		 << "count=" << setfill(' ') << setw(7) 
		 << mw.cntLearn() << ", "
		 << "err=" << setprecision(10) << sqrt(err) << "]") ;
	 mw.save(OutNetFile) ;
      }
   }
   if(TestN > 0) {
      Logging(LogP,logstr,True,
	      "#Trial["
	      << "n=" << TestN << ", "
	      << "format=" << "\"sensor/expect/action/sensor+/expect+\"" << "]") ;
   }
   for(UInt i = 0 ; i < TestN ; i++) {
      //mw.run2frame(tset, AdaptN, Beta) ;
      mw.runNframe(tset, FrameN, AdaptN, Beta) ;
      {
	 // for test
	 //MwTrainSetBase::Frame & nset = tset.getFrame(mw.worktime+1) ;
	 //mw.frame[FrameN].setSensor(nset.sensor(),nset.senMask()) ;
      }

      Logging(LogP,logstr,True,
	      "~" << mw.frame[FrameN-1].sensor->out() << 
	      "/" << mw.frame[FrameN-1].expect->out() << 
	      "/" << mw.frame[FrameN-0].action->out() << 
	      "/" << mw.sensorTch(tset,2) << 		/* "2" = worktime+1 */
	      "/" << mw.frame[FrameN-0].expect->out()) ;
	 
   }
} ;

//======================================================================
// planning

void planning(MentalWorld & mw, SoccerTSet & tset, ostream & logstr) {
   //Val err = 0.0 ;

   ValVector cpos(tset.sensorN()) ;
   ValVector gpos(tset.sensorN()) ;
   ValVector onmask(tset.sensorN()) ;
   ValVector onmask2(tset.sensorN()) ;

   cpos[0] = -1.0 ; cpos[1] = -1.0 ;
   gpos[0] = 1.0 ; gpos[1] = 1.0 ;

   onmask.set(0.0) ; onmask[0] = onmask[1] = 1.0 ;
   onmask2.set(0.0) ; onmask2[0] = onmask2[1] = 5.0 ;

   Int duration = ExpandN - 2 ;

   mw.simplePlanning(cpos,onmask2,gpos,onmask,duration,
		     (duration * PlanN),Beta) ;

   // distplay!!
   for(UInt i = 1 ; i < ExpandN ; i++) {
      ITK_DBG(i);
      ITK_DBG(mw.frame[i].expect->out()) ;
      ITK_DBG(mw.frame[i].action->out()) ;
   } ;
}


//======================================================================
// load_run_save

void load_run_save() {

   ofstream logstr ;
   if(LogP) {
      if(NewLogP) 
	 logstr.open(LogFile) ;
      else
	 logstr.open(LogFile,ios::app) ;
   }

   MentalWorld mw ;
   mw.load(InNetFile) ;

   Logging(LogP,logstr,True,
	   "#Load["
	   << "file=\"" << InNetFile << "\", "
	   << "time=\"" << TimeVal::now().stamp() << "\"]") ;

   mw.expand(ExpandN) ;
   
   SoccerTSet tset ;

   try {
      for(UInt l = 0 ; l < LoopN ; l++) {
	 run_learn_test(mw,tset,logstr) ;
      }
   } catch (Nnn::Exception & ex) {
      mw.save(TmpDumpFile,SL_AllState) ;
      Logging(LogP,logstr,True,
	      "#Xpt"
	      << "[" << "time=" << TimeVal::now().stamp() 
	      << "," << "xpt=" << ex 
	      << "]") ;
      exit(1) ;
   } 
   mw.save(OutNetFile) ;
   Logging(LogP,logstr,True,
	   "#Save["
	   << "file=\"" << OutNetFile << "\", "
	   << "time=\"" << TimeVal::now().stamp() << "\"]") ;
} ;

//======================================================================
// test planining

void load_planning() {

   ofstream logstr ;
   if(LogP) {
      if(NewLogP) 
	 logstr.open(LogFile) ;
      else
	 logstr.open(LogFile,ios::app) ;
   }

   MentalWorld mw ;
   mw.load(InNetFile) ;

   Logging(LogP,logstr,True,
	   "#Load["
	   << "file=\"" << InNetFile << "\", "
	   << "time=\"" << TimeVal::now().stamp() << "\"]") ;

   mw.expand(ExpandN) ;
   
   SoccerTSet tset ;

   planning(mw,tset,logstr) ;
} ;

//======================================================================
// sub for test

//------------------------------
// test 1

void test1() {
   SoccerTSet tset ;
   ValVector sensor(tset.sensorN()) ;
   ValVector senMask(tset.sensorN()) ;
   ValVector action(tset.actionN()) ;
   ValVector actMask(tset.actionN()) ;

   Val vmax = 0.0 ;
   Val kmax = 0.0 ;

   try {
      for(UInt i = 0 ; i < 10000 ; i++) {
	 tset.cycle(sensor,senMask,action,actMask) ;
	 //cout << i << ":" << tset.entry().time << ":" 
	 //     << sensor << "/" << action << endl ;
	 //tset.entry().describe(cout) ;
#if 0
	 cout << "p:" << NNN_VAL_FORM(8,4) << sensor[0] 
	      << ","  << NNN_VAL_FORM(8,4) << sensor[1] << " " ;
	 cout << "v:" << NNN_VAL_FORM(8,4) << sensor[2] 
	      << ","  << NNN_VAL_FORM(8,4) << sensor[3] << " " ;
	 cout << "a:" << NNN_VAL_FORM(8,4) << action[0] 
	      << ","  << NNN_VAL_FORM(8,4) << action[1] << endl ;
	 if(Abs(vmax) < Abs(sensor[2])) vmax = sensor[2] ;
	 if(Abs(vmax) < Abs(sensor[3])) vmax = sensor[3] ;
#endif
	 if(Abs(action[6])+Abs(action[7]) > 0.1) {
	    cout << "k:" << NNN_VAL_FORM(8,4) << action[6] 
		 << ","  << NNN_VAL_FORM(8,4) << action[7] << endl ;
	 }
	 if(Abs(kmax) < Abs(action[6])) kmax = action[6] ;
	 if(Abs(kmax) < Abs(action[7])) kmax = action[7] ;
      } 
   } catch (Itk::Exception * ex) {
      ITK_DBG(*ex) ;
   } 

   //cout << "vmax = " << vmax << endl ;
   cout << "kmax = " << kmax << endl ;
   exit(0) ;
}

//------------------------------
// test 2

void test2() {

   MentalWorld mw ;
   mw.load(InNetFile) ;
   mw.expand(ExpandN) ;
   
   SoccerTSet tset ;

   for(UInt i = 0 ; i < TestN ; i++) {
      mw.runNframe(tset, FrameN, AdaptN, Beta) ;
      ITK_DBG(mw.frame[FrameN-1].ollinp->out()) ;
      ITK_DBG(mw.frame[FrameN-1].ollout->out()) ;
   }

   exit(0) ;
} 

//------------------------------
// test 3

void test3() {
   SoccerTSet tset ;

   try {
      for(UInt i = 0 ; i < 10000 ; i++) {
	 tset.step() ;
	 if(tset.entry().player.validp && tset.entry().coach.validp) {
	    Vec diff = 
	       tset.entry().player.ppos 
	       - tset.entry().coach.ppos ;
	    cout << tset.entry().coach.ppos.x 
		 << "\t" << tset.entry().coach.ppos.y
		 << "\t" << diff.x 
		 << "\t" << diff.y << endl ;
	 }
      } 
   } catch (Itk::Exception * ex) {
      ITK_DBG(*ex) ;
   } 
   exit(0) ;
}

//======================================================================
// main routine

int main(int argc, char* argv[]) {

   getopt(argc,argv) ;
   
   if(TestP) {
      test1() ;
      //test2() ;
      //test3() ;
   }

   if(PlanN == 0) {
      load_run_save() ;
   } else {
      load_planning() ;
   }

   return 0;
}










