// -*- Mode:C++ -*-
//Header:
//File: libsclient.h
//Author: NDOA, Itsuki
//Date: 1999/04/15
//

//ModifyHistory:
// 1999/04/15: Create this file
//EndModifyHistory:

#ifndef _CycleStep_h_
#define _CycleStep_h_
////////////////////////////////////////////////////////////////////////

namespace Saloo {

   using namespace Itk ;

   //======================================================================
   // Count & Time

   class CycleStep : public WithDescriber {
      public:
	 UInt count ;
	 TimeVal rtime;

	 //--------------------------------------------------
	 // constructor
      public:
	 CycleStep() { set(0,True) ; } ;
	 CycleStep(UInt c) { init(c) ; } ;
	 CycleStep(const CycleStep& src) {
	    set(src) ;
	 } ;

	 //--------------------------------------------------
	 // init
      public:
	 void init(UInt c = 0) {
	    set(c) ;
	 } ;

	 //--------------------------------------------------
	 // update rtime
      public:
	 void update() { rtime.update() ; } ;

	 //--------------------------------------------------
	 // set
      public:
	 CycleStep& set(UInt c, Bool updatep = True) {
	    count = c ;
	    if(updatep) update() ;
	    return *this ;
	 } ;
	 CycleStep& set(const CycleStep& src) {
	    count = src.count ;
	    rtime = src.rtime ;
	    return *this ;
	 } ;

	 CycleStep& operator=(const CycleStep& src) { return set(src) ; } ;
	 CycleStep& operator=(const UInt c) { return set(c) ; } ;

	 CycleStep& setupdate(const CycleStep& src) {
	    return set(src.count) ;
	 } ;

	 Bool operator==(const CycleStep& ref) { 
	    return count == ref.count ; 
	 } ;

	 //--------------------------------------------------
	 // increment
      public:
	 CycleStep& operator++() { return inc() ; } ;
	 CycleStep& operator++(int) { return inc() ; } ;

	 CycleStep& inc(Bool updatep = True) {
	    count++ ;
	    if(updatep) update() ;
	    return *this ;
	 } ;

	 //--------------------------------------------------
	 // comparison
      public:
	 Bool isNewer(const CycleStep & ref, Int margin = 0) const {
	    return (count + margin) > ref.count ;
	 } ;
	 Bool isOlder(const CycleStep & ref, Int margin = 0) const {
	    return (count + margin) < ref.count ;
	 } ;

	 Bool operator<(const CycleStep & ref) const {
	    return count < ref.count ;
	 } ;
	 Bool operator<=(const CycleStep & ref) const {
	    return count <= ref.count ;
	 } ;
	 Bool operator>(const CycleStep & ref) const {
	    return count > ref.count ;
	 } ;
	 Bool operator>=(const CycleStep & ref) const {
	    return count >= ref.count ;
	 } ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#CycleStep[" << count << ":" ;
	    rtime.describe(ostr,detailp) ;
	    ostr << "]" ;
	    if(detailp) ostr << endl ;
	 } ;
   } ;
} ;
////////////////////////////////////////////////////////////////////////
#endif
