// -*- Mode:C++ -*-
//Header:
//File: OfflineCoach.h
//Title: OfflineCoach interface
//Author: NODA, Itsuki
//Date: 2001/03/12
//

//ModifyHistory:
// 2001/03/12: Start to create this file
//EndModifyHistory:

#ifndef _saloo_OfflineCoach_h_
#define _saloo_OfflineCoach_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"

namespace Saloo {

   using namespace Itk ;
   using namespace Soccer ;

   //----------------------------------------------------------------------
   // constants

   //======================================================================
   // class OfflineCoach

   class OfflineCoach : public WithDescriber {
      public:
	 virtual const char * _classname() const { return "OfflineCoach" ; } ;

	 //--------------------------------------------------
	 // constants

	 static const UInt playerN = SC_N_PLAYER_PER_TEAM ;
	 static const UInt playerM = 2 * SC_N_PLAYER_PER_TEAM ;

	 //----------------------------------------
	 // for debug
   
	 static const VbsLevel Vbs_Comm = 5 ;
	 static const VbsLevel Vbs_Test = 1 ;

	 //==================================================
	 // PlayMode
	 
	 typedef ScWord * PlayMode ;
	 static char * playmodeName(const PlayMode & pm) {
	    return pm->name ;
	 } ;

#define PLAYMODE(id) &(SC_WORD(id))

	 //==================================================
	 // class OfflineCoach::SeeInfo
      public:
	 class SeeInfo : public WithDescriber {
	       //========================================
	       // class OffineCoach::SeeInfo::Object
	    public:
	       class Object : public WithDescriber {
		     //------------------------------
		     // Type
		  public:
		     enum Type {
			Ball,
			Player,
			Other
		     } ;

		     static const char * typestr(Type type) {
			switch(type) {
			   case Ball:	return "ball" ;
			   case Player: return "player" ;
			   case Other:  return "other" ;
			   default:     return "unknown" ;
			} ;
		     } ;

		     //==============================
		     // body (Object)

		     //------------------------------
		     // member
		  public:
		     Type type ;
		     Int  aux ;
		     PosDir _posdir ;
		     Vec  _vel ;
 
		     //------------------------------
		     // access
		  public:
		     ITK_DEF_ACCESS(PosDir & posdir(), { 
			return _posdir ; 
		     }) ;
		     ITK_DEF_ACCESS(Pos & pos(), { 
			return posdir().pos ; 
		     }) ;
		     ITK_DEF_ACCESS(Angle & bdir(), { 
			return  posdir().bdir ;
		     }) ;
		     ITK_DEF_ACCESS(Angle & hdir(), { 
			return  posdir().hdir ;
		     }) ;
		     ITK_DEF_ACCESS(Vec & vel(), { 
			return  _vel ;
		     }) ;

		     //------------------------------
		     // describe
		  public:
		     void describe(ostream & ostr, 
				   const Bool detailp = True) const {
			ostr << "#Object[" << typestr(type) 
			     << ":" << aux << "]" ;

			if(detailp) {
			   ostr << endl ; 
			   ostr << "\t" << "posdir=" << posdir() << endl ;
			   ostr << "\t" << "vel=" << vel() << endl ;
			}
		     } ;
	       } ;
	       //========================================
	       // body (SeeInfo)

	       //----------------------------------------
	       // member
	    public:
	       Object _ball ;
	       Object _teammate[playerN] ;
	       Object _opponent[playerN] ;
	       
	       SubString teamname ;
	       Side side ;
	       CycleStep time ;
	       
	       SimpleSexp::Heap sxheap ;

	       //----------------------------------------
	       // setup
	    public:
	       void setup(SubString & tname, Side s) {
		  teamname = tname ;
		  side = s ;

		  ball().type = Object::Ball ;
		  for(UInt i = 0 ; i < playerN ; i++) {
		     teammate(i).type = Object::Player ;
		     teammate(i).aux = i ;
	     
		     opponent(i).type = Object::Player ;
		     opponent(i).aux = i ;
		  }
	       } ;

	       //----------------------------------------
	       // access
	    public:
	       ITK_DEF_ACCESS(Object & ball(), { return _ball ;}) ;
	       
	       ITK_DEF_ACCESS(Object * teammate(), { return _teammate ;});
	       ITK_DEF_ACCESS(Object & teammate(UInt i), {
		  return _teammate[i] ;
	       }) ;

	       ITK_DEF_ACCESS(Object * opponent(), { return _opponent ;});
	       ITK_DEF_ACCESS(Object & opponent(UInt i), {
		  return _teammate[i] ;
	       }) ;

	       //----------------------------------------
	       // scan
	    public:
	       Bool scan(SimpleSexp & s)  ;
	       Bool scan(SimpleSexp * s) { return scan(*s) ; } ;
	       Bool scan(Buffer & buffer) {
		  return scan(SimpleSexp::scan(buffer,&sxheap)) ;
	       } ;
	       
	       //----------------------------------------
	       // utility
	    public:
	       Bool isLeft() { return Soccer::isLeft(side) ; } ;
	       Bool isRight() { return Soccer::isRight(side) ; } ;

	       Bool isMyTeam(SimpleSexp * s) { return s->equal(teamname) ;} ;

	       //----------------------------------------
	       // describe
	    public:
	       void describe(ostream & ostr, 
			     const Bool detailp = True) const {
		  ostr << "#SeeInfo[" << (void *)this <<  "]" ;

		  if(detailp) {
		     ostr << endl ;

		     ostr << "\t" << "ball:" ;
		     ball().describe(ostr,detailp) ;

		     ostr << "\t" << "teammate:" << endl ;
		     for(UInt i = 0 ; i < playerN ; i++) {
			ostr << "\t\t" << "[" << i << "] :" ;
			teammate(i).describe(ostr,detailp) ;
		     }

		     ostr << "\t" << "opponent:" << endl ;
		     for(UInt i = 0 ; i < playerN ; i++) {
			ostr << "\t\t" << "[" << i << "] :" ;
			opponent(i).describe(ostr,detailp) ;
		     }
		  }
	       } ;
	 } ;
   
	 //==================================================
	 // body (OfflineCoach)
   
	 //----------------------------------------
	 // member
      public:
	 ClientSocket socket ;
	 SubString sschost ;
	 UInt sscport ;
	 Flt sscproto ;
	 
	 Buffer sendbuf ;
	 Buffer recvbuf ;
	 SimpleSexp * recvsexp ;
	 SimpleSexp::Heap sxheap ;
	 
	 CycleStep sendCount ;
	 CycleStep recvCount ;

	 MutexCondVar cvSensor ;
	 MutexCondVar cvSee ;

	 MutexCondVar cvSend ;

	 SeeInfo sinfo ;
	 Thread * thrdSensor ;

	 //----------------------------------------
	 // descructor
      public:
	 virtual ~OfflineCoach() {} ;

	 //----------------------------------------
	 // init
      public:
	 Bool init(const char * const shost = DfltSSHost,
		   UInt sport = DfltSSOffCPort,
		   Flt proto = DfltSSProto,
		   Bool openp = True) {
	    sschost = shost ;
	    sscport = sport ;
	    sscproto = proto ;

	    sendbuf.alloc(DfltBufSize) ;
	    recvbuf.alloc(DfltBufSize) ;

	    Bool r = False ;
	    if(openp) r = open() ;
	    return r ;
	 } ;
	    

	 //----------------------------------------
	 // open connection
      public:
	 Bool open() {
	    socket.init((const char *)sschost,Socket::UDP,sscport) ;
	    socket.setBlocking() ;

	    return True ;
	 } ;

	 //----------------------------------------
	 // send change_mode
	 //   possible pm ::= PLAYMODE(mode)
	 //            mode ::= PlayOn | TimeOver | BeforeKickOff
	 //                   | KickOffL | KickOffR | KickInL | KickInR
	 //                   | FreeKickL | FreeKickR 
	 //                   | CornerKickL | CornerKickR
	 //                   | GoalKickL | GoalKickR
      public:
	 Bool sendChangeMode(const PlayMode & pm) ;

	 //----------------------------------------
	 // send move
      public:
	 Bool sendMoveBall(const Flt x, const Flt y) ;
	 Bool sendMoveBall(const Pos & pos) {
	    return sendMoveBall(pos.x,pos.y) ;
	 } ;

	 Bool sendMovePlayer(const SubString & teamname, const Int unum,
			     const Flt x, const Flt y, const Flt dir) ;
	 Bool sendMovePlayer(const SubString & teamname, const Int unum,
			     const Flt x, const Flt y) ;

	 Bool sendMovePlayer(const SubString & teamname, const Int unum,
			     const Pos & pos, const Flt dir) {
	    return sendMovePlayer(teamname,unum, pos.x, pos.y, dir) ;
	 } ;
	 Bool sendMovePlayer(const SubString & teamname, const Int unum,
			     const Pos & pos) {
	    return sendMovePlayer(teamname,unum, pos.x, pos.y) ;
	 } ;

	 Bool sendMovePlayer(const SubString & teamname, const Int unum,
			     const PosDir & posdir) {
	    return sendMovePlayer(teamname,unum, posdir.pos, posdir.bdir) ;
	 } ;

	 //----------------------------------------
	 // sensor
      public:
	 Bool cycleSensor() ;
	 Bool scanSensor(SimpleSexp & s) ;
	 Bool scanSensor(SimpleSexp * s) { return scanSensor(*s) ; } ;

	 //----------------------------------------
	 // sensor thread
      public:
	 Thread * forkSensor() ;
	 static void * runSensor(void *p) ;
	 void * sensorMainLoop() ;

	 //----------------------------------------
	 // basic utility
      public:
	 Bool send() ;
	 Bool recv() ;

	 //----------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#" << _classname() << "[" << (void*)this << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "\t" "sshost:port=" 
		    << sschost << ":" << sscport << endl ;
	       ostr << "\t" "protocol_version=" << sscproto << endl ;
	       describeAux(ostr,detailp) ;
	    } 
	 } ;

	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;


} ;

////////////////////////////////////////////////////////////////////////
#endif

