// -*- Mode:C++ -*-
//Header:
//File: SimplePlayer.h
//Title: Simple Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/10
//

//ModifyHistory:
// 2001/03/10: Start to create this file
//EndModifyHistory:

#ifndef _saloo_SimplePlayer_h_
#define _saloo_SimplePlayer_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "BasePlayer.h"
#include "estimatePos.h"
#include "FieldState.h"
#include "ThreadedPlayer.h"
#include "WorldModel.h"

namespace Saloo {

   using namespace Itk ;
   using namespace Soccer ;

   //======================================================================
   // Simple Player

   class SimplePlayer : public ThreadedPlayer {
      public:
	 virtual const char * _classname() const { return "SimplePlayer";};

	 //==================================================
	 // class StateHistory

	 class StateHistory : public RingBufferT<FieldState> {
	    public:
	 } ;

	 //==================================================
	 // body (SimplePlayer)

	 //--------------------------------------------------
	 // member
      public:

	 //------------------------------
	 // timestamps

	 CycleStep _stime ;	/* most recent sensor timestamp */
	 CycleStep _atime ;	/* most recent audio sensor timestamp */
	 //CycleStep _vtime ;   /* most recent visual sensor timestamp */
	 CycleStep _btime ;	/* most recent sense body timestamp */
	 CycleStep _fstime ;    /* most recent fullstate timestamp */

	 CycleStep _ctime ;     /* most recent command timestamp */
	 CycleStep _bctime ;	/* most recent body command timestamp */

	 //------------------------------
	 // FieldState

	 FieldState _currentState ;	/* by sensor information */
	 FieldState _taughtState ;	/* by fullstate information */

	 StateHistory _stateHistory ;
	 CycleStep senseStateTime ;
	 CycleStep estimateStateTime ;

	 /* <<< [I.Noda:2001.07.26] <<< */
	 //------------------------------
	 // WorldModel

	 WorldModel worldmodel ;
	 /* >>> [I.Noda:2001.07.26] >>> */

	 //------------------------------
	 // major landmark

	 Landmark * _goal ;
	 Landmark & goal() { return *_goal ; } ;
	 Vec & goalpos() { return goal().pos ; } ;

	 Landmark * _owngoal ;
	 Landmark & owngoal() { return *_owngoal ; } ;
	 Vec & owngoalpos() { return owngoal().pos ; } ;

	 //------------------------------
	 // messages and playmode

	 MutexCondVar cvPlayMode ;

	 //--------------------------------------------------
	 // constructor
      public:
	 SimplePlayer() {
	    _goal = getLandmark(LMID_GR) ;
	    _owngoal = getLandmark(LMID_GL) ;
	    senseStateTime.set(0,True) ;
	    estimateStateTime.set(0,True) ;
	 } ;
	 
	 //--------------------------------------------------
	 // access

	 //------------------------------
	 // taught state by fullstate info
      public:
	 ITK_DEF_ACCESS(FieldState & taughtState(), {
	    return _taughtState ;
	 }) ;

	 //------------------------------
	 // field state
      public:
	 ITK_DEF_ACCESS(virtual FieldState & currentState(), {
	    return _currentState ;
	 }) ;

	 ITK_DEF_ACCESS(CycleStep & stime(), { 	// sensor time
	    return _stime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & atime(), { 	// hear time
	    return _atime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & btime(), { 	// sense body time
	    return _btime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & fstime(), { // full state time
	    return _fstime ;
	 });

	 ITK_DEF_ACCESS(CycleStep & vtime(), { 	// visual sensor time
	    return currentState().time();
	 });

	 ITK_DEF_ACCESS(CycleStep & epostime(), { // estimate sensor time
	    return currentState().time();
	 });

	 ITK_DEF_ACCESS(CycleStep & ctime(), { // last command time
	    return _ctime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & bctime(), { // last body command time
	    return _bctime ;
	 });

	 ITK_DEF_ACCESS(PosDir & myposdir(), { 
	    return currentState().myposdir() ;
	 });
	 ITK_DEF_ACCESS(Vec & mypos(), { 
	    return currentState().mypos() ; 
	 });
	 ITK_DEF_ACCESS(Angle & myhdir(), { 
	    return currentState().myhdir() ; 
	 });
	 ITK_DEF_ACCESS(Angle & mybdir(), { 
	    return currentState().mybdir() ; 
	 });

	 /* now, analyzeSensorSenseBody() maintain myneck and myneckangle
	  */
	 ITK_DEF_ACCESS(Angle & myneck(), {
	    //return sinfo.senseBody.headAngle ;  
	    return currentState().neckAngle() ;
	 });

	 ITK_DEF_ACCESS(Flt & myspeed(), {
	    return sinfo.senseBody.speed ;
	 }) ;

	 ITK_DEF_ACCESS(Flt & posConf(), { 
	    return currentState().posConf() ;
	 });
	 ITK_DEF_ACCESS(Vec & myvel(), {
	    return currentState().myvel() ;
	 }) ;
	 ITK_DEF_ACCESS(Flt & velConf(), {
	    return currentState().velConf() ;
	 }) ;
	 ITK_DEF_ACCESS(Bool & outofFieldP(), {
	    return currentState().outofFieldP() ;
	 }) ;
	 ITK_DEF_ACCESS(FieldState::Object & ball(), { 
	    return currentState().ball() ;
	 }) ;
	 ITK_DEF_ACCESS(Vec & ballpos(), {
	    return currentState().ballpos() ;
	 }) ;
	 ITK_DEF_ACCESS(Vec & ballvel(), {
	    return currentState().ballvel() ;
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & teammate(UInt i), {
	    return currentState().teammate(i) ; 
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & opponent(UInt i), {
	    return currentState().opponent(i) ;
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & player(UInt i), {
	    return currentState().player(i) ;
	 }) ;

	 //------------------------------
	 // field state history
      public:
	 ITK_DEF_ACCESS(FieldState & stateEntry(UInt cycle), {
	    return _stateHistory[cycle] ;
	 }) ;

	 //FieldState & estimateState(UInt cycle) { } ; /////

	 FieldState & pushState(const CycleStep & time, const FieldState & s) ;

	 /* <<< [I.Noda:2001.07.26] <<< */
	 //------------------------------
	 // access to world model
      public:
	 WorldModel::State & estState() { return estState(stime().count) ;};

	 WorldModel::State & estState(Int t) { return worldmodel.estState(t);};
	 
	 //--------------------
	 // access to world model (self)
      public:
	 WorldModel::Object & estSelf() { return estSelf(stime().count) ;} ;
	 WorldModel::Object & estSelf(Int t) { return worldmodel.estSelf(t) ;};

	 //--------------------
	 // access to world model (ball)
      public:
	 WorldModel::Object & estBall() { return estBall(stime().count) ;};
	 WorldModel::Object & estBall(Int t) { return worldmodel.estBall(t) ;};

	 //--------------------
	 // access to world model (selfpos & ballvel)
      public:
	 Vec & estSelfPos() { return estSelfPos(stime().count) ;} ;
	 Vec & estSelfPos(Int t) { return estSelf(t).pos() ;} ;
      public:
	 Vec & estSelfVel() { return estSelfVel(stime().count) ;} ;
	 Vec & estSelfVel(Int t) { return estSelf(t).vel() ;} ;

	 //--------------------
	 // access to world model (ballpos & ballvel)
      public:
	 Vec & estBallPos() { return estBallPos(stime().count) ;} ;
	 Vec & estBallPos(Int t) { return estBall(t).pos() ;} ;

      public:
	 Vec & estBallVel() { return estBallVel(stime().count) ;} ;
	 Vec & estBallVel(Int t) { return estBall(t).vel() ;} ;
	       
	 /* >>> [I.Noda:2001.07.26] >>> */

	 //--------------------------------------------------
	 // analyze see/hear info
      public:

	 virtual Bool analyzeSensorSee() ;
	 virtual Bool analyzeSensorHear() ;
	 virtual Bool analyzeSensorSenseBody() ;
	 virtual Bool analyzeSensorFullState() ;

	 //--------------------------------------------------
	 // command timing
      public:
	 Bool waitNextBestComTime(UInt marginMS = 50) ;
	 // Bool waitNextBestComTime(UInt marginMS = 100) ;

	 //--------------------------------------------------
	 // basic command
      public:

	 virtual Bool move(const Flt & x, const Flt & y) ;
	 Bool move(const Vec & pos) { return move(pos.x,pos.y) ; } ;

	 virtual Bool turn(const Flt & moment) ;
	 virtual Bool dash(const Flt & power) ;
	 virtual Bool kick(const Flt & power, const Flt & dir) ;
	 virtual Bool gcatch(const Flt & dir) ; //goalie catch

	 virtual Bool head(const Flt & dir) ;  // turn_neck

	 virtual Bool say(const SubString & msg) ;
	 Bool say(const char * msg) { 
	    return say(SubString(const_cast<char *>(msg))) ; 
	 } ;
	 
	 //--------------------------------------------------
	 // low level command

	 //------------------------------
	 // wrappter for superclass
      public:
	 virtual Bool sendCommandPost(Bool bodyp)  ;

	 //------------------------------
	 // turn
      public:
	 virtual Bool turnTo(const Flt adir) ;
	 virtual Bool turnToRel(const Vec & rpos) ;
	 virtual Bool turnToAbs(const Vec & apos) ;

	 //------------------------------
	 // kick
      public:
	 virtual Bool kickTo(const Flt & power, const Flt & adir) ;
	 virtual Bool kickToRel(const Flt & power, const Vec & rpos) ;
	 virtual Bool kickToAbs(const Flt & power, const Vec & apos) ;

	 //------------------------------
	 // kick
      public:
	 virtual Bool gcatchTo(const Flt & adir) ;
	 virtual Bool gcatchToRel(const Vec & rpos) ;
	 virtual Bool gcatchToAbs(const Vec & apos) ;

	 //------------------------------
	 // neck
      public:
	 virtual Bool headTo(const Angle & adir) ;
	 virtual Bool headToRel(const Vec & rpos) ;
	 virtual Bool headToAbs(const Vec & apos) ;
	 
	 //--------------------------------------------------
	 // timestamp check
      public:
	 static const Int SeenMargin = 5 ;

	 Bool isBallSeenRecently(Int margin = SeenMargin) const {
	    return ball().time().isNewer(stime(),margin) ;
	 } ;

	 Bool isMyPosEstimatedRecently(Int margin = SeenMargin) const {
	    return epostime().isNewer(stime(),margin) ;
	 } ;

	 //--------------------------------------------------
	 // kickable check
      public:
	 Bool kickable() const {
	    return ballpos().distance(mypos()) < SC_KICKABLE_AREA ;
	 } ;

	 //--------------------------------------------------
	 // catchable check
      public:
	 Bool isInPenaltyArea(const Pos & pos, Flt margin = 0.0) const {
	    return (pos.x < -((SC_PITCH_LENGTH/2.0) 
			      - SC_PENALTY_AREA_LENGTH) - margin &&
		    Abs(pos.y) < (SC_PENALTY_AREA_WIDTH/2.0) - margin) ;
	 } ;

      public:
	 Bool gcatchable() const {
	    if(!isPlayOn()) return False ;
	    return (isInPenaltyArea(ballpos()) &&
		    ballpos().distance(mypos()) < SC_CATCHABLE_AREA_L) ;
	 } ;

	 //--------------------------------------------------
	 // mode check

	 //------------------------------
	 // can use move ?
      public:
	 Bool canMove() const {
	    if(isNull(playmode())) return False ;
	    switch(playmode()->id) {
	       case SC_WID_BeforeKickOff:
	       case SC_WID_HalfTime:
	       case SC_WID_GoalL:
	       case SC_WID_GoalR:
	       case SC_WID_OffSideL:
	       case SC_WID_OffSideR:
		  return True ;
	       default:
		  return False ;
	    }
	 } ;
	 
	 //------------------------------
	 // should be init pos?
      public:
	 Bool shouldBeInitPos() const {
	    if(isNull(playmode())) return False ;
	    switch(playmode()->id) {
	       case SC_WID_BeforeKickOff:
	       case SC_WID_HalfTime:
	       case SC_WID_GoalL:
	       case SC_WID_GoalR:
		  return True ;
	       default:
		  return False ;
	    }
	 } ;

	 //------------------------------
	 // my goal kick?
      public:
	 Bool isMyGoalKick() const {
	    return playmode()->id == (isLeft(side)  ? 
				      SC_WID_GoalKickL :
				      SC_WID_GoalKickR) ;
	 } ;

	 //------------------------------
	 // my kick off?
      public:
	 Bool isMyKickOff() const {
	    return playmode()->id == (isLeft(side)  ? 
				      SC_WID_KickOffL :
				      SC_WID_KickOffR) ;
	 } ;

	 //------------------------------
	 // opponent kick off?
      public:
	 Bool isOpponentKickOff() const {
	    return playmode()->id == (isLeft(side)  ? 
				      SC_WID_KickOffR :
				      SC_WID_KickOffL) ;
	 } ;

	 //------------------------------
	 // playOn?
      public:
	 Bool isPlayOn() const {
	    switch(playmode()->id) {
	       case SC_WID_PlayOn:
		  return True ;
	       default:
		  return False ;
	    }
	 } ;

	 //------------------------------
	 // restart mode by side s ?
      public:
	 Bool isRestart(Side s) const {
	    if(isLeft(s)) {
	       switch(playmode()->id) {
		  case SC_WID_KickOffL:
		  case SC_WID_FreeKickL:
		  case SC_WID_CornerKickL:
		  case SC_WID_GoalKickL:
		  case SC_WID_GoalieCatchL:
		  case SC_WID_OffSideR:
		  case SC_WID_DropBall:
		     return True ;
		  default:
		     return False ;
	       }
	    } else {
	       switch(playmode()->id) {
		  case SC_WID_KickOffR:
		  case SC_WID_FreeKickR:
		  case SC_WID_CornerKickR:
		  case SC_WID_GoalKickR:
		  case SC_WID_GoalieCatchR:
		  case SC_WID_OffSideL:
		  case SC_WID_DropBall:
		     return True ;
		  default:
		     return False ;
	       }
	    }
	 } ;
	 
	 //------------------------------
	 // my restart mode?
      public:
	 Bool isMyRestart() const { return isRestart(side) ; } ;
	 
	 //------------------------------
	 // opponent restart mode?
      public:
	 Bool isOpponentRestart() const {
	    return (isRestart(otherside(side)) && 
		    (playmode()->id != SC_WID_DropBall)) ;
	 } ;

	 //------------------------------
	 // can play 
      public:
	 Bool canPlay() const { return isPlayOn() || isMyRestart() ; } ;

	 //------------------------------
	 // (can use dash?)
      public:
	 Bool canDash() const { 
	    switch(playmode()->id) {
	       case SC_WID_BeforeKickOff:
	       case SC_WID_HalfTime:
		  return False ;
	       default:
		  return True ;
	    }
	 } ;

	 //--------------------------------------------------
	 // describe Aux
      public:
	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;
} ;
////////////////////////////////////////////////////////////////////////
#endif
	 


   
