// -*- Mode:C++ -*-
//Header:
//File: WorldModel.cc
//Title: WorldModel.cc
//Author: NODA, Itsuki
//Date: 2001/07/27
//

//ModifyHistory:
// 2001/07/27: Start to create this file
//EndModifyHistory:

#include "WorldModel.h"

namespace Saloo {

   //======================================================================
   // WorldModel:: class State

   //======================================================================
   // class WorldModel

   //------------------------------------------------------------
   // WorldModel:: estState

   WorldModel::State & WorldModel::estState(Int t, Bool lockp) {
      State * r = ITK_NULLPTR ;

      if(lockp) mutex.lock() ;
      {
	 if(t < worktime - stateN) {
	    ITK_WRN("Too old world state") ;
	    ITK_DBG(t) ;
	    ITK_DBG(worktime) ;
	    ITK_DBG(stateN) ;
	 }

	 // checking old state
	 if      (t < worktime) {
	    r = &state(t) ;
	 }
	 else {
	    if (t > worktime) {
	       estState(t-1,False) ;
	    }
	    r = &estimate1(t,False) ;
	 }
      }
      if(lockp) mutex.unlock();
      
      return *r ;
   }
      
   //------------------------------------------------------------
   // WorldModel:: estimate1
   //   estimate state(t) from state(t-1)
   //   suppose state(t-1) is calculated
   //   return state(t)

   WorldModel::State & WorldModel::estimate1(Int t, Bool lockp) {
      State * r ;

      if(lockp) mutex.lock() ;
      {
	 State & pstate = state(t-1) ;
	 State & cstate = state(t) ;
	 r = &cstate ;

	 cstate.reset() ;

	 cstate.time() = t ;

	 cstate.self().copyFrom(pstate.self()) ;
	 cstate.self().pos() += cstate.self().vel() ;
	 cstate.self().vel() *= decayPlayerVel ;
	 cstate.self().posConf *= decayConf ;
	 cstate.self().velConf *= decayConf ;
	    
	 cstate.ball().copyFrom(pstate.ball()) ;
	 cstate.ball().pos() += cstate.ball().vel() ;
	 cstate.ball().vel() *= decayBallVel ;
	 cstate.ball().posConf *= decayConf ;
	 cstate.ball().velConf *= decayConf ;

	 worktime = t + 1 ;
      }
      if(lockp) mutex.unlock() ;
	    
      return *r ;
   } ;

   //------------------------------------------------------------
   // WorldModel:: put sensored data

   void WorldModel::putSensedData(Int t, const FieldState & fs, Bool lockp) {
      if(lockp) mutex.lock() ;

      {
	 State & pstate = estState(t-1,False) ;
	 State & cstate = estState(t,False) ;

	 if(fs.self().posConf > 0.0) {
	    cstate.self().bdir() = fs.self().bdir() ;
	    cstate.self().hdir() = fs.self().hdir() ;
	 }
	 if(fs.self().posConf > cstate.self().posConf) {
	    cstate.self().pos() = fs.self().pos() ;
	    cstate.self().posConf = fs.self().posConf ;
	 }
	 if(fs.self().velConf > cstate.self().velConf) {
	    cstate.self().vel() = fs.self().vel() ;
	    cstate.self().velConf = fs.self().velConf ;
	 }

	 if(fs.ball().posConf > cstate.ball().posConf) {
	    cstate.ball().pos() = fs.ball().pos() ;
	    cstate.ball().posConf = fs.ball().posConf ;

	    Vec v ; v.set(0,0) ;
	    Flt n = 0.0 ;
	    Flt decay = 1.0 ;
	    Flt conf = 1.0 ;
	    Flt confDecay = 1.0 ; // !!!
	    for(UInt i = 0 ; i < 3 ; i++) {
	       v += (estState(t-i,False).ball().pos() - 
		     estState(t-i-1,False).ball().pos()) * decay * conf ;
	       n += conf ;

	       conf *= confDecay ; 
	       decay *= decayBallVel ;
	    }

	    cstate.ball().vel() = v * (decayBallVel / n) ;

	    cstate.ball().velConf = 
	       cstate.ball().posConf * pstate.ball().posConf ;
	 } ;

	 worktime = t + 1 ;

      } 
      if(lockp) mutex.unlock() ;
   } ;
} ;


