// -*- Mode:C++ -*-
//Header:
//File: WorldModel.h
//Title: WorldModel.h
//Author: NODA, Itsuki
//Date: 2001/07/27
//

//ModifyHistory:
// 2001/07/27: Start to create this file
//EndModifyHistory:

#ifndef _saloo_WorldModel_h_
#define _saloo_WorldModel_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"
#include "SensorInfo.h"
#include "BasePlayer.h"
#include "estimatePos.h"
#include "FieldState.h"

namespace Saloo {
   
   using namespace Itk ;
   using namespace Soccer ;

   //======================================================================
   // class WorldModel

   class WorldModel : public WithDescriber {

	 //============================================================
	 // WorldModel:: class State
      public:
	 class State : public FieldState {

	       //----------------------------------------
	       // WorldModel::State:: member

	    public:
	       Int time_ ;
	       ITK_DEF_ACCESS(Int & time(), { return time_ ; }) ;

	       //----------------------------------------
	       // WorldModel::State:: reset
	    public:
	       void reset() {
		  time() = -1 ;

		  self().time().init() ;
		  self().posConf = 0 ;
		  self().velConf = 0 ;

		  ball().time().init() ;
		  ball().posConf = 0 ;
		  ball().velConf = 0 ;
	       } ;
	 } ;
	 
	 typedef State::Object Object ;
      
	 //------------------------------------------------------------
	 // WorldModel:: member

	 class DummyMutex {
	    public: void lock() {} ;
	    public: void unlock() {} ;
	 } ;

      public:
	 Int worktime ;
	 DummyMutex mutex ;

	 Flt decayConf ;
	 Flt decayPlayerVel ;
	 Flt decayBallVel ;

	 //----------------------------------------
	 // WorldModel:: state, ball, self
      public:
	 static const Int stateN = 64 ;
	 State state_[stateN] ;

	 ITK_DEF_ACCESS(State & state(Int t), { 
	    if(t < 0) 
	       return state_[0] ;
	    else 
	       return state_[t % stateN]; 
	 }) ;
	 ITK_DEF_ACCESS(Object & ball(Int t), { return state(t).ball() ; });
	 ITK_DEF_ACCESS(Object & self(Int t), { return state(t).self() ; });

	 //------------------------------------------------------------
	 // WorldModel:: constructor
      public:
	 WorldModel() {
	    init() ;
	 } 

	 //------------------------------------------------------------
	 // WorldModel:: init
      public:
	 void init() {
	    worktime = 0 ;
	    decayConf = 0.5 ;
	    decayPlayerVel = 0.4 ;
	    decayBallVel = 0.94 ;
	 } ;

	 //------------------------------------------------------------
	 // WorldModel:: estState, estBall, estSelf
      public:
	 virtual State & estState(Int t, Bool lockp = True) ;

	 Object & estBall(Int t, Bool lockp = True) { 
	    return estState(t,lockp).ball() ; 
	 } ;

	 Object & estSelf(Int t, Bool lockp = True) { 
	    return estState(t,lockp).self() ; 
	 } ;

	 //------------------------------------------------------------
	 // WorldModel:: estimate1
	 //   estimate state(t) from state(t-1)
	 //   suppose state(t-1) is calculated
	 //   return state(t)
      public:
	 virtual State & estimate1(Int t, Bool lockp = True) ;
	    
	 //------------------------------------------------------------
	 // WorldModel:: put sensored data
      public:
	 void putSensedData(Int t, const FieldState & fs, Bool lockp = True) ;

   } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif
