// -*- Mode:C++ -*-
//Header:
//File: estimatePos.cc
//Title: estimate position
//Author: NODA, Itsuki
//Date: 2001/03/05
//

//ModifyHistory:
// 2001/03/05: Start to create this file
//EndModifyHistory:

#include "estimatePos.h"

namespace Saloo {

   //======================================================================
   // estimate pos dir vec (1) in abstruct orientation

   Bool estimatePosDirVelAbs1(PosDir& posdir, Vec& vel, 
			      Bool& outOfFieldP,
			      const SeeInfo& sinf) {
      Flt dist = 2.0 * SC_PITCH_LENGTH ;
      const SeeInfo::Object * line = ITK_NULLPTR ;
      const SeeInfo::Object * flag = ITK_NULLPTR ;

      outOfFieldP = True ;

      for(UInt i = 0 ; i < sinf.objN ; i++) {
	 const SeeInfo::Object & obj = sinf.obj[i] ;
	 if(obj.type == SeeInfo::Object::Flag || 
	    obj.type == SeeInfo::Object::Goal) {
	    if(!obj.inViewConeP) continue ;
	    if(Abs(dist) > Abs(obj.dist)) {
	       flag = &obj ;
	       dist = obj.dist ;
	    }
	 } else if (obj.type == SeeInfo::Object::Line) {
	    if(isNull(line)) {
	       outOfFieldP = False ;
	       line = &obj ; 
	    } else {
	       outOfFieldP = True ;
	    }
	    i++ ;
	 }
      }

      if(isNull(line) || isNull(flag)) return False ;

      Vec flag_pos = flag->id.flag->pos ;

      switch(line->id.line) {
	 case LID_t: {
	    if (outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       posdir.pos.y = (- SC_HALF_PITCH_W 
			       - Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	       posdir.pos.y = (- SC_HALF_PITCH_W 
			       + Abs(line->dist * Sin(line->dir))) ;
	    }

	    Flt diff_x = sqrt(Abs(sq(flag->dist) -
				  sq(flag_pos.y - posdir.pos.y))) ;
	    Flt obj_dir = normalizeAngle(flag->dir + posdir.hdir) ;
     
	    if(Abs(obj_dir) < 90.0)
	       posdir.pos.x = flag_pos.x - diff_x ;
	    else
	       posdir.pos.x = flag_pos.x + diff_x ;
	    break ;
	 }
	 case LID_b: {
	    if (outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	       posdir.pos.y = (  SC_HALF_PITCH_W 
				 + Abs(line->dir * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       posdir.pos.y = (  SC_HALF_PITCH_W 
				 - Abs(line->dir * Sin(line->dir))) ;
	    }

	    Flt diff_x = sqrt(Abs(sq(flag->dist) -
				  sq(flag_pos.y - posdir.pos.y))) ;
	    Flt obj_dir = normalizeAngle(flag->dir + posdir.hdir) ;
      
	    if(Abs(obj_dir) < 90.0)
	       posdir.pos.x = flag_pos.x - diff_x ;
	    else
	       posdir.pos.x = flag_pos.x + diff_x ;
	    break ;
	 }
	 case LID_l: {
	    if (outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	       posdir.pos.x = (- SC_HALF_PITCH_L 
			       - Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	       posdir.pos.x = (- SC_HALF_PITCH_L 
			       + Abs(line->dist * Sin(line->dir))) ;
	    }

	    Flt diff_y = sqrt(Abs(sq(flag->dist) -
				  sq(flag_pos.x - posdir.pos.x))) ;
	    Flt obj_dir = normalizeAngle(flag->dir + posdir.hdir) ;
      
	    if(obj_dir > 0.0)
	       posdir.pos.y = flag_pos.y - diff_y ;
	    else
	       posdir.pos.y = flag_pos.y + diff_y ;
	    break ;
	 }
	 case LID_r: {
	    if (outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	       posdir.pos.x = (  SC_HALF_PITCH_L 
				 + Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	       posdir.pos.x = (  SC_HALF_PITCH_L 
				 - Abs(line->dist * Sin(line->dir))) ;
	    }

	    Flt diff_y = sqrt(Abs(sq(flag->dist) -
				  sq(flag_pos.x - posdir.pos.x))) ;
	    Flt obj_dir = normalizeAngle(flag->dir + posdir.hdir) ;
      
	    if(obj_dir > 0.0)
	       posdir.pos.y = flag_pos.y - diff_y ;
	    else
	       posdir.pos.y = flag_pos.y + diff_y ;
	    break ;
	 }
	 default: return False ;
      }

      return True ;
   }

   //--------------------------------------------------
   // adjust team orientation

   Bool estimatePosDirVelTeam1(PosDir& posdir, Vec& vel,
			       Bool& outOfFieldP,
			       const SeeInfo& sinf, const Side side) {
      if(estimatePosDirVelAbs1(posdir,vel,outOfFieldP,sinf)) {
	 orientationAbs2Team(posdir,vel,side) ;
	 return True ;
      } else {
	 return False ;
      } 
   }


//======================================================================
// estimate pos dir vec (2) in abstruct orientation

   static inline Flt confidentFactor(const Flt dist) { 
      return (1.0 / (sq(dist)+0.1)) ;
   } ;

   Bool estimatePosDirVelAbs2(PosDir& posdir, Flt& posConf, 
			      Vec& vel, Flt& velConf,
			      Bool& outOfFieldP,
			      const SeeInfo& sinf) {
      Vec p(0.0,0.0) ;
      Vec pN(0.0,0.0) ;
      Vec v(0.0,0.0) ;
      Flt vN = 0.0 ;

      const SeeInfo::Object *line = ITK_NULLPTR ;
   
      outOfFieldP = True ;

      //------------------------------
      // find a reference line 

      for(UInt i = 0 ; i < sinf.objN ; i++) {
	 if(sinf.obj[i].type == SeeInfo::Object::Line) {
	    if(isNull(line)) { 
	       line = &(sinf.obj[i]) ; 
	       outOfFieldP = False ;
	    } else { 
	       outOfFieldP = True ;
	    }
	 }
      }
   
      if(isNull(line)) return False ;

      //------------------------------
      // estimate hdir from line

      switch(line->id.line) {
	 case LID_t :{
	    Flt ey ;
	    if(outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = (- SC_HALF_PITCH_W -
		     Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	       ey = (- SC_HALF_PITCH_W +
		     Abs(line->dist * Sin(line->dir))) ;
	    }
	    Flt w = confidentFactor(line->dist) ;
	    p.y += ey * w ; pN.y += w ;
	    break ;
	 }
	 case LID_b : {
	    Flt ey ;
	    if(outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	       ey = (  SC_HALF_PITCH_W -
		       Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = (  SC_HALF_PITCH_W +
		       Abs(line->dist * Sin(line->dir))) ;
	    }
	    break ;
	    Flt w = confidentFactor(line->dist) ;
	    p.y += ey * w ; pN.y += w ;
	 }
	 case LID_l : {
	    Flt ex ;
	    if(outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	       ex = (- SC_HALF_PITCH_L -
		     Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	       ex = (- SC_HALF_PITCH_L +
		     Abs(line->dist * Sin(line->dir))) ;
	    }
	    Flt w = confidentFactor(line->dist) ;
	    p.x += ex * w ; pN.x += w ;
	    break ;
	 }
	 case LID_r : {
	    Flt ex ;
	    if(outOfFieldP) {
	       posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	       ex = (  SC_HALF_PITCH_L +
		       Abs(line->dist * Sin(line->dir))) ;
	    } else {
	       posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	       ex = (  SC_HALF_PITCH_L -
		       Abs(line->dist * Sin(line->dir))) ;
	    }
	    Flt w = confidentFactor(line->dist) ;
	    p.x += ex * w ; pN.x += w ;
	    break ;
	 }
	 default: return False ;
      }   


      //------------------------------
      // estimate pos from flags

      for(UInt i = 0 ; i < sinf.objN ; i++) {
	 const SeeInfo::Object& obj = sinf.obj[i] ;

	 if(!(obj.type == SeeInfo::Object::Flag || 
	      obj.type == SeeInfo::Object::Goal)) continue ;
	 if(!(obj.inViewConeP)) continue ;

	 Vec flag_pos = obj.id.flag->pos ;

	 Flt obj_dir = obj.dir + posdir.hdir ;
	 Flt obj_dist = obj.dist ;
	 Vec ep(flag_pos.x - obj_dist * Cos(obj_dir),
		flag_pos.y - obj_dist * Sin(obj_dir)) ;
	 Flt w = confidentFactor(obj_dist) ;

	 p.x += w * ep.x ; p.y += w * ep.y ; 
	 pN.x += w ; pN.y += w ;

	 if(obj.know_dist_chng() && obj.know_dir_chng()) {
	    Flt r = obj.dist_chng ;
	    Flt a = deg2rad(obj.dir_chng) * obj_dist ;

	    v.x += -(w * (r * Cos(obj_dir) - a * Sin(obj_dir))) ;
	    v.y += -(w * (r * Sin(obj_dir) + a * Cos(obj_dir))) ;
	    vN += w ;
	 }
      }

      if(pN.x <= 0.0 || pN.y <= 0.0) return False ;

      if(!(p.finite() && pN.finite())) {
	 ITK_WRN( ITK_HERE << "something wrong") ;
	 ITK_DBG(p) ;
	 ITK_DBG(pN) ;
	 return False ;
      } ;

      posdir.pos.set((p.x / pN.x), (p.y / pN.y)) ;
      posConf = Min(pN.x,pN.y) ;
      
      if(vN < 0.00001) { vel.set(0.0,0.0) ; } 
      else             { vel.set(v.x/vN, v.y/vN) ; }
      velConf = vN ;

      return True ;
   }

   //--------------------------------------------------
   // adjust team orientation

   Bool estimatePosDirVelTeam2(PosDir& posdir, Flt& posConf,
			       Vec& vel, Flt& velConf,
			       Bool& outOfFieldP,
			       const SeeInfo& sinf, const Side side) {
      if(estimatePosDirVelAbs2(posdir,posConf,vel,velConf,outOfFieldP,sinf)) {
	 orientationAbs2Team(posdir,vel,side) ;
	 return True ;
      } else {
	 return False ;
      } 
   }

//======================================================================
// estimate pos dir vec (3) in abstruct orientation

   static const Flt estPosStepDelta = 0.5 ;
   static const Flt estPosEps = 0.001 ;
   static const Flt estPosCutOff = 0.05 ;

   Bool estimatePosDirVelAbs3(PosDir& posdir, Flt& posConf, 
			      Vec& vel, Flt& velConf,
			      Bool& outOfFieldP,
			      const SeeInfo& sinf) {

      const Flt qstep = 0.01 ;
      const UInt loop_max = 1000 ;

      Vec &ppos = posdir.pos ;
      // ppos.x = 0.0 ; ppos.y = 0.0 ; // use preset value 

      UInt c = 0 ;
      while(True) {
	 Vec ded ;
	 ded.x = 0.0 ; ded.y = 0.0 ;

	 Flt max_rdr = 0.0 ;
	 for(UInt i = 0 ; i < sinf.objN ; i++) {
	    const SeeInfo::Object& obj = sinf.obj[i] ;

	    if(!(obj.type == SeeInfo::Object::Flag || 
		 obj.type == SeeInfo::Object::Goal)) continue ;
	    if(!obj.inViewConeP) continue ;

	    const Vec &fpos = obj.id.flag->pos ;
	    const Flt &r0 = obj.dist ;

	    Flt r = distance(ppos,fpos) ;
	    Flt dr = Abs(r-r0) ;
	    Flt rdr = dr / (r0 + estPosEps);
	    if(dr > r0 * qstep) rdr = dr ;
	    if(rdr > max_rdr) max_rdr = rdr ;

	    Flt a = 2.0 * (r - r0) / (sq(r0) * r + estPosEps) ;

	    ded.x += a * (ppos.x - fpos.x) ;
	    ded.y += a * (ppos.y - fpos.y) ;
	 }

	 //ppos -= (ded * scEstPosStepDelta) ;
	 ppos -= (ded.unit() * max_rdr) ;

	 //if(ded.norm() < scEstPosCutOff) break ;
	 if(max_rdr < qstep) break ;

	 if(c >= loop_max) break ;
	 c++ ;
      }

      // Because of a gap between -180 and 180 degrees in angle, 
      // it is better to use x-y coordinate rather than angle value directory
      // ScFloat hdir = 0.0 ;  

      Vec hdirp(0.0,0.0) ;
      Int n = 0 ;

      for(UInt i = 0 ; i < sinf.objN ; i++) {
	 const SeeInfo::Object& obj = sinf.obj[i] ;
	 const Vec &fpos = obj.id.flag->pos ;
      
	 if(!(obj.type == SeeInfo::Object::Flag || 
	      obj.type == SeeInfo::Object::Goal)) continue ;
	 if(!obj.inViewConeP) continue ;

	 Flt adir = fpos.angleFrom(ppos) ;
	 Flt rdir = adir - obj.dir ;
	 hdirp.x += Cos(rdir) ;
	 hdirp.y += Sin(rdir) ;
	 n++ ;
      }
      if(n == 0) {
	 // do not change hdir
      } else {
	 posdir.hdir = ATan(hdirp.y, hdirp.x) ;
      }
      if(n <= 2) {
	 posConf = 0.0 ;
      } else {
	 posConf = 1.0 ;
      }
   
      return True ;
   }

   //--------------------------------------------------
   // adjust team orientation

   Bool estimatePosDirVelTeam3(PosDir& posdir, Flt& posConf,
			       Vec& vel, Flt& velConf,
			       Bool& outOfFieldP,
			       const SeeInfo& sinf, const Side side) {
      if(estimatePosDirVelAbs3(posdir,posConf,vel,velConf,outOfFieldP,sinf)) {
	 orientationAbs2Team(posdir,vel,side) ;
	 return True ;
      } else {
	 return False ;
      } 
   }
}
