// -*- Mode:C++ -*-
//Header:
//File: saloo.cc
//Title: Saloo 
//Author: NODA, Itsuki
//Date: 2001/04/20
//

//ModifyHistory:
// 2001/04/20: Start to create this file
//EndModifyHistory:


#include <stl.h>
#include "SimplePlayer.h"
#include "OnlineCoach.h"
//#include "OfflineCoach.h"

#include "NnPredictor.h"

using namespace Saloo ;
using namespace Itk ;

using namespace Nnn ;

//typedef Geo2D::Vec Vec ;

//======================================================================
// Option 

#define DFLT_TEAMNAME "saloo"

char * TeamName = DFLT_TEAMNAME ;
char * ParamFile = DFLT_TEAMNAME ".param" ;

char * SHost = "localhost" ;
Int SPort = 6000 ;

Bool GoalieP = False ;

Int BallSenseMargin = 10 ;
Int PosSenseMargin = 10 ;
Int PlayerSenseMargin = 10 ;
Int SayBufSize = 4096 ;

Flt PlayerSpeedMax = 1.0 ;

Int DebugLevel = 0 ;

Flt Test2DashPower = 100 ;

void getopt(int & argc, char ** & argv) {
   OptArg opt[] = {
      {"shost",     OptArg::cstringVal, &SHost, 0,   "sserver_hostname"},
      {"sport",	    OptArg::IntVal,     &SPort, 0,   "sserver_portnum"},
      {"goalie",    OptArg::OnOff,	&GoalieP,0,  "[on/off]"},
      {"team",	    OptArg::cstringVal, &TeamName,0, "teamname"},
      {"param",	    OptArg::cstringVal, &ParamFile,0, "parameter_filename"},
      {"debug",	    OptArg::IntVal,	&DebugLevel,0,  "debug_level [0-100]"},

      {"test2",     OptArg::FltVal,	&Test2DashPower,0,  "power"},

      {ITK_NULLPTR}} ;

   myGetOpt(opt,argc,argv) ;
   ITK_SET_VBS(DebugLevel)
   
} 

//======================================================================
// class Param Table

class Param {

      //==================================================
      // Param:: class RoleInfo
   public:
      class RoleInfo {
	    //----------------------------------------
	    // Param::RoleInfo:: type
	 public:
	    enum Type {
	       NoType,
	       Forward,
	       MidFielder,
	       Defender,
	       Goalie,
	       TypeMax } ;

	    static const char * type2name(Type t, Bool shortp = True) {
	       switch(t) {
		  case Forward:		return shortp ? "fw" : "forward" ;
		  case MidFielder:	return shortp ? "mf" : "midfielder" ;
		  case Defender:	return shortp ? "df" : "defender" ;
		  case Goalie:		return shortp ? "gk" : "goalie" ;
		  default:		return shortp ? "--" : "unknown" ;
	       } 
	    } ;

	    static Type name2type(const SimpleSexp * s, Bool shortp = True) {
	       for(UInt i = 0 ; i < static_cast<Int>(TypeMax) ; i++) {
		  Type t = static_cast<Type>(i) ;
		  if(s->equal(type2name(t,shortp))) return t ;
	       }
	       return NoType ;
	    } ;

	    //----------------------------------------
	    // Param::RoleInfo:: member
	 public:
	    UInt unum ;
	    Type type ;
	    Vec init ;
	    Vec home ;

	    //----------------------------------------
	    // Param::RoleInfo:: scanner
	 public:
	    Bool scan(SimpleSexp * sexp) ;
      } ;

      //==================================================
      // Param:: class RoleInfoTable
   public:
      class RoleInfoTable : public vector<RoleInfo> {
	    //----------------------------------------
	    // Param::RoleInfoTable:: iterators
	 public:
	    typedef iterator Iterator ;
	    typedef const_iterator C_Iterator ;
	    typedef reverse_iterator RevIterator ;
	    typedef const_reverse_iterator C_RevIterator ;

      } ;

      //==================================================
      // Param:: body

      //--------------------------------------------------
      // Param:: member
   public:
      RoleInfoTable rinfo ;

      Int ballSenseMargin ;
      Int posSenseMargin ;
      Int playerSenseMargin ;

      Int saybufsize ;

      Int comTimeMargin ; // margin for waitNextBestComTime

      Flt playerMaxAccel ;
      Int maxContDash ;
      
      //--------------------------------------------------
      // Param:: constructor
   public:
      Param() { init() ; } ;

      //--------------------------------------------------
      // Param:: init
   public:
      void init() {
	 ballSenseMargin = BallSenseMargin ;
	 posSenseMargin = PosSenseMargin ;
	 playerSenseMargin = PlayerSenseMargin ;
	 saybufsize = SayBufSize ;

	 comTimeMargin = 50 ;

	 playerMaxAccel = 0.65 ;
	 maxContDash = 3 ;
      } ;

      //--------------------------------------------------
      // Param:: scanner
   public:
      Bool scan(SimpleSexp * sexp) ;
      Bool scan(istream & str) {
	 return scan(SimpleSexp::scan(str)) ;
      } ;
      Bool scanFile(const char * const filename) {
	 ifstream istr(filename);
	 return scan(istr) ;
      } ;

      //--------------------------------------------------
      // Param:: get role by unum
   public:
      RoleInfo * getRoleByUNum(UInt unum) {
	 for(RoleInfoTable::Iterator r = rinfo.begin() ;
	     r < rinfo.end() ; r++) {
	    if((*r).unum == unum) return &(*r) ;
	 }
	 return ITK_NULLPTR ;
      }
      
} ;

//--------------------------------------------------
// Param::scan()

Bool Param::scan(SimpleSexp * sexp) {
   if(!sexp->first()->equal("param")) {
      ITK_WRN("wrong parameter header:" << *sexp) ;
      return False ;
   }

   SimpleSexp * body = sexp->rest() ;

   //------------------------------
   // scan role list

   SimpleSexp *rolelist = body->assoc("role-list")->rest() ;
   UInt len = rolelist->length() ;
   
   rinfo.resize(len) ;

   for(UInt i = 0 ; i < len ; i++) {
      if(!rinfo[i].scan(rolelist->nth(i))) {
	 ITK_WRN("wrong posinfo at " << i << "th entry") ;
	 return False ;
      }
   }
   
   //------------------------------
   // scan misc param

   SimpleSexp * miscparam = body->assoc("misc")->rest() ;
   SimpleSexp * s ;

   s = miscparam->assocValue("ballSenseMargin") ;
   if(!s->isNil()) ballSenseMargin = s->scanAsInt() ;

   s = miscparam->assocValue("posSenseMargin") ;
   if(!s->isNil()) posSenseMargin = s->scanAsInt() ;

   s = miscparam->assocValue("playerSenseMargin") ;
   if(!s->isNil()) playerSenseMargin = s->scanAsInt() ;

   s = miscparam->assocValue("saybufsize") ;
   if(!s->isNil()) saybufsize = s->scanAsInt() ;

   s = miscparam->assocValue("comTimeMargin") ;
   if(!s->isNil()) comTimeMargin = s->scanAsInt() ;

   s = miscparam->assocValue("playerMaxAccel") ;
   if(!s->isNil()) playerMaxAccel = s->scanAsFlt() ;

   s = miscparam->assocValue("maxContDash") ;
   if(!s->isNil()) maxContDash = s->scanAsInt() ;

   return True ;
} ;

//--------------------------------------------------
// Param::RoleInfo::scan()

Bool Param::RoleInfo::scan(SimpleSexp * sexp) {
   if(!sexp->first()->equal("role")) {
      ITK_WRN("wrong roleinfo heading") ;
      return False ;
   }
   
   SimpleSexp * body = sexp->rest() ;
   
   SimpleSexp * v;

   v = body->assocValue("type") ;
   if(v->isNil()) { ITK_WRN("roleinfo has no type") ; return False ; }
   type = name2type(v->first()) ;
   if(type == NoType) { ITK_WRN("unknown role type:" << *v) ; }

   v = body->assocValue("unum") ;
   if(v->isNil()) { ITK_WRN("roleinfo has no unum") ; return False ; }
   unum = v->scanAsInt() ;

   v = body->assocValue("init") ;
   if(v->isNil()) { ITK_WRN("roleinfo has no init") ; return False ; }
   init.x = v->nth(0)->scanAsFlt() ;
   init.y = v->nth(1)->scanAsFlt() ;

   v = body->assocValue("home") ;
   if(v->isNil()) { ITK_WRN("roleinfo has no home") ;  return False ; }
   home.x = v->nth(0)->scanAsFlt() ;
   home.y = v->nth(1)->scanAsFlt() ;

   return True ;

} ;

//======================================================================
// class PlayPhase

static VbsLevel VbsPlayPhase = 5 ;

class PlayPhase : public WithDescriber {

   public:
      enum Type {
	 Nop,
	 Dango,
	 FixedPosition,
	 DeployToHome,
	 PP_Max,
      } ;
      static const char * tname(Type t) {
	 switch(t) {
	    case Nop :			return "nop" ;
	    case Dango : 		return "dango" ;
	    case FixedPosition : 	return "fixed_pos" ;
	    case DeployToHome : 	return "deploy_to_home" ;
	    default :			return "(none)" ;
	 }
      } ;

   public:
      CycleStep time ;
      Type type ;
      Flt _farg1 ;  Flt _farg2 ;  Flt _farg3 ;
      Int _iarg1 ;  Int _iarg2 ;  Int _iarg3 ;

      //--------------------------------------------------
      // access
   public:
      ITK_DEF_ACCESS(Flt & farg1(), { return _farg1 ; }); 
      ITK_DEF_ACCESS(Flt & farg2(), { return _farg2 ; }); 
      ITK_DEF_ACCESS(Flt & farg3(), { return _farg3 ; }); 
      ITK_DEF_ACCESS(Int & iarg1(), { return _iarg1 ; }); 
      ITK_DEF_ACCESS(Int & iarg2(), { return _iarg2 ; }); 
      ITK_DEF_ACCESS(Int & iarg3(), { return _iarg3 ; }); 

      //--------------------------------------------------
      // describe
   public:
      virtual void describe(ostream & ostr, 
			    const Bool detailp = True) const {
	 ostr << "#PlayPhase[" << tname(type) << "]" ;

	 if(detailp) {
	    ostr << endl ;
	    ostr << "\t" "farg1=" << farg1() << endl ;
	    ostr << "\t" "farg2=" << farg2() << endl ;
	    ostr << "\t" "farg3=" << farg3() << endl ;

	    ostr << "\t" "iarg1=" << iarg1() << endl ;
	    ostr << "\t" "iarg2=" << iarg2() << endl ;
	    ostr << "\t" "iarg3=" << iarg3() << endl ;
	 } 
      } ;
} ;

class PlayPhaseStack : public StackT<PlayPhase> { 
   public:
      MutexCondVar mtxcv ;
      
   public:
      virtual ~PlayPhaseStack() {} ;

   public:
      void lock() { return mtxcv.lock() ; } ;
      void unlock() { return mtxcv.unlock() ; } ;
      void signal() { return mtxcv.signal() ; } ;
      void broadcast() { return mtxcv.broadcast() ; } ;
      Int wait() { return mtxcv.wait() ; } ;
      Int lockWait() { return mtxcv.lockWait() ; } ;
} ;


//======================================================================
// class Saloo

class Saloo0 : public SimplePlayer {

      //============================================================
      // Saloo0:: class TargetInfo
   public:
      class TargetInfo : public WithDescriber {
	 public:
	    Bool dribblep ;
	    Pos pos ;
	    Flt score ;
	    Flt power ;
	    char * name ;
	    Int aux ;

	 public:
	    virtual void describe(ostream & ostr, 
				  const Bool detailp = True) const {
	       ostr << "#TargetInfo[" << pos << (dribblep ? "(D)" : "(P)")
		    << ":score=" << score 
		    << "|" << name << "(" << aux << ")]" ;

	       if(detailp) ostr << endl ;
	    } ;
      } ;

      //============================================================
      // Saloo0:: class CutPosInfo
   public:
      class CutPosInfo : public WithDescriber {
	 public:
	    Vec pos ;
	    Int maxdash ;

	 public:
	    CutPosInfo() {} ;
	    CutPosInfo(const Vec & p, const Int & m) {
	       pos = p ; maxdash = m ;
	    } ;

	 public:
	    virtual void describe(ostream & ostr, 
				  const Bool detailp = True) const {
	       ostr << "#CutPosInfo[" << pos << ",maxdash=" << maxdash << "]" ;
	    }
      } ;

      //------------------------------------------------------------
      // Saloo0:: member
   public:
      Param param ;
      Param::RoleInfo * role ;

      PlayPhaseStack ppstack ;

      //------------------------------------------------------------
      // Saloo0:: hear utility
   public:
      SimpleSexp::Heap hearItemHeap ;
      Mutex mtxHear ;

      //------------------------------------------------------------
      // Saloo0:: say process and utility
   public:
      SimpleSexp::Heap sayItemHeap ;
      SimpleSexp * saylist ;
      Mutex mtxSay ;

      Thread * thrdSay ;

   public:
      void clearSay(Bool lockp = False) {
	 if(lockp) mtxSay.lock() ;
	 {
	    sayItemHeap.clear() ;
	    saylist = SimpleSexp::Nil ;
	 }
	 if(lockp) mtxSay.unlock() ;
      } ;

      SimpleSexp * addSayItem(SimpleSexp * s) {
	 return sayItemHeap.addFirst(s,saylist) ;
      } ;

      Thread * forkSay() ;
      static void * runSay(void *p) ;
      void * sayMainLoop() ;

      //------------------------------------------------------------
      // Saloo0:: member: NnPredictor
   public:
      BallPredictor ballPredictor ;

      //------------------------------------------------------------
      // Saloo0:: constructor
   public:
      Saloo0() { ballPredictor.setup(worldmodel) ; } ;

      //------------------------------------------------------------
      // Saloo0:: run 
   public:
      void run() ;
      void mainloop() ;

      void playDango() ;
      void playFixedPosition() ;
      
      void playGoalie() ;

      //------------------------------------------------------------
      // Saloo0:: test run
   public:
      void playTest1() ;
      void playTest2() ;
      void playTest3() ;
      void playTest4() ;

      //------------------------------------------------------------
      // Saloo0:: utility
   public:
      Bool checkKickOff() ;

      Bool checkPos() ;
      Bool checkBall() ;

      Flt badnessOfTarget(const Pos & target) ;
      Flt goodnessOfTarget(const Pos & target) ;
      Flt scoreOfTarget(const Pos & target) ;

      Bool compareTargetScore(TargetInfo & target, const Pos & newpos,
			      char * name = ITK_NULLPTR, Int aux = 0,
			      Bool dribblep = False);
      TargetInfo findBestTargetPos() ; // used in playFixedPosition

      Bool shouldChase() ;      // used in playFixedPosition

      Bool isClosestToBall(Int predict = 0, Bool ignoreGoalie = True) ;

      Pos homepos() ;

      //------------------------------------------------------------
      // Saloo0:: goalie's utility
   public:
      Bool maybeSuccessCatch() {
	 if(isLeft(side)) {
	    return (playmode()->id == SC_WID_GoalieCatchL ||
		    playmode()->id == SC_WID_FreeKickL) ;
	 } else {
	    return (playmode()->id == SC_WID_GoalieCatchR ||
		    playmode()->id == SC_WID_FreeKickR) ;
	 }
      } ;
      
      //------------------------------------------------------------
      // Saloo0:: find ball cut pos
   public:
      CutPosInfo cutBallPos() ;

      //------------------------------------------------------------
      // Saloo0:: exactRun : do exact run to the dest position
   public:
      void exactRun(Pos destpos, Flt maxpower,
		    Bool needWaitFirstp = False, 
		    Int maxdash = 0) ;

      //------------------------------------------------------------
      // Saloo0:: lookAroundWell
   public:
      void lookAroundWell() ;

      //------------------------------------------------------------
      // Saloo0:: PlayPhase manipuration

      //----------------------------------------
      // Saloo0:: push base
   public:
      PlayPhase & pushTopPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pushTop() ; 
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;

      //----------------------------------------
      // Saloo0:: push content
   public:
      PlayPhase & pushPlayPhase(const PlayPhase & p, Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pushTop() ; 
	 pp = p ;
	 ITK_VBS(VbsPlayPhase,*this << "push PlayPhase:" << pp) ;
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;
      
      //----------------------------------------
      // Saloo0:: pop

      PlayPhase & popPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pop() ;
	 ITK_VBS(VbsPlayPhase,*this << "pop PlayPhase:" << pp) ;
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;

      //----------------------------------------
      // Saloo0:: push with Flt args
   public:
      PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				Flt arg1,
				Flt arg2 = 0.0,
				Flt arg3 = 0.0,
				Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.farg1() = arg1 ;
	 pp.farg2() = arg2 ;
	 pp.farg3() = arg3 ;

	 PlayPhase & r = pushPlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Saloo0:: push with Int args
   public:
      PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				Int arg1 = 0,
				Int arg2 = 0,
				Int arg3 = 0,
				Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.iarg1() = arg1 ;
	 pp.iarg2() = arg2 ;
	 pp.iarg3() = arg3 ;

	 PlayPhase & r = pushPlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Saloo0:: get current PlayPhase
   public:
      PlayPhase & currentPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.top() ;
	 if(lockp) ppstack.unlock() ;
	 return pp ;
      } ;

      //----------------------------------------
      // Saloo0:: change current PlayPhase
   public:
      PlayPhase & changePlayPhase(const PlayPhase & newpp, Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = currentPlayPhase(False) ;
	 ITK_VBS(VbsPlayPhase,*this << "change PlayPhase from" << pp 
		 << " to " << newpp) ;
	 pp = newpp ;
	 ppstack.broadcast() ;
	 if(lockp) ppstack.unlock() ;
	 return pp ;
      } ;

      //----------------------------------------
      // Saloo0:: change current PlayPhase with Flt args
   public:
      PlayPhase & changePlayPhase(PlayPhase::Type type,
				  Flt arg1,
				  Flt arg2 = 0.0,
				  Flt arg3 = 0.0,
				  Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.farg1() = arg1 ;
	 pp.farg2() = arg2 ;
	 pp.farg3() = arg3 ;

	 PlayPhase & r = changePlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Saloo0:: change current PlayPhase with Int args
   public:
      PlayPhase & changePlayPhase(PlayPhase::Type type,
				  Int arg1 = 0,
				  Int arg2 = 0,
				  Int arg3 = 0,
				  Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.iarg1() = arg1 ;
	 pp.iarg2() = arg2 ;
	 pp.iarg3() = arg3 ;

	 PlayPhase & r = changePlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Saloo0:: confirm or push current PlayPhase
      // 	return True if new PlayPhase is pushed
   public:
      Bool confirmPushPlayPhase(PlayPhase::Type type,
				Int arg1 = 0,
				Int arg2 = 0,
				Int arg3 = 0,
				Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 
	 if(ppstack.isEmpty()) {
	    pushPlayPhase(type,arg1,arg2,arg3,False) ;
	    r = True ;
	 } else {
	    PlayPhase & pp = ppstack.top() ;
	    if(pp.type == type && pp.iarg1() == arg1 && pp.iarg2() == arg2 
	       && pp.iarg3() == arg3) {
	       r = False ;
	    } else {
	       pushPlayPhase(type,arg1,arg2,arg3,False) ;
	       r = True ;
	    }
	 }
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;

      Bool confirmPushPlayPhase(PlayPhase::Type type,
				Flt arg1,
				Flt arg2 = 0.0,
				Flt arg3 = 0.0,
				Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 if(ppstack.isEmpty()) {
	    pushPlayPhase(type,arg1,arg2,arg3,False) ;
	    r = True ;
	 } else {
	    PlayPhase & pp = ppstack.top() ;
	    if(pp.type == type && pp.farg1() == arg1 && pp.farg2() == arg2 
	       && pp.farg3() == arg3) {
	       r = False ;
	    } else {
	       pushPlayPhase(type,arg1,arg2,arg3,False) ;
	       r = True ;
	    }
	 }
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;
	    
      //----------------------------------------
      // Saloo0:: confirm and pop current PlayPhase
      // 	return True if PlayPhase is poped
   public:
      Bool confirmPopPlayPhase(PlayPhase::Type type, Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.top() ;
	 if(pp.type == type) {
	    popPlayPhase(False) ;
	    r = True ;
	 } else {
	    r = False ;
	 } 
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;

      //----------------------------------------
      // Saloo0:: virtual analyzeSensorSee()
   public:
      virtual Bool analyzeSensorSee() ;

      //----------------------------------------
      // Saloo0:: virtual analyzeSensorHear() ;
   public:
      virtual Bool analyzeSensorHear() ;

      //----------------------------------------
      // Saloo0:: for debug
   public:
      virtual Bool send() {
//	 ITK_MSG("sb:" << *this << btime().rtime) ;
//	 ITK_MSG("send:" << *this << TimeVal::now()) ;
	 return SimplePlayer::send() ;
      } ;

} ;

//----------------------------------------------------------------------
// Saloo0::run()

void Saloo0::run() {

   param.scanFile(ParamFile) ;

   init(SHost,SPort,TeamName,GoalieP) ;
   forkSensor() ;
   forkSay() ;

   if(False) {					/* narrow view ? */
      sinfo.senseBody.viewWidth = VW_normal ;
      while(sinfo.senseBody.viewWidth != VW_narrow) {
	 sendChangeView(VW_narrow,VQ_high) ;
	 sleep(0,100) ;
      }
   }

   role = param.getRoleByUNum(unum) ;

   sendMove(role->init) ;

   mainloop() ;
} ;

//----------------------------------------------------------------------
// Saloo0::mainloop()

void Saloo0::mainloop() {
   
   confirmPushPlayPhase(PlayPhase::Nop) ;
   confirmPushPlayPhase(PlayPhase::Dango) ;
   confirmPushPlayPhase(PlayPhase::FixedPosition) ;

   while(True) {
      if(checkKickOff()) { 
	 confirmPushPlayPhase(PlayPhase::DeployToHome) ;
	 continue ; 
      }

      //playTest4() ; continue ;
      //playTest2() ; continue ;
      //playDango() ;
      if(goaliep) {
	 playGoalie() ;
      } else {
	 playFixedPosition() ;
      }
   }
}

//--------------------------------------------------
// footOfPerpendicularLine
//
//	find A foot of a perpendicular line
// 	linepos   O = (Ox,Oy)		; lpos
//	linevec   V = (Vx,Vy)		; pvec
//	point     P = (Px,Py)		; ppos
//	footpoint Q = (Qx,Qy)
//
//	Q = O + r * V    ;   Q is on line (O->V)
//	[P - Q]*V = 0      ;   PQ -| V
//
//	r * |V|^2 = [P - O] * V
//	r = [P - O] * V / |V|^2

static Flt _dummyFltVar_ ;

Vec footOfPerpendicularLine(const Vec & lpos, const Vec & lvec,
			    const Vec & ppos, Flt & a = _dummyFltVar_) {

   if(lvec.norm() < ITK_EPS) return lpos ; /* if lvec is zero vector
					    * return lpos 
					    */
   Vec dpos = ppos - lpos ;

   a = dpos.innerProd(lvec) / lvec.sqnorm() ;
   
   return lpos + (lvec * a) ;
} ;
   

   

//--------------------------------------------------
// Saloo0::cutBallpos

VbsLevel VbsCutBallPos = 6 ;

Saloo0::CutPosInfo Saloo0::cutBallPos() {

   //Int CutBallPredictFactor = 5 ;	/* predict factor */
   Int CutBallPredictFactor = 2 ;	/* predict factor */
   Int CutBallPredictChaseMargin = 5 ;  /* used the ball is going fast */

   /* bvelConf is too low
    * then chase the current ball pos
    */
   //if(estBall(stime().count).velConf < 0.000001) {
   if(estBall(stime().count).velConf < 0.00000001) { 
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "velConf is too small") ;
      return CutPosInfo(estBallPos(stime().count),1) ;
   }

   /* the ball is too far
    * then chase the current ball pos
    */
   if((estBallPos(stime().count) - mypos()).norm() > 30.0) {
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "ball is too far") ;
      return CutPosInfo(estBallPos(stime().count),1) ;
   }


   Int t = stime().count + CutBallPredictFactor ;		
   Vec & bpos = estBallPos(t) ;
   Vec & bvel = estBallVel(t) ;
   Vec & ppos = estSelfPos(t) ;

   /* if ball stops
    * then chase the future ball pos
    */
   if(bvel.norm() < ITK_EPS) { 
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "ball is stopping") ;
      return CutPosInfo(bpos,0) ;	       
   }

   Flt a ;  /* magnify factor of velocity.  
	     * calucurated in the following func */
   Vec footpos = footOfPerpendicularLine(bpos,bvel,ppos,a) ;

   /* if footpoint is behind of the ball or the footpos is very close to
    * the player,
    * then chase the future ball pos
    */
   if(a < 0.0 ||
      (footpos - mypos()).norm() < 2.0) {
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "foot point is behind the ball") ;
      return CutPosInfo(bpos,1) ; 
   }

   Flt l_b2f = (bpos - footpos).norm() ; // dist(ballPos - footPos)
   Flt l_p2f = (ppos - footpos).norm() ; // dist(playerPos - footPos)

   Flt t_b2f = l_b2f / bvel.norm() ; /* same as a? */
   Flt t_p2f = l_p2f / PlayerSpeedMax ;

   /* if player can reach footpoint 10 step before the ball
    * then chase the midpoint of ball pos and footpoint
    */
   if(t_b2f - t_p2f > 10) {
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "player can reach 10 step before the ball") ;
      return CutPosInfo((bpos + footpos) / 2.0,0) ;	
   } 
   /* else if player can reach footpoint before the ball
    * then chase the foot point
    */
   else if (t_b2f - t_p2f > 0) {
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "player can reach before the ball") ;
      return CutPosInfo(footpos,0) ;	
   } 
   /* otherwise, the player can not reach the footpoint before the ball
    * then chase the far future ball pos
    */
   else { 
      ITK_VBS(VbsCutBallPos,*this << "cutBallPos: " 
	      << "ball is faster than player") ;
      return CutPosInfo(estBallPos(t + CutBallPredictFactor 
				   + CutBallPredictChaseMargin),
			0) ;
   }
}

//--------------------------------------------------
// play Fixed Position

Flt HomePosMargin = 5.0 ;
Flt GoaliePosMargin = 1.0 ;

Flt DashFullPower = 100 ;
Flt DashFastPower = 95 ;
Flt DashMiddlePower = 60 ;
Flt DashSlowPower = 30 ;

Flt KickFullPower = 100 ;

VbsLevel VbsPlayType = 6 ;

void Saloo0::playFixedPosition() {

   waitNextBestComTime(param.comTimeMargin) ;

   checkBall() ;
   if      (checkPos()) { 
      ITK_VBS(VbsPlayType,*this << "cheking pos") ;
   } 
   else if (checkBall()) { 
      ITK_VBS(VbsPlayType,*this << "checking ball") ;
   } 
   else if (kickable() && !isOpponentRestart()) {
      ITK_VBS(VbsPlayType,*this << "kicking") ;

      if(isMyRestart()) {
	 lookAroundWell() ;
      }

      TargetInfo target = findBestTargetPos() ;

      kickToAbs(target.power, target.pos) ;

      /* after kick, dash to the same directon 
       */
      waitNextBestComTime(param.comTimeMargin) ;
      if(turnToAbs(target.pos)) waitNextBestComTime(param.comTimeMargin) ;
      dash(DashFastPower) ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
   } 
   else if (shouldChase()) {
      ITK_VBS(VbsPlayType,*this << "chasing") ;
      //turnToAbs(ballpos()) ;
      CutPosInfo cutpos = cutBallPos() ;

      exactRun(cutpos.pos,DashFullPower,False,cutpos.maxdash) ;
#if 0
      if(turnToAbs(cutpos.pos)) waitNextBestComTime(param.comTimeMargin) ;
      Flt loopN = Min(3.0,(dist / PlayerSpeedMax )) ;
      for(Flt i = 0.0 ; i < loopN ; i += 1.0) {
	 if(i > 0.0) waitNextBestComTime(param.comTimeMargin) ;
	 dash(DashFullPower) ;
	 sleep(0,10) ;
	 headToAbs(ballpos()) ;
      }
#endif
   } 
   else if (mypos().distance(homepos()) > HomePosMargin) {
      ITK_VBS(VbsPlayType,*this << "homing") ;
      Flt power = 0.0 ;
      if        (role->type == Param::RoleInfo::Defender && 
		 mypos().x > homepos().x) {
	 power = DashFullPower ;
      } else if (role->type == Param::RoleInfo::Forward && 
		 mypos().x < homepos().x) {
	 power = DashFullPower ;
      } else if (currentPlayPhase().type == PlayPhase::DeployToHome) {
	 power = DashMiddlePower ;
      } else {
	 power = DashSlowPower ;
      }
      exactRun(homepos(),power,False) ;
#if 0
      if(turntoabs(homepos())) waitnextbestcomtime(param.comtimemargin) ;
      dash(power) ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
#endif
   } 
   else {
      ITK_VBS(VbsPlayType,*this << "lookingBall") ;
      confirmPopPlayPhase(PlayPhase::DeployToHome);
      turnToAbs(ballpos()) ;
      headToAbs(ballpos()) ;
   }
}

//----------------------------------------
// play Goalie

void Saloo0::playGoalie() {
   
   waitNextBestComTime(param.comTimeMargin) ;

   checkBall() ;
   
   if (checkPos()) {
      ITK_VBS(VbsPlayType,*this << "checkPos") ;
   } 
   else if (checkBall()) {
      ITK_VBS(VbsPlayType,*this << "checkBall") ;
   }
   else if (gcatchable()) {
      ITK_VBS(VbsPlayType,*this << "catchBall") ;

      gcatchToAbs(ballpos()) ;

      waitNextBestComTime(param.comTimeMargin) ;
      if(maybeSuccessCatch()) { /* catch may succeed */

	 /* the number of move commands should be less than 2.
	  */
	 if(intRand(2) == 0)
	    for(Int i = 0 ; i < 2 ; i++) move(-40,-10) ;// ???
	 else
	    for(Int i = 0 ; i < 2 ; i++) move(-40,10) ;// ???

	 lookAroundWell() ;

	 //ITK_DBG(myposdir()) ;
	 waitNextBestComTime(param.comTimeMargin) ;
	 TargetInfo target = findBestTargetPos() ;
	 //target.power = Max(target.power,100.0) ;
	 target.power = Max(target.power,50.0) ;
	 if(target.pos.x < -40.0) {
	    ITK_MSG("!!!!!!!! why target is in behind ?????" << target.pos) ;
	 }

	 kickToAbs(target.power, target.pos) ;
	 waitNextBestComTime(param.comTimeMargin) ;
	 kickToAbs(target.power, target.pos) ;

	 exactRun(homepos(),DashFullPower,True) ;
      }
   }
   else if (kickable()) {
      ITK_VBS(VbsPlayType,*this << "kick") ;

      if(isMyRestart()) {
	 lookAroundWell() ;
      }

      TargetInfo target = findBestTargetPos() ;
      kickToAbs(target.power, target.pos) ;
   }
   else if (isInPenaltyArea(ballpos(),2.0) ||
	    (ballpos() - owngoalpos()).norm() < 17 && shouldChase()) {
      							// chase !!
      ITK_VBS(VbsPlayType,*this << "chase") ;

      CutPosInfo cutpos = cutBallPos() ;

      exactRun(cutpos.pos,DashFullPower,False,cutpos.maxdash) ;
#if 0
      if(turnToAbs(cutpos.pos)) waitNextBestComTime(param.comTimeMargin) ;
      dash(DashFullPower) ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
#endif
   } 
   else if (mypos().distance(homepos()) > GoaliePosMargin) {
      ITK_VBS(VbsPlayType,*this << "homing") ;
      if(turnToAbs(homepos())) waitNextBestComTime(param.comTimeMargin) ;
      dash(DashMiddlePower) ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
   } 
   else {
      ITK_VBS(VbsPlayType,*this << "lookingBall") ;
      //confirmPopPlayPhase(PlayPhase::DeployToHome);
      turnToAbs(ballpos()) ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
   }
} ;      
      
//----------------------------------------
// look around well

void Saloo0::lookAroundWell() {
   for(Int i = 0 ; i < 3 ; i++) {
      if(!isMyRestart()) return ;
      waitNextBestComTime(param.comTimeMargin) ;
      turnTo(0.0) ;
   }
   
   for(UInt i = 0 ; i < 5 ; i++) {
      for(Flt dir = -80.0 ; dir <= 80.0 + ITK_EPS ; dir += 20.0) {
	 if(!isMyRestart()) return ;
	 waitNextBestComTime(param.comTimeMargin) ;
	 turnTo(0.0) ;
	 sleep(0,10) ;
	 headTo(dir) ;
	 cvSee.lockWait() ;
      }
   }
   for(Int i = 0 ; i < 10 ; i++) {
      if(!isMyRestart()) return ;
      waitNextBestComTime(param.comTimeMargin) ;
      turnTo(0.0) ;
      sleep(0,10) ;
      headTo(0.0) ;
   }
} ;

//----------------------------------------
// exact Run

void Saloo0::exactRun(Pos destpos, Flt maxpower,
		      Bool needWaitFirstp, Int maxdash) {

   // for debug
   //Pos origpos = mypos() ;

   if(maxdash == 0) { maxdash = param.maxContDash ; } ;

   if(needWaitFirstp) waitNextBestComTime(param.comTimeMargin) ;

   if(turnToAbs(destpos)) waitNextBestComTime(param.comTimeMargin) ;
   
   Flt maxacc = param.playerMaxAccel ;
   Flt amp = maxacc / 100.0 ; /* magnification factor for dash power 
			       * dash 100 -> v +=  0.65 m/step
			       */

   Int i = 0 ;
   Flt dist = destpos.distance(mypos()) ;
   for(i = 0 ; i < maxdash && dist > 0.1 ; i++) {
      if(i != 0) waitNextBestComTime(param.comTimeMargin) ;

      /* if the ball is kickable, then emergency exit
       */
      if((goaliep && gcatchable()) || kickable()) return ;

//      Bool needSecondDash = False ;
      Flt pow ;
      if(dist < myspeed()) { /* need negative dash */
	 //ITK_MSG("negative dash!!") ;
	 pow = (dist - myspeed()) / amp ;
      } else if ((dist - myspeed()) < maxacc) {
	 pow = (dist - myspeed()) / amp ;
//	 needSecondDash = True ;
      } else {
	 pow = maxpower ;
      }
      dash(pow) ;
      Flt run = (myspeed() + pow) * amp ;
      dist -= run ;
      sleep(0,10) ;
      headToAbs(ballpos()) ;
   }
} ;



//----------------------------------------
// home position

Flt ShiftYPosMax = 5.0 ;

Pos Saloo0::homepos() {
   Pos r = role->home ;

   /* modify y-pos by ballpos
    */
   r.y += ShiftYPosMax * (ballpos().y/(SC_PITCH_WIDTH/2.0)) ;

   /* modify by role
    */
   switch(role->type) {
      case Param::RoleInfo::Forward : {
	 if(isOpponentRestart()) {
	    r.x = Max(ballpos().x - 10,0.0) ;
	 } 
	 else if(ballpos().x > -10) {
	    r.x = Min(ballpos().x + 10,(SC_PITCH_LENGTH/2.0)) ;
	 } else {
	    r.x = 0.0 ;
	 }
	 break;
      } ;
      case Param::RoleInfo::MidFielder : {
	 r.x = ballpos().x ;
	 if(isMyRestart()) { r.x = Min(r.x + 20.0, 45.0) ; }
	 if(isOpponentRestart()) { r.x = Max(r.x - 20.0, -45.0) ; } 
	 break ;
      }
      case Param::RoleInfo::Defender : {
	 if(ballpos().x < 10) {
	    r.x = Max(ballpos().x - 10,(-(SC_PITCH_LENGTH/2.0))) ;
	 } else {
	    r.x = 0.0 ;
	 }
	 if(isMyRestart()) r.x += 10.0 ;
	 break ;
      }
      case Param::RoleInfo::Goalie : {
	 Vec g1 = owngoalpos() ; g1.y -= (SC_GOAL_WIDTH/2.0);/* goalpostL */
	 Vec g2 = owngoalpos() ; g2.y += (SC_GOAL_WIDTH/2.0);/* goalpostR */
	 if(ballpos().x < r.x) {
	    if        (ballpos().y < g1.y) {  // left side
	       r = g1 ;
	       r.x -= 1.0 ;
	    } else if (ballpos().y > g2.y) {  // right side
	       r = g2 ;
	       r.x -= 1.0 ;
	    } else {			      // middle
	       r = ballpos() ;
	    }
	 } else { /* goalie should place on the center of two lines
		   * of ball-goalpostL and ball-goalpostR
		   * x-pos is in the same level of homepos()
		   */
	    Vec u1 = g1 - ballpos() ; u1.normalize(1.0) ;
	    Vec u2 = g2 - ballpos() ; u2.normalize(1.0) ;
	    Vec p0 = r - ballpos() ;
	    Flt a = p0.x / (u1.x + u2.x) ;
	    r.y = ballpos().y + a * (u1.y + u2.y) ;
	 }
	 break ;
      }
      default : {
	 ITK_WRN(*this << "unknown role?") ;
	 ITK_DBG(role->type) ;
	 break ;
      }
   }

   if(isMyKickOff()) {
      r.x = Min(r.x,0.0) ;
   } else if (isOpponentKickOff() && Abs(r.y) < 10.0) {
      r.x = Min(r.x, -10.0) ;
   }
      
   if(playmode()->id == SC_WID_KickOffL || playmode()->id == SC_WID_KickOffR) {
   }
   
   return r ;
} ;

//----------------------------------------
// Flt functions

/* sigmoid
 *               >----- 1.0
 *              /
 *             /	    dy/dx|_{x=0} = t/4
 *   0.0 -----<
 *              ^[thr]	
 */
inline static Flt sigmoid(Flt x, Flt thr = 0.0 , Flt t = 1.0) {
   return 1.0 / (1.0 + exp(- (x - thr)/t)) ;
} ;

/* lsigmoid
 *               +----- height
 *              /
 *             /	    
 *   0.0 -----+
 *     [lower]^  ^[upper]	
 */
inline static Flt lsigmoid(Flt x, Flt lower = -1.0, Flt upper = 1.0, 
			   Flt height = 1.0) {
   if      ( x < lower ) return 0.0 ;
   else if ( x < upper)  return height * (x - lower) / (upper - lower) ;
   else                  return height ;
} ;

/* cutoff
 *   1.0 -----<
 *             \
 *              \	    dy/dx|_{x=0} = -t/4
 *               >------
 *             ^[thr]
 */
inline static Flt cutoff(Flt x, Flt thr = 0.0, Flt t = 1.0) {
   return sigmoid(-x,-thr,t) ;
} ;

/* rampUp
 *               /
 *              /           dy/dx|{x>thr} = a
 *             /
 *   0.0 -----+
 *            ^[thr]
 */
inline static Flt rampUp(Flt x, Flt thr = 0.0, Flt a = 1.0) {
   return (x < thr) ? 0.0 : a * (x - thr);
} ;

/* rampDown
 *       \
 *        \                 dy/dx|{x>thr} = -a
 *         \
 *          +--------- 0.0
 *           ^[thr]
 */
inline static Flt rampDown(Flt x, Flt thr = 0.0, Flt a = 1.0) {
   return (x > thr) ? 0.0 : a * (thr - x);
} ;

/* rampUpDown
 *  [center]v
 *          + [height]
 *         / \        
 *        /   \
 *  -----+     +--------- 0.0
 *             ^[center+width]
 */
inline static Flt rampUpDown(Flt x, Flt center = 0.0,  Flt width = 1.0,
			     Flt height = 1.0) {
   if      (x < center - width) return 0.0 ;
   else if (x > center + width) return 0.0 ;
   else                         return height * (1.0 - Abs(x - center)/width) ;
} ;

/* trapzoid
 *           [tL]v   v[tR]
 *               +---+ [height]
 *              /     \        
 *             /       \
 *  ----------+         +--------- 0.0
 *        [bL]^         ^[bR]
 */
inline static Flt trapzoid(Flt x, /* DAIKEI */
			   Flt bottomL = -2.0, Flt topL = -1.0,
			   Flt topR = 1.0, Flt bottomR = 2.0,
			   Flt height = 1.0) {
   if      (x < bottomL) return 0.0 ;
   else if (x < topL)    return height * (x - bottomL)/(topL - bottomL) ;
   else if (x < topR)    return height ;
   else if (x < bottomR) return height * (x - topR)/(bottomR - topR) ;
   else                  return 0.0 ;
} ;

//----------------------------------------
// calc badness (based on opponent position)
//   the value how opponent is cutting the pass cource

Flt Saloo0::badnessOfTarget(const Pos & target) {
   Pos & refpos = mypos() ;
   Flt score = 0.0 ;
   Flt l = refpos.distance(target) ;
   for(UInt i = 0 ; i < SC_N_PLAYER_PER_TEAM ; i++) {
      FieldState::Object & p = opponent(i) ;
      Int oldness = stime().count - p.time().count ;
      //Flt oldfact = cutoff((Flt)(oldness),10.0,10.0) ; //!!!
      Flt oldfact = cutoff((Flt)(oldness),5.0,1.0) ; //!!!
      Flt angdiff = refpos.angleTo(p.pos(),target) ;
      if(refpos.distance(p.pos()) <= l + 2.0) { //!!! margin
	 //score += oldfact * rampDown(Abs(angdiff),10,2) ; // !!! 
	 score += oldfact * rampDown(Abs(angdiff),5,4) ; // !!! 
      }
   } 
   if(refpos.distance(owngoalpos()) < -40.0) //!!!
      score *= 3.0 ; //!!!

   return score ;
}

//----------------------------------------
// calc goodness (based on field position)

Flt Saloo0::goodnessOfTarget(const Pos & target) {
   Pos & refpos = mypos() ;
   Flt score = 0.0 ;

   Flt r = refpos.distance(target) ;
   //score += rampUpDown(r, 15, 30, 5) ; //!!! 15 is best dist ?
   score += trapzoid(r, 0, 15, 20, 40) ;

   Flt dx = target.x - refpos.x ;
   //score += 5 * sigmoid(dx,-20.0,20.0) ; //!!! fore is best?
   score += lsigmoid(dx,-20.0,20.0,10.0) - 0.5 ; //!!! fore is best?

   Flt dist2goal = target.distance(goalpos()) ;
   score += 5 * cutoff(dist2goal,10,5) ; // !!! near to goal is good?

   Flt dist2owngoal = target.distance(owngoalpos()) ;
   score -= rampDown(dist2owngoal, 30) ; // !!! near to owngoal is bad?

   if(Abs(target.y) > SC_PITCH_WIDTH/2.0) {
      score -= 30.0 ;	//!!! badness of outside
      score -= 100.0 * sigmoid(target.x,30.0,10.0) ;// !!!
   }

   if(isInPenaltyArea(ballpos()) ||
      ballpos().x < -40.0 ||
      ballpos().distance(owngoalpos()) < 30.0) { /* when in dangerous area */

      Pos dpos = target - ballpos() ;

      if(dpos.x <= 0.0) {
	 score -= 100.0 ; /* never backpass */
      }
      else {
	 if        (ballpos().y > 1.0) { // !!!
	    if((dpos.y / dpos.x) < -0.5) score -= 100.0 ; /* never kick to 
							   * center */
	 } else if (ballpos().y < -1.0) { // !!!
	    if((dpos.y / dpos.x) > 0.5)  score -= 100.0 ; /* never kick to
							   * center */
	 }
      }
   }

   return score ;
} 

Flt Saloo0::scoreOfTarget(const Pos & target) {
   Flt good = goodnessOfTarget(target) ;
   Flt bad = badnessOfTarget(target) ;
   ITK_VBS(10,ITK_HERE << endl 
	   << "\t" << "I am " << *this << endl 
	   << "\t" << "target=" << target << endl 
	   << "\t" << "goodness=" << good << endl
	   << "\t" << "badness=" << bad) ;
   return good - bad ;
} ;

//----------------------------------------
// Find Base Target Pos

Bool Saloo0::compareTargetScore(TargetInfo & target, const Pos & newpos,
				char * name, Int aux, Bool dribblep) {
   Flt newscore = scoreOfTarget(newpos) ;
   if(newscore > target.score) {
      target.pos = newpos ;
      target.score = newscore ;
      target.name = name ;
      target.aux = aux ;
      target.power = Min(KickFullPower,
			 3.0 * target.pos.distance(mypos())) ; // !!!
      target.dribblep = dribblep ;
      return True ;
   } else {
      return False ;
   }
} 

Saloo0::TargetInfo Saloo0::findBestTargetPos() {
//   return goalpos() ;

   TargetInfo target ;
   target.dribblep = False ;
   target.pos = goalpos() ;
   target.score = scoreOfTarget(goalpos()) ;
   target.name = "goal" ;
   target.aux = 0 ;

   //goal corner
   compareTargetScore(target,goalpos() + Pos(0,-5.0),"goalL") ; // !!!
   compareTargetScore(target,goalpos() + Pos(0, 5.0),"goalR") ; // !!!

   // outside
   compareTargetScore(target,Pos(ballpos().x + 10, -30),"clearL") ; //!!!
   compareTargetScore(target,Pos(ballpos().x + 10,  30),"clearR") ; //!!!

   // here , power should be full power
   target.power = KickFullPower ;

   // teammate
   for(UInt i = 0 ; i < SC_N_PLAYER_PER_TEAM ; i++) {
      if(i == 0) continue ; // skip goalie !!!
      FieldState::Object & p = teammate(i) ;
      if(p.time().isNewer(stime(),10)) { //!!!
	 compareTargetScore(target,p.pos() + Pos(1,0),"teammate",i+1) ; // !!!
      }
   }

   // dribble possibility
   if(mypos().distance(owngoalpos()) > 30.0 && 
      mypos().distance(goalpos()) > 15) {
      target.score *= 0.5 ;
      for(Flt ang = -60.0 ; ang <= 60.0 ; ang += 10.0) {
	 Pos drpos = mypos() + Pos(Cos(ang),Sin(ang)) * 5.0 ;
	 compareTargetScore(target, drpos, "dribble",0,True) ;
      }
   }

   ITK_VBS(0,*this << endl << "   " << " target=" << target) ;
   ITK_DBG(myposdir()) ;
   ITK_DBG(ballpos()) ;

   return target ;
}

//----------------------------------------
// should chase?

Bool Saloo0::shouldChase() {
   if(isOpponentRestart()) return False ;
   if(isMyRestart()) {
      if(isClosestToBall(5,False)) return True ;
      else return False ;
   } else {
      if(isClosestToBall(5)) return True ;	// predict 5 step future
      else return False ;
   }
} 

//----------------------------------------
// closest to ball

Bool Saloo0::isClosestToBall(Int predict, Bool ignoreGoalie) {
   //Pos bpos = ballpos() ;
   Pos bpos = estBallPos(stime().count + predict) ;
   Flt bdist = mypos().distance(bpos) ;
   for(UInt i = 0 ; i < SC_N_PLAYER_PER_TEAM ; i++) { //
      if(ignoreGoalie && i == 0) continue ;	// suppose goalie' unum=1
      if(i+1 != unum) {
	 FieldState::Object & p = teammate(i) ;
	 if(p.time().isNewer(stime(),param.playerSenseMargin)) { 
	    if(p.pos().distance(bpos) < bdist)
	       return False ;
	 }
      }
   }

   return True ;
}

//------------------------------
// Saloo0::checkKickOff()

Bool Saloo0::checkKickOff() {
   if(shouldBeInitPos()) {
      if(mypos().distance(role->init) > 1.0) {
	 sendMove(role->init) ;
	 sleep(0,100) ;
      } else {
	 turn(30) ;
      }
      return True ;
   } else {
      return False ;
   }
}

//------------------------------
// Saloo0::findBall()

Bool Saloo0::checkBall() {

//   ITK_VBS(1,"checking ball:" << ballpos() << "(" << ball().time() << "/" << stime() << ")") ;

   if(isBallSeenRecently(param.ballSenseMargin)) {
      return False ;
   } else {
      //turnToAbs(ballpos()) ;
      headToAbs(ballpos()) ;
      for(UInt i = 0 ; i < 8 ; i++) {
	 cvSee.lockWait() ;
	 if(isBallSeenRecently(param.ballSenseMargin)) return True ;
	 if(sinfo.senseBody.viewWidth == VW_narrow) {
	    turn(40) ;
	 } else {
	    turn(80) ;
	 }
      }
      ITK_WRN(*this << "can not find the ball") ;
      return False ;
   }
} ;

//------------------------------
// Saloo0::confirmPos()

Bool Saloo0::checkPos() {

//   ITK_VBS(10,"checking pos:" << myposdir() << "(" << vtime() << ")" ) ;

   if(isMyPosEstimatedRecently(param.posSenseMargin)) {
      return False ;
   } else {
      for(UInt i = 0 ; i < 8 ; i++) {
	 cvSee.lockWait() ;
	 if(isMyPosEstimatedRecently(param.posSenseMargin)) return True ;
	 if(sinfo.senseBody.viewWidth == VW_narrow) {
	    turn(40) ;
	 } else {
	    turn(80) ;
	 }
      }
      ITK_WRN(*this << "can not estimate the position") ;
      return False ;
   }
} ;

//----------------------------------------------------------------------
// Saloo0::virtual analyzeSensorSee()

static VbsLevel VbsChangeView = 5 ;

Bool Saloo0::analyzeSensorSee() {
   Bool r = SimplePlayer::analyzeSensorSee() ;

   /* --------------------
    * changing view mode
    */
   if      (sinfo.senseBody.viewWidth == VW_normal &&
	    //currentState().time().count == stime().count) {
	    ball().time().count == stime().count) {
      sendChangeView(VW_narrow,VQ_high) ;
      ITK_VBS(VbsChangeView,*this << "change_view:normal->narrow") ;
   }
   else if (sinfo.senseBody.viewWidth == VW_narrow &&
	    //currentState().time().count == stime().count) {
	    ball().time().count != stime().count) {
      sendChangeView(VW_normal,VQ_high) ;
      ITK_VBS(VbsChangeView,*this << "change_view:narrow->normal") ;
   }

   /* -------------------- 
    * Generating Messages
    */
   FieldState & cstate = currentState() ;
   if(cstate.self().posConf > 0.001 &&    //*
      cstate.ball().posConf > 0.001) {
      mtxSay.lock() ; {
	 addSayItem(
	    sayItemHeap.list(
	       sayItemHeap.newSymbol("ball"),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("time"),
		  sayItemHeap.newInt(cstate.ball().time().count)),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("pos"),
		  sayItemHeap.newFlt(cstate.ball().pos().x),
		  sayItemHeap.newFlt(cstate.ball().pos().y)),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("posConf"),
		  sayItemHeap.newFlt(cstate.ball().posConf)))) ;
	 addSayItem(
	    sayItemHeap.list(
	       sayItemHeap.newSymbol("self"),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("time"),
		  sayItemHeap.newInt(cstate.ball().time().count)),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("pos"),
		  sayItemHeap.newFlt(cstate.self().pos().x),
		  sayItemHeap.newFlt(cstate.self().pos().y)),
	       sayItemHeap.list(
		  sayItemHeap.newSymbol("posConf"),
		  sayItemHeap.newFlt(cstate.self().posConf)))) ;
      } ; mtxSay.unlock() ;
   }

#if 0
   if(stime().count % 20 == 5 && unum == 3) {
      ITK_DBG(estBallPos(stime().count)) ;
      ITK_DBG(estBallPos(stime().count+1)) ;
      ITK_DBG(ballPredictor.estimatePos(stime().count+1,5,20,0.1)) ;
   }
#endif

   return r ;
} ;


//----------------------------------------------------------------------
// Saloo0::virtual analyzeSensorHear()

Bool Saloo0::analyzeSensorHear() {
   Bool r = SimplePlayer::analyzeSensorHear() ;

   if(sinfo.hear.sender == HearInfo::Player) {
      /* message() begins by '"' because the server quote the orig. messsage
       */
      SubString message = sinfo.hear.message().shiftedSubString(1) ;

      if(message.hasPrefix("(") && 
	 message.shiftedSubString(1).hasPrefix(teamname)) {

	 mtxHear.lock() ; 
	 {
	    Buffer b(message) ;
	    SimpleSexp * sexp = hearItemHeap.scan(b) ;
	    
	    //ITK_VBS(5,*this << " receive our message." << sexp) ;

	    /* message sample
	     * (saloo 11 (time 37) 
	     *   (self (time 37) (pos -0.0349464 -0.398392) (posConf 3.84887))
	     *   (ball (time 37) (pos -8.40887e-05 8.57101e-05)
	     *         (posConf 9.62217)))
	     */

	    { /* analyze ball info */

	       SimpleSexp * binfo = sexp->nthcdr(2)->assoc("ball") ;
	       UInt btime = binfo->rest()->assocValue("time")->scanAsInt() ;
	       Pos bpos ; 
	       bpos.x = binfo->rest()->assoc("pos")->nth(1)->scanAsFlt() ;
	       bpos.y = binfo->rest()->assoc("pos")->nth(2)->scanAsFlt() ;
	       Flt bconf = binfo->rest()->assocValue("posConf")->scanAsFlt() ;

	       if(ball().time() < btime ||
		  (ball().time() == btime && ball().posConf < bconf)) {
		  ball().time() = btime ;
		  ball().pos() = bpos ;
		  ball().posConf = bconf ;

		  /*
		   * test
		   */
		  //worldmodel.putSensedData(btime,currentState()) ; // tricky ?
	       }
	    }

	    { /* analyze player info */

	       UInt unum = sexp->nth(1)->scanAsInt() ;

	       SimpleSexp * pinfo = sexp->nthcdr(2)->assoc("self") ;
	       UInt ptime = pinfo->rest()->assocValue("time")->scanAsInt() ;
	       Pos ppos ; 
	       ppos.x = pinfo->rest()->assoc("pos")->nth(1)->scanAsFlt() ;
	       ppos.y = pinfo->rest()->assoc("pos")->nth(2)->scanAsFlt() ;
	       //Flt pconf = pinfo->rest()->assocValue("posConf")->scanAsFlt() ;

	       FieldState::Object & p = teammate(unum-1) ;
	       if(p.time() <= ptime) {
		  p.time() = ptime ;
		  p.pos() = ppos ;
	       }
	    }
	 }
	 mtxHear.unlock() ;
	    
      }
   }

   return r ;
}


//----------------------------------------------------------------------
// Saloo0::say processes

//--------------------------------------------------
// Saloo0::forkSay()

Thread * Saloo0::forkSay() {
   return thrdSay = new Thread(Saloo0::runSay,this) ;
} ;

//--------------------------------------------------
// Saloo0::runSay

void * Saloo0::runSay(void *p) {
   Saloo0 * player = static_cast<Saloo0 *>(p) ;
   return player->sayMainLoop() ;
};

//--------------------------------------------------
// Saloo0::sayMainLoop

void * Saloo0::sayMainLoop() {

   Buffer saybuf(param.saybufsize) ;
   
   clearSay(True) ;
   
   while(True) {
      sleep(0,100) ;	// should be change

      mtxSay.lock() ;
      {
	 if(!saylist->isNil()) {
	    sayItemHeap.addFirst(
	       sayItemHeap.list(sayItemHeap.newSymbol("time"),
				sayItemHeap.newInt(stime().count)),
	       saylist) ;
	    sayItemHeap.addFirst(sayItemHeap.newInt(unum),saylist) ;
	    sayItemHeap.addFirst(sayItemHeap.newSymbol(teamname),saylist) ;
	    {
	       saybuf.clear() ;
	       ostrstream ostr(saybuf.tail,saybuf.restLength()) ;
	       ostr << saylist ;
	       saybuf.setLength(ostr.pcount()) ;
	       saybuf.terminate() ;

	       say(saybuf) ;
	    }
	    clearSay(False) ;
	 }
      }
      mtxSay.unlock() ;
   }

   return this ;
} ;

//======================================================================
// for Test

//--------------------------------------------------
// play Test

//------------------------------
// test 1

void Saloo0::playTest1() {
   head(-90) ;
   for(UInt i = 0 ; i < 180 ; i++) {
      head(1) ;
      sleep(0,100) ;
      ITK_DBG(mypos()) ;
      ITK_DBG(mybdir()) ;
      ITK_DBG(myhdir()) ;
      ITK_DBG(myneck()) ;
   } 
}


//------------------------------
// test 2
//   to test dash effect

void Saloo0::playTest2() {

   OnlineCoach coach ;
   coach.init(SHost, 6002, TeamName) ;
   coach.forkSensor() ;

   waitNextBestComTime(param.comTimeMargin) ;
   turnTo(0.0) ;
   waitNextBestComTime(param.comTimeMargin) ;
   turnTo(0.0) ;
   waitNextBestComTime(param.comTimeMargin) ;
   turnTo(0.0) ;

   for(UInt i = 0 ; i < 30 ; i++) {
      waitNextBestComTime(param.comTimeMargin) ;
      ITK_DBG(vtime()) ;
      ITK_DBG(coach.sinfo.teammate(unum-1).pos()) ;
      ITK_DBG(myspeed()) ;
      dash(Test2DashPower) ;
   }

   exit(0) ;
}

//------------------------------
// test 3
//   to test dash effect

void Saloo0::playTest3() {

   OnlineCoach coach ;
   coach.init(SHost, 6002, TeamName) ;
   coach.forkSensor() ;

   for(Flt bdir = -180.0 ; bdir <= 181.0 ; bdir += 10.0) {
      for(Flt hdir = -90 ; hdir <= 91.0 ; hdir += 10.0) {
	 waitNextBestComTime(param.comTimeMargin) ;
	 turnTo(bdir) ;
	 waitNextBestComTime(param.comTimeMargin) ;
	 ITK_DBG(hdir - myneck()) ;
	 head(hdir - myneck()) ;
	 sleep(0,100) ;
	 ITK_DBG(myposdir()) ;
      }
   }
   exit(0) ;
}

//------------------------------
// test 4
//   to test dash effect

void Saloo0::playTest4() {

#if 0
   OnlineCoach coach ;
   coach.init(SHost, 6002, TeamName) ;
   coach.forkSensor() ;
#endif

   lookAroundWell() ;
   sleep(1,0) ;
   ITK_DBG(myhdir()) ;
   ITK_DBG(mybdir()) ;
   ITK_DBG(myneck()) ;

   exit(0) ;
}


//--------------------------------------------------
// play Dango

void Saloo0::playDango() {
   if(checkBall()) {
      /* do nothing */
   } else if(kickable()) {
      kickToAbs(100,goalpos()) ;
   } else {
      turnToAbs(ballpos()) ;
      dash(100) ;
   }
}

//======================================================================
// play Saloo

int playSaloo(int argc, char** argv) {
   Saloo0 player ;
   
   player.run() ;

   return 0 ;
 
} ;

//======================================================================
// main

int main(int argc, char** argv) {
   getopt(argc,argv) ;

   playSaloo(argc,argv) ;

   return 0 ;
} ;



