// -*- Mode:C++ -*-
//Header:
//File: sctypes.cc
//Title: Base Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/04
//

//ModifyHistory:
// 2001/03/04: Start to create this file
//EndModifyHistory:

#include "sctypes.h"

namespace Soccer {

   //------------------------------------------------------------
   // Side

   Side scanAsSide(const SimpleSexp & s) {
      if(s.equal("l") || s.equal("left")) return Left ;
      if(s.equal("r") || s.equal("right")) return Right ;
      return NoSide ;
   } 

   //------------------------------------------------------------
   // Referee Message 

   ScWord ScWordTable[] = SC_WORD_TABLE ;

   ScWord * scanAsScWord(const SubString & s) {
      for(ScWord * word = ScWordTable ; !isNull(word->name) ; word++) {
	 if(s == word->name) return word ;
      }

      if      (s.hasPrefix("goal_l_")) return &(SC_WORD(GoalL)) ;
      else if (s.hasPrefix("goal_r_")) return &(SC_WORD(GoalR)) ;

      return ITK_NULLPTR ;
   } 

   ScWord * scanAsScWord(const SimpleSexp & s) {
      for(ScWord * word = ScWordTable ; !isNull(word->name) ; word++) {
	 if(s.equal(word->name)) return word ;
      }

      if      (s.scanAsConstSymbol().hasPrefix("goal_l_")) 
	 return &(SC_WORD(GoalL)) ;
      else if (s.scanAsConstSymbol().hasPrefix("goal_r_")) 
	 return &(SC_WORD(GoalR)) ;

      return ITK_NULLPTR ;
   } 

   ScWord * getScWord(const ScWordID & id) {
      for(ScWord * word = ScWordTable ; !isNull(word->name) ; word++) {
	 if(word->id == id) return word ;
      }
      return ITK_NULLPTR ;
   } 

   //------------------------------------------------------------
   // Landmark

   Landmark LandmarkTable[] = LANDMARK_TABLE ;

   Landmark * scanAsLandmark(const SimpleSexp & s) {
      for(Landmark * l = LandmarkTable ; !isNull(l->namestr) ; l++) {
	 if(s.equal(l->namestr)) return l ;
      }
      return ITK_NULLPTR ;
   } 

   Landmark * scanAsLandmark(const SubString & s) {
      for(Landmark * l = LandmarkTable ; !isNull(l->namestr) ; l++) {
	 if(s == l->namestr) return l ;
      }
      return ITK_NULLPTR ;
   } 

   Landmark * getLandmark(LandmarkID id) {
      for(Landmark * l = LandmarkTable ; !isNull(l->namestr) ; l++) {
	 if(id == l->id) return l ;
      }
      return ITK_NULLPTR ;
   } 

      

} 
